"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const task_names_1 = require("hardhat/builtin-tasks/task-names");
const glob_1 = __importDefault(require("glob"));
const fs_1 = __importDefault(require("fs"));
const scripts_runner_1 = require("hardhat/internal/util/scripts-runner");
const sdk_1 = require("@graphprotocol/sdk");
const gre_1 = require("@graphprotocol/sdk/gre");
const CONFIG_TESTS = 'test/e2e/deployment/config/**/*.test.ts';
const INIT_TESTS = 'test/e2e/deployment/init/**/*.test.ts';
// Built-in test & run tasks don't support GRE arguments
// so we pass them by overriding GRE config object
const setGraphConfig = (args, hre) => {
    const greArgs = [
        'graphConfig',
        'l1GraphConfig',
        'l2GraphConfig',
        'addressBook',
        'disableSecureAccounts',
        'fork',
    ];
    for (const arg of greArgs) {
        if (args[arg]) {
            if (arg === 'graphConfig') {
                const l1 = (0, sdk_1.isGraphL1ChainId)(hre.config.networks[hre.network.name].chainId);
                hre.config.graph[l1 ? 'l1GraphConfig' : 'l2GraphConfig'] = args[arg];
            }
            else {
                hre.config.graph[arg] = args[arg];
            }
        }
    }
};
(0, gre_1.greTask)('e2e', 'Run all e2e tests')
    .addFlag('skipBridge', 'Skip bridge tests')
    .setAction(async (args, hre) => {
    let testFiles = [
        ...new glob_1.default.GlobSync(CONFIG_TESTS).found,
        ...new glob_1.default.GlobSync(INIT_TESTS).found,
    ];
    if (args.skipBridge) {
        testFiles = testFiles.filter(file => !/l1|l2/.test(file));
    }
    // Disable secure accounts, we don't need them for this task
    hre.config.graph.disableSecureAccounts = true;
    setGraphConfig(args, hre);
    await hre.run(task_names_1.TASK_TEST, {
        testFiles: testFiles,
    });
});
(0, gre_1.greTask)('e2e:config', 'Run deployment configuration e2e tests').setAction(async (args, hre) => {
    const files = new glob_1.default.GlobSync(CONFIG_TESTS).found;
    // Disable secure accounts, we don't need them for this task
    hre.config.graph.disableSecureAccounts = true;
    setGraphConfig(args, hre);
    await hre.run(task_names_1.TASK_TEST, {
        testFiles: files,
    });
});
(0, gre_1.greTask)('e2e:init', 'Run deployment initialization e2e tests').setAction(async (args, hre) => {
    const files = new glob_1.default.GlobSync(INIT_TESTS).found;
    // Disable secure accounts, we don't need them for this task
    hre.config.graph.disableSecureAccounts = true;
    setGraphConfig(args, hre);
    await hre.run(task_names_1.TASK_TEST, {
        testFiles: files,
    });
});
(0, gre_1.greTask)('e2e:scenario', 'Run scenario scripts and e2e tests')
    .addPositionalParam('scenario', 'Name of the scenario to run')
    .addFlag('skipScript', 'Don\'t run scenario script')
    .setAction(async (args, hre) => {
    setGraphConfig(args, hre);
    const script = `test/e2e/scenarios/${args.scenario}.ts`;
    const test = `test/e2e/scenarios/${args.scenario}.test.ts`;
    console.log(`> Running scenario: ${args.scenario}`);
    console.log(`- script file: ${script}`);
    console.log(`- test file: ${test}`);
    if (!args.skipScript) {
        if (fs_1.default.existsSync(script)) {
            await (0, scripts_runner_1.runScriptWithHardhat)(hre.hardhatArguments, script, [
                args.addressBook,
                args.graphConfig,
                args.l1GraphConfig,
                args.l2GraphConfig,
                args.disableSecureAccounts,
            ]);
        }
        else {
            console.log(`No script found for scenario ${args.scenario}`);
        }
    }
    if (fs_1.default.existsSync(test)) {
        await hre.run(task_names_1.TASK_TEST, {
            testFiles: [test],
        });
    }
    else {
        throw new Error(`No test found for scenario ${args.scenario}`);
    }
});
(0, gre_1.greTask)('e2e:upgrade', 'Run upgrade tests')
    .addPositionalParam('upgrade', 'Name of the upgrade to run')
    .addFlag('post', 'Wether to run pre/post upgrade scripts')
    .setAction(async (args, hre) => {
    setGraphConfig(args, hre);
    await runUpgrade(args, hre, args.post ? 'post' : 'pre');
});
// eslint-disable-next-line @typescript-eslint/no-explicit-any
async function runUpgrade(args, hre, type) {
    const script = `test/e2e/upgrades/${args.upgrade}/${type}-upgrade.ts`;
    const test = `test/e2e/upgrades/${args.upgrade}/${type}-upgrade.test.ts`;
    console.log(`> Running ${type}-upgrade: ${args.upgrade}`);
    console.log(`- script file: ${script}`);
    console.log(`- test file: ${test}`);
    // Run script
    if (fs_1.default.existsSync(script)) {
        console.log(`> Running ${type}-upgrade script: ${script}`);
        await (0, scripts_runner_1.runScriptWithHardhat)(hre.hardhatArguments, script, [
            args.addressBook,
            args.graphConfig,
            args.l1GraphConfig,
            args.l2GraphConfig,
            args.disableSecureAccounts,
            args.fork,
        ]);
    }
    // Run test
    if (fs_1.default.existsSync(test)) {
        console.log(`> Running ${type}-upgrade test: ${test}`);
        await hre.run(task_names_1.TASK_TEST, {
            testFiles: [test],
        });
    }
}
