"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const hardhat_1 = __importDefault(require("hardhat"));
const chai_1 = require("chai");
const ethers_1 = require("ethers");
const fixtures_1 = require("../lib/fixtures");
const sdk_1 = require("@graphprotocol/sdk");
const { AddressZero } = ethers_1.constants;
const MAX_PPM = 1000000;
describe('Curation:Config', () => {
    let me;
    let governor;
    let fixture;
    let contracts;
    const graph = hardhat_1.default.graph();
    const defaults = graph.graphConfig.defaults;
    before(async function () {
        [me] = await graph.getTestAccounts();
        ({ governor } = await graph.getNamedAccounts());
        fixture = new fixtures_1.NetworkFixture(graph.provider);
        contracts = await fixture.load(governor);
    });
    beforeEach(async function () {
        await fixture.setUp();
    });
    afterEach(async function () {
        await fixture.tearDown();
    });
    describe('defaultReserveRatio', function () {
        it('should set `defaultReserveRatio`', async function () {
            // Set right in the constructor
            (0, chai_1.expect)(await contracts.Curation.defaultReserveRatio()).eq(defaults.curation.reserveRatio);
            // Can set if allowed
            const newValue = (0, sdk_1.toBN)('100');
            await contracts.Curation.connect(governor).setDefaultReserveRatio(newValue);
            (0, chai_1.expect)(await contracts.Curation.defaultReserveRatio()).eq(newValue);
        });
        it('reject set `defaultReserveRatio` if out of bounds', async function () {
            const tx1 = contracts.Curation.connect(governor).setDefaultReserveRatio(0);
            await (0, chai_1.expect)(tx1).revertedWith('Default reserve ratio must be > 0');
            const tx2 = contracts.Curation.connect(governor).setDefaultReserveRatio(MAX_PPM + 1);
            await (0, chai_1.expect)(tx2).revertedWith('Default reserve ratio cannot be higher than MAX_PPM');
        });
        it('reject set `defaultReserveRatio` if not allowed', async function () {
            const tx = contracts.Curation.connect(me).setDefaultReserveRatio(defaults.curation.reserveRatio);
            await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
        });
    });
    describe('minimumCurationDeposit', function () {
        it('should set `minimumCurationDeposit`', async function () {
            // Set right in the constructor
            (0, chai_1.expect)(await contracts.Curation.minimumCurationDeposit()).eq(defaults.curation.minimumCurationDeposit);
            // Can set if allowed
            const newValue = (0, sdk_1.toBN)('100');
            await contracts.Curation.connect(governor).setMinimumCurationDeposit(newValue);
            (0, chai_1.expect)(await contracts.Curation.minimumCurationDeposit()).eq(newValue);
        });
        it('reject set `minimumCurationDeposit` if out of bounds', async function () {
            const tx = contracts.Curation.connect(governor).setMinimumCurationDeposit(0);
            await (0, chai_1.expect)(tx).revertedWith('Minimum curation deposit cannot be 0');
        });
        it('reject set `minimumCurationDeposit` if not allowed', async function () {
            const tx = contracts.Curation.connect(me).setMinimumCurationDeposit(defaults.curation.minimumCurationDeposit);
            await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
        });
    });
    describe('curationTaxPercentage', function () {
        it('should set `curationTaxPercentage`', async function () {
            const curationTaxPercentage = defaults.curation.curationTaxPercentage;
            // Set new value
            await contracts.Curation.connect(governor).setCurationTaxPercentage(0);
            await contracts.Curation.connect(governor).setCurationTaxPercentage(curationTaxPercentage);
        });
        it('reject set `curationTaxPercentage` if out of bounds', async function () {
            const tx = contracts.Curation.connect(governor).setCurationTaxPercentage(MAX_PPM + 1);
            await (0, chai_1.expect)(tx).revertedWith('Curation tax percentage must be below or equal to MAX_PPM');
        });
        it('reject set `curationTaxPercentage` if not allowed', async function () {
            const tx = contracts.Curation.connect(me).setCurationTaxPercentage(0);
            await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
        });
    });
    describe('curationTokenMaster', function () {
        it('should set `curationTokenMaster`', async function () {
            const newCurationTokenMaster = contracts.Curation.address;
            await contracts.Curation.connect(governor).setCurationTokenMaster(newCurationTokenMaster);
        });
        it('reject set `curationTokenMaster` to empty value', async function () {
            const newCurationTokenMaster = AddressZero;
            const tx = contracts.Curation.connect(governor).setCurationTokenMaster(newCurationTokenMaster);
            await (0, chai_1.expect)(tx).revertedWith('Token master must be non-empty');
        });
        it('reject set `curationTokenMaster` to non-contract', async function () {
            const newCurationTokenMaster = (0, sdk_1.randomAddress)();
            const tx = contracts.Curation.connect(governor).setCurationTokenMaster(newCurationTokenMaster);
            await (0, chai_1.expect)(tx).revertedWith('Token master must be a contract');
        });
        it('reject set `curationTokenMaster` if not allowed', async function () {
            const newCurationTokenMaster = contracts.Curation.address;
            const tx = contracts.Curation.connect(me).setCurationTokenMaster(newCurationTokenMaster);
            await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
        });
    });
});
