"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const hardhat_1 = __importDefault(require("hardhat"));
const ethers_1 = require("ethers");
const chai_1 = require("chai");
const fixtures_1 = require("../lib/fixtures");
const sdk_1 = require("@graphprotocol/sdk");
const { AddressZero } = ethers_1.constants;
const MAX_PPM = 1000000;
describe('DisputeManager:Config', () => {
    let me;
    let governor;
    let arbitrator;
    const graph = hardhat_1.default.graph();
    const defaults = graph.graphConfig.defaults;
    let fixture;
    let contracts;
    let disputeManager;
    before(async function () {
        [me] = await graph.getTestAccounts();
        ({ governor, arbitrator } = await graph.getNamedAccounts());
        fixture = new fixtures_1.NetworkFixture(graph.provider);
        contracts = await fixture.load(governor);
        disputeManager = contracts.DisputeManager;
    });
    beforeEach(async function () {
        await fixture.setUp();
    });
    afterEach(async function () {
        await fixture.tearDown();
    });
    describe('configuration', () => {
        describe('arbitrator', function () {
            it('should set `arbitrator`', async function () {
                // Set right in the constructor
                (0, chai_1.expect)(await disputeManager.arbitrator()).eq(arbitrator.address);
                // Can set if allowed
                await disputeManager.connect(governor).setArbitrator(me.address);
                (0, chai_1.expect)(await disputeManager.arbitrator()).eq(me.address);
            });
            it('reject set `arbitrator` if not allowed', async function () {
                const tx = disputeManager.connect(me).setArbitrator(arbitrator.address);
                await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
            });
            it('reject set `arbitrator` to address zero', async function () {
                const tx = disputeManager.connect(governor).setArbitrator(AddressZero);
                await (0, chai_1.expect)(tx).revertedWith('Arbitrator must be set');
            });
        });
        describe('minimumDeposit', function () {
            it('should set `minimumDeposit`', async function () {
                const oldValue = defaults.dispute.minimumDeposit;
                const newValue = (0, sdk_1.toBN)('1');
                // Set right in the constructor
                (0, chai_1.expect)(await disputeManager.minimumDeposit()).eq(oldValue);
                // Set new value
                await disputeManager.connect(governor).setMinimumDeposit(newValue);
                (0, chai_1.expect)(await disputeManager.minimumDeposit()).eq(newValue);
            });
            it('reject set `minimumDeposit` if not allowed', async function () {
                const newValue = (0, sdk_1.toBN)('1');
                const tx = disputeManager.connect(me).setMinimumDeposit(newValue);
                await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
            });
        });
        describe('fishermanRewardPercentage', function () {
            it('should set `fishermanRewardPercentage`', async function () {
                const newValue = defaults.dispute.fishermanRewardPercentage;
                // Set right in the constructor
                (0, chai_1.expect)(await disputeManager.fishermanRewardPercentage()).eq(newValue);
                // Set new value
                await disputeManager.connect(governor).setFishermanRewardPercentage(0);
                await disputeManager.connect(governor).setFishermanRewardPercentage(newValue);
            });
            it('reject set `fishermanRewardPercentage` if out of bounds', async function () {
                const tx = disputeManager.connect(governor).setFishermanRewardPercentage(MAX_PPM + 1);
                await (0, chai_1.expect)(tx).revertedWith('Reward percentage must be below or equal to MAX_PPM');
            });
            it('reject set `fishermanRewardPercentage` if not allowed', async function () {
                const tx = disputeManager.connect(me).setFishermanRewardPercentage(50);
                await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
            });
        });
        describe('slashingPercentage', function () {
            it('should set `slashingPercentage`', async function () {
                const qryNewValue = defaults.dispute.qrySlashingPercentage;
                const idxNewValue = defaults.dispute.idxSlashingPercentage;
                // Set right in the constructor
                (0, chai_1.expect)(await disputeManager.qrySlashingPercentage()).eq(qryNewValue);
                (0, chai_1.expect)(await disputeManager.idxSlashingPercentage()).eq(idxNewValue);
                // Set new value
                await disputeManager.connect(governor).setSlashingPercentage(0, 0);
                await disputeManager.connect(governor).setSlashingPercentage(qryNewValue, idxNewValue);
            });
            it('reject set `slashingPercentage` if out of bounds', async function () {
                const tx1 = disputeManager.connect(governor).setSlashingPercentage(0, MAX_PPM + 1);
                await (0, chai_1.expect)(tx1).revertedWith('Slashing percentage must be below or equal to MAX_PPM');
                const tx2 = disputeManager.connect(governor).setSlashingPercentage(MAX_PPM + 1, 0);
                await (0, chai_1.expect)(tx2).revertedWith('Slashing percentage must be below or equal to MAX_PPM');
            });
            it('reject set `slashingPercentage` if not allowed', async function () {
                const tx = disputeManager.connect(me).setSlashingPercentage(50, 50);
                await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
            });
        });
    });
});
