"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const hardhat_1 = __importDefault(require("hardhat"));
const chai_1 = require("chai");
const sdk_1 = require("@graphprotocol/sdk");
describe('EpochManager', () => {
    const graph = hardhat_1.default.graph();
    const defaults = graph.graphConfig.defaults;
    let me;
    let governor;
    let epochManager;
    const epochLength = (0, sdk_1.toBN)('3');
    before(async function () {
        [me, governor] = await graph.getTestAccounts();
        ({ governor } = await graph.getNamedAccounts());
    });
    beforeEach(async function () {
        const { contract: controller } = await (0, sdk_1.deploy)(sdk_1.DeployType.DeployAndSave, governor, {
            name: 'Controller',
        }, graph.addressBook);
        await (0, sdk_1.deploy)(sdk_1.DeployType.DeployAndSave, governor, {
            name: 'GraphProxyAdmin',
        }, graph.addressBook);
        const epochManagerResult = await (0, sdk_1.deploy)(sdk_1.DeployType.DeployWithProxy, governor, {
            name: 'EpochManager',
            args: [controller.address, defaults.epochs.lengthInBlocks],
        }, graph.addressBook, {
            name: 'GraphProxy',
        });
        epochManager = epochManagerResult.contract;
    });
    describe('configuration', () => {
        it('should set `epochLength', async function () {
            // Set right in the constructor
            (0, chai_1.expect)(await epochManager.epochLength()).eq(defaults.epochs.lengthInBlocks);
            // Update and check new value
            const newEpochLength = (0, sdk_1.toBN)('4');
            const currentEpoch = await epochManager.currentEpoch();
            const tx = epochManager.connect(governor).setEpochLength(newEpochLength);
            await (0, chai_1.expect)(tx)
                .emit(epochManager, 'EpochLengthUpdate')
                .withArgs(currentEpoch, newEpochLength);
            (0, chai_1.expect)(await epochManager.epochLength()).eq(newEpochLength);
        });
        it('reject set `epochLength` if zero', async function () {
            // Update and check new value
            const newEpochLength = (0, sdk_1.toBN)('0');
            const tx = epochManager.connect(governor).setEpochLength(newEpochLength);
            await (0, chai_1.expect)(tx).revertedWith('Epoch length cannot be 0');
        });
    });
    describe('epoch lifecycle', function () {
        // Use epochs every three blocks
        // Blocks -> (1,2,3)(4,5,6)(7,8,9)
        // Epochs ->   1    2    3
        beforeEach(async function () {
            await epochManager.connect(governor).setEpochLength(epochLength);
        });
        describe('calculations', () => {
            it('first epoch should be 1', async function () {
                const currentEpoch = await epochManager.currentEpoch();
                (0, chai_1.expect)(currentEpoch).eq(1);
            });
            it('should return correct block number', async function () {
                const currentBlock = await sdk_1.helpers.latestBlock();
                (0, chai_1.expect)(await epochManager.blockNum()).eq(currentBlock);
            });
            it('should return same starting block if we stay on the same epoch', async function () {
                // Move right to the start of a new epoch
                const blocksSinceEpochStart = await epochManager.currentEpochBlockSinceStart();
                const blocksToNextEpoch = epochLength.sub(blocksSinceEpochStart);
                await sdk_1.helpers.mineUpTo((await epochManager.blockNum()).add(blocksToNextEpoch));
                const beforeCurrentEpochBlock = await epochManager.currentEpochBlock();
                // Advance block - will not jump to next epoch
                await sdk_1.helpers.mine();
                const afterCurrentEpochBlock = await epochManager.currentEpochBlock();
                (0, chai_1.expect)(afterCurrentEpochBlock).equal(beforeCurrentEpochBlock);
            });
            it('should return next starting block if we move to the next epoch', async function () {
                const beforeCurrentEpochBlock = await epochManager.currentEpochBlock();
                // Advance blocks to move to the next epoch
                await sdk_1.helpers.mineUpTo(beforeCurrentEpochBlock.add(epochLength));
                const afterCurrentEpochBlock = await epochManager.currentEpochBlock();
                (0, chai_1.expect)(afterCurrentEpochBlock).not.eq(beforeCurrentEpochBlock);
            });
            it('should return next epoch if advance > epochLength', async function () {
                const nextEpoch = (await epochManager.currentEpoch()).add((0, sdk_1.toBN)('1'));
                // Advance blocks and move to the next epoch
                const currentEpochBlock = await epochManager.currentEpochBlock();
                await sdk_1.helpers.mineUpTo(currentEpochBlock.add(epochLength));
                const afterCurrentEpoch = await epochManager.currentEpoch();
                (0, chai_1.expect)(afterCurrentEpoch).eq(nextEpoch);
            });
        });
        describe('progression', () => {
            beforeEach(async function () {
                const currentEpochBlock = await epochManager.currentEpochBlock();
                await sdk_1.helpers.mineUpTo(currentEpochBlock.add(epochLength));
            });
            context('> epoch not run', function () {
                it('should return that current epoch is not run', async function () {
                    (0, chai_1.expect)(await epochManager.isCurrentEpochRun()).eq(false);
                });
                it('should run new epoch', async function () {
                    // Run epoch
                    const currentEpoch = await epochManager.currentEpoch();
                    const tx = epochManager.connect(me).runEpoch();
                    await (0, chai_1.expect)(tx).emit(epochManager, 'EpochRun').withArgs(currentEpoch, me.address);
                    // State
                    const lastRunEpoch = await epochManager.lastRunEpoch();
                    (0, chai_1.expect)(lastRunEpoch).eq(currentEpoch);
                });
            });
            context('> epoch run', function () {
                beforeEach(async function () {
                    await epochManager.runEpoch();
                });
                it('should return current epoch is already run', async function () {
                    (0, chai_1.expect)(await epochManager.isCurrentEpochRun()).eq(true);
                });
                it('reject run new epoch', async function () {
                    const tx = epochManager.runEpoch();
                    await (0, chai_1.expect)(tx).revertedWith('Current epoch already run');
                });
            });
        });
    });
});
