"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const hardhat_1 = __importDefault(require("hardhat"));
const chai_1 = require("chai");
const ethers_1 = require("ethers");
const fixtures_1 = require("../lib/fixtures");
const sdk_1 = require("@graphprotocol/sdk");
describe('BridgeEscrow', () => {
    const graph = hardhat_1.default.graph();
    let governor;
    let tokenReceiver;
    let spender;
    let fixture;
    let contracts;
    let grt;
    let bridgeEscrow;
    const nTokens = (0, sdk_1.toGRT)('1000');
    before(async function () {
        [tokenReceiver, spender] = await graph.getTestAccounts();
        ({ governor } = await graph.getNamedAccounts());
        fixture = new fixtures_1.NetworkFixture(graph.provider);
        contracts = await fixture.load(governor);
        grt = contracts.GraphToken;
        bridgeEscrow = contracts.BridgeEscrow;
        // Give some funds to the Escrow
        await grt.connect(governor).mint(bridgeEscrow.address, nTokens);
    });
    beforeEach(async function () {
        await fixture.setUp();
    });
    afterEach(async function () {
        await fixture.tearDown();
    });
    describe('approveAll', function () {
        it('cannot be called by someone other than the governor', async function () {
            const tx = bridgeEscrow.connect(tokenReceiver).approveAll(spender.address);
            await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
        });
        it('allows a spender to transfer GRT held by the contract', async function () {
            (0, chai_1.expect)(await grt.allowance(bridgeEscrow.address, spender.address)).eq(0);
            const tx = grt
                .connect(spender)
                .transferFrom(bridgeEscrow.address, tokenReceiver.address, nTokens);
            await (0, chai_1.expect)(tx).revertedWith('ERC20: transfer amount exceeds allowance');
            await bridgeEscrow.connect(governor).approveAll(spender.address);
            await (0, chai_1.expect)(grt.connect(spender).transferFrom(bridgeEscrow.address, tokenReceiver.address, nTokens)).to.emit(grt, 'Transfer');
            (0, chai_1.expect)(await grt.balanceOf(tokenReceiver.address)).to.eq(nTokens);
        });
    });
    describe('revokeAll', function () {
        it('cannot be called by someone other than the governor', async function () {
            const tx = bridgeEscrow.connect(tokenReceiver).revokeAll(spender.address);
            await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
        });
        it('revokes a spender\'s permission to transfer GRT held by the contract', async function () {
            await bridgeEscrow.connect(governor).approveAll(spender.address);
            await bridgeEscrow.connect(governor).revokeAll(spender.address);
            // We shouldn't be able to transfer _anything_
            const tx = grt
                .connect(spender)
                .transferFrom(bridgeEscrow.address, tokenReceiver.address, ethers_1.BigNumber.from('1'));
            await (0, chai_1.expect)(tx).revertedWith('ERC20: transfer amount exceeds allowance');
        });
    });
});
