"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const hardhat_1 = __importDefault(require("hardhat"));
const chai_1 = require("chai");
const ethers_1 = require("ethers");
const fixtures_1 = require("../lib/fixtures");
const sdk_1 = require("@graphprotocol/sdk");
const { AddressZero } = ethers_1.constants;
describe('L1GraphTokenGateway', () => {
    const graph = hardhat_1.default.graph();
    let governor;
    let tokenSender;
    let l2Receiver;
    let pauseGuardian;
    let fixture;
    let grt;
    let l1GraphTokenGateway;
    let bridgeEscrow;
    let bridgeMock;
    let inboxMock;
    let outboxMock;
    let routerMock;
    let l2GRTMock;
    let l2GRTGatewayMock;
    let fixtureContracts;
    let l2MockContracts;
    const senderTokens = (0, sdk_1.toGRT)('1000');
    const maxGas = (0, sdk_1.toBN)('1000000');
    const maxSubmissionCost = (0, sdk_1.toBN)('7');
    const gasPriceBid = (0, sdk_1.toBN)('2');
    const defaultEthValue = maxSubmissionCost.add(maxGas.mul(gasPriceBid));
    const emptyCallHookData = '0x';
    const defaultData = ethers_1.utils.defaultAbiCoder.encode(['uint256', 'bytes'], [maxSubmissionCost, emptyCallHookData]);
    const defaultDataNoSubmissionCost = ethers_1.utils.defaultAbiCoder.encode(['uint256', 'bytes'], [(0, sdk_1.toBN)(0), emptyCallHookData]);
    const notEmptyCallHookData = '0x12';
    const defaultDataWithNotEmptyCallHookData = ethers_1.utils.defaultAbiCoder.encode(['uint256', 'bytes'], [maxSubmissionCost, notEmptyCallHookData]);
    before(async function () {
        [tokenSender, l2Receiver] = await graph.getTestAccounts();
        ({ governor, pauseGuardian } = await graph.getNamedAccounts());
        fixture = new fixtures_1.NetworkFixture(graph.provider);
        // Deploy L1
        fixtureContracts = await fixture.load(governor);
        grt = fixtureContracts.GraphToken;
        l1GraphTokenGateway = fixtureContracts.L1GraphTokenGateway;
        bridgeEscrow = fixtureContracts.BridgeEscrow;
        ({ bridgeMock, inboxMock, outboxMock, routerMock } = await fixture.loadL1ArbitrumBridge(governor));
        // Deploy L2 mock
        l2MockContracts = await fixture.loadMock(true);
        l2GRTMock = l2MockContracts.L2GraphToken;
        l2GRTGatewayMock = l2MockContracts.L2GraphTokenGateway;
        // Give some funds to the token sender/router mock
        await grt.connect(governor).mint(tokenSender.address, senderTokens);
        await sdk_1.helpers.setBalance(routerMock.address, ethers_1.utils.parseEther('1'));
    });
    beforeEach(async function () {
        await fixture.setUp();
    });
    afterEach(async function () {
        await fixture.tearDown();
    });
    context('> immediately after deploy', function () {
        describe('calculateL2TokenAddress', function () {
            it('should return address zero as it was not set', async function () {
                (0, chai_1.expect)(await l1GraphTokenGateway.calculateL2TokenAddress(grt.address)).eq(AddressZero);
            });
        });
        describe('outboundTransfer', function () {
            it('reverts because it is paused', async function () {
                const tx = l1GraphTokenGateway
                    .connect(tokenSender)
                    .outboundTransfer(grt.address, l2Receiver.address, (0, sdk_1.toGRT)('10'), maxGas, gasPriceBid, defaultData, {
                    value: defaultEthValue,
                });
                await (0, chai_1.expect)(tx).revertedWith('Paused (contract)');
            });
        });
        describe('finalizeInboundTransfer', function () {
            it('revert because it is paused', async function () {
                const tx = l1GraphTokenGateway
                    .connect(tokenSender)
                    .finalizeInboundTransfer(grt.address, l2Receiver.address, tokenSender.address, (0, sdk_1.toGRT)('10'), defaultData);
                await (0, chai_1.expect)(tx).revertedWith('Paused (contract)');
            });
        });
        describe('setArbitrumAddresses', function () {
            it('is not callable by addreses that are not the governor', async function () {
                const tx = l1GraphTokenGateway
                    .connect(tokenSender)
                    .setArbitrumAddresses(inboxMock.address, routerMock.address);
                await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
            });
            it('rejects setting an EOA as router or inbox', async function () {
                let tx = l1GraphTokenGateway
                    .connect(governor)
                    .setArbitrumAddresses(tokenSender.address, routerMock.address);
                await (0, chai_1.expect)(tx).revertedWith('INBOX_MUST_BE_CONTRACT');
                tx = l1GraphTokenGateway
                    .connect(governor)
                    .setArbitrumAddresses(inboxMock.address, tokenSender.address);
                await (0, chai_1.expect)(tx).revertedWith('ROUTER_MUST_BE_CONTRACT');
            });
            it('sets inbox and router address', async function () {
                const tx = l1GraphTokenGateway
                    .connect(governor)
                    .setArbitrumAddresses(inboxMock.address, routerMock.address);
                await (0, chai_1.expect)(tx)
                    .emit(l1GraphTokenGateway, 'ArbitrumAddressesSet')
                    .withArgs(inboxMock.address, routerMock.address);
                (0, chai_1.expect)(await l1GraphTokenGateway.l1Router()).eq(routerMock.address);
                (0, chai_1.expect)(await l1GraphTokenGateway.inbox()).eq(inboxMock.address);
            });
        });
        describe('setL2TokenAddress', function () {
            it('is not callable by addreses that are not the governor', async function () {
                const tx = l1GraphTokenGateway.connect(tokenSender).setL2TokenAddress(l2GRTMock.address);
                await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
            });
            it('sets l2GRT', async function () {
                const tx = l1GraphTokenGateway.connect(governor).setL2TokenAddress(l2GRTMock.address);
                await (0, chai_1.expect)(tx).emit(l1GraphTokenGateway, 'L2TokenAddressSet').withArgs(l2GRTMock.address);
                (0, chai_1.expect)(await l1GraphTokenGateway.l2GRT()).eq(l2GRTMock.address);
            });
        });
        describe('setL2CounterpartAddress', function () {
            it('is not callable by addreses that are not the governor', async function () {
                const tx = l1GraphTokenGateway
                    .connect(tokenSender)
                    .setL2CounterpartAddress(l2GRTGatewayMock.address);
                await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
            });
            it('sets l2Counterpart which can be queried with counterpartGateway()', async function () {
                const tx = l1GraphTokenGateway
                    .connect(governor)
                    .setL2CounterpartAddress(l2GRTGatewayMock.address);
                await (0, chai_1.expect)(tx)
                    .emit(l1GraphTokenGateway, 'L2CounterpartAddressSet')
                    .withArgs(l2GRTGatewayMock.address);
                (0, chai_1.expect)(await l1GraphTokenGateway.l2Counterpart()).eq(l2GRTGatewayMock.address);
                (0, chai_1.expect)(await l1GraphTokenGateway.counterpartGateway()).eq(l2GRTGatewayMock.address);
            });
        });
        describe('setEscrowAddress', function () {
            it('is not callable by addreses that are not the governor', async function () {
                const tx = l1GraphTokenGateway.connect(tokenSender).setEscrowAddress(bridgeEscrow.address);
                await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
            });
            it('sets escrow', async function () {
                const tx = l1GraphTokenGateway.connect(governor).setEscrowAddress(bridgeEscrow.address);
                await (0, chai_1.expect)(tx)
                    .emit(l1GraphTokenGateway, 'EscrowAddressSet')
                    .withArgs(bridgeEscrow.address);
                (0, chai_1.expect)(await l1GraphTokenGateway.escrow()).eq(bridgeEscrow.address);
            });
        });
        describe('addToCallhookAllowlist', function () {
            it('is not callable by addreses that are not the governor', async function () {
                const tx = l1GraphTokenGateway
                    .connect(tokenSender)
                    .addToCallhookAllowlist(fixtureContracts.RewardsManager.address);
                await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
                (0, chai_1.expect)(await l1GraphTokenGateway.callhookAllowlist(fixtureContracts.RewardsManager.address)).eq(false);
            });
            it('rejects adding an EOA to the callhook allowlist', async function () {
                const tx = l1GraphTokenGateway.connect(governor).addToCallhookAllowlist(tokenSender.address);
                await (0, chai_1.expect)(tx).revertedWith('MUST_BE_CONTRACT');
            });
            it('adds an address to the callhook allowlist', async function () {
                const tx = l1GraphTokenGateway
                    .connect(governor)
                    .addToCallhookAllowlist(fixtureContracts.RewardsManager.address);
                await (0, chai_1.expect)(tx)
                    .emit(l1GraphTokenGateway, 'AddedToCallhookAllowlist')
                    .withArgs(fixtureContracts.RewardsManager.address);
                (0, chai_1.expect)(await l1GraphTokenGateway.callhookAllowlist(fixtureContracts.RewardsManager.address)).eq(true);
            });
        });
        describe('removeFromCallhookAllowlist', function () {
            it('is not callable by addreses that are not the governor', async function () {
                await l1GraphTokenGateway
                    .connect(governor)
                    .addToCallhookAllowlist(fixtureContracts.RewardsManager.address);
                const tx = l1GraphTokenGateway
                    .connect(tokenSender)
                    .removeFromCallhookAllowlist(fixtureContracts.RewardsManager.address);
                await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
                (0, chai_1.expect)(await l1GraphTokenGateway.callhookAllowlist(fixtureContracts.RewardsManager.address)).eq(true);
            });
            it('removes an address from the callhook allowlist', async function () {
                await l1GraphTokenGateway
                    .connect(governor)
                    .addToCallhookAllowlist(fixtureContracts.RewardsManager.address);
                const tx = l1GraphTokenGateway
                    .connect(governor)
                    .removeFromCallhookAllowlist(fixtureContracts.RewardsManager.address);
                await (0, chai_1.expect)(tx)
                    .emit(l1GraphTokenGateway, 'RemovedFromCallhookAllowlist')
                    .withArgs(fixtureContracts.RewardsManager.address);
                (0, chai_1.expect)(await l1GraphTokenGateway.callhookAllowlist(fixtureContracts.RewardsManager.address)).eq(false);
            });
        });
        describe('Pausable behavior', () => {
            it('cannot be paused or unpaused by someone other than governor or pauseGuardian', async () => {
                let tx = l1GraphTokenGateway.connect(tokenSender).setPaused(false);
                await (0, chai_1.expect)(tx).revertedWith('Only Governor or Guardian');
                tx = l1GraphTokenGateway.connect(tokenSender).setPaused(true);
                await (0, chai_1.expect)(tx).revertedWith('Only Governor or Guardian');
            });
            it('cannot be unpaused if some state variables are not set', async function () {
                let tx = l1GraphTokenGateway.connect(governor).setPaused(false);
                await (0, chai_1.expect)(tx).revertedWith('INBOX_NOT_SET');
                await l1GraphTokenGateway
                    .connect(governor)
                    .setArbitrumAddresses(inboxMock.address, routerMock.address);
                tx = l1GraphTokenGateway.connect(governor).setPaused(false);
                await (0, chai_1.expect)(tx).revertedWith('L2_COUNTERPART_NOT_SET');
                await l1GraphTokenGateway
                    .connect(governor)
                    .setL2CounterpartAddress(l2GRTGatewayMock.address);
                tx = l1GraphTokenGateway.connect(governor).setPaused(false);
                await (0, chai_1.expect)(tx).revertedWith('ESCROW_NOT_SET');
            });
            it('can be paused and unpaused by the governor', async function () {
                await fixture.configureL1Bridge(governor, fixtureContracts, l2MockContracts);
                let tx = l1GraphTokenGateway.connect(governor).setPaused(true);
                await (0, chai_1.expect)(tx).emit(l1GraphTokenGateway, 'PauseChanged').withArgs(true);
                (0, chai_1.expect)(await l1GraphTokenGateway.paused()).eq(true);
                tx = l1GraphTokenGateway.connect(governor).setPaused(false);
                await (0, chai_1.expect)(tx).emit(l1GraphTokenGateway, 'PauseChanged').withArgs(false);
                (0, chai_1.expect)(await l1GraphTokenGateway.paused()).eq(false);
            });
            describe('setPauseGuardian', function () {
                it('cannot be called by someone other than governor', async function () {
                    const tx = l1GraphTokenGateway
                        .connect(tokenSender)
                        .setPauseGuardian(pauseGuardian.address);
                    await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
                });
                it('sets a new pause guardian', async function () {
                    const currentPauseGuardian = await l1GraphTokenGateway.pauseGuardian();
                    const tx = l1GraphTokenGateway.connect(governor).setPauseGuardian(pauseGuardian.address);
                    await (0, chai_1.expect)(tx)
                        .emit(l1GraphTokenGateway, 'NewPauseGuardian')
                        .withArgs(currentPauseGuardian, pauseGuardian.address);
                });
                it('allows a pause guardian to pause and unpause', async function () {
                    await fixture.configureL1Bridge(governor, fixtureContracts, l2MockContracts);
                    await l1GraphTokenGateway.connect(governor).setPauseGuardian(pauseGuardian.address);
                    let tx = l1GraphTokenGateway.connect(pauseGuardian).setPaused(true);
                    await (0, chai_1.expect)(tx).emit(l1GraphTokenGateway, 'PauseChanged').withArgs(true);
                    (0, chai_1.expect)(await l1GraphTokenGateway.paused()).eq(true);
                    tx = l1GraphTokenGateway.connect(pauseGuardian).setPaused(false);
                    await (0, chai_1.expect)(tx).emit(l1GraphTokenGateway, 'PauseChanged').withArgs(false);
                    (0, chai_1.expect)(await l1GraphTokenGateway.paused()).eq(false);
                });
            });
        });
    });
    context('> after configuring and unpausing', function () {
        const createMsgData = function (callHookData) {
            const selector = l1GraphTokenGateway.interface.getSighash('finalizeInboundTransfer');
            const params = ethers_1.utils.defaultAbiCoder.encode(['address', 'address', 'address', 'uint256', 'bytes'], [grt.address, tokenSender.address, l2Receiver.address, (0, sdk_1.toGRT)('10'), callHookData]);
            const outboundData = ethers_1.utils.hexlify(ethers_1.utils.concat([selector, params]));
            const msgData = ethers_1.utils.solidityPack([
                'uint256',
                'uint256',
                'uint256',
                'uint256',
                'uint256',
                'uint256',
                'uint256',
                'uint256',
                'uint256',
                'bytes',
            ], [
                (0, sdk_1.toBN)(l2GRTGatewayMock.address),
                (0, sdk_1.toBN)('0'),
                defaultEthValue,
                maxSubmissionCost,
                (0, sdk_1.applyL1ToL2Alias)(tokenSender.address),
                (0, sdk_1.applyL1ToL2Alias)(tokenSender.address),
                maxGas,
                gasPriceBid,
                ethers_1.utils.hexDataLength(outboundData),
                outboundData,
            ]);
            return msgData;
        };
        const createInboxAccsEntry = function (msgDataHash) {
            // The real bridge would emit the InboxAccs entry that came before this one, but our mock
            // emits this, making it easier for us to validate here that all the parameters we sent are correct
            const expectedInboxAccsEntry = ethers_1.utils.keccak256(ethers_1.utils.solidityPack(['address', 'uint8', 'address', 'bytes32'], [inboxMock.address, 9, l1GraphTokenGateway.address, msgDataHash]));
            return expectedInboxAccsEntry;
        };
        const testValidOutboundTransfer = async function (signer, data, callHookData) {
            const tx = l1GraphTokenGateway
                .connect(signer)
                .outboundTransfer(grt.address, l2Receiver.address, (0, sdk_1.toGRT)('10'), maxGas, gasPriceBid, data, {
                value: defaultEthValue,
            });
            // Our bridge mock returns an incrementing seqNum starting at 1
            const expectedSeqNum = 1;
            await (0, chai_1.expect)(tx)
                .emit(l1GraphTokenGateway, 'DepositInitiated')
                .withArgs(grt.address, tokenSender.address, l2Receiver.address, expectedSeqNum, (0, sdk_1.toGRT)('10'));
            const msgData = createMsgData(callHookData);
            const msgDataHash = ethers_1.utils.keccak256(msgData);
            const expectedInboxAccsEntry = createInboxAccsEntry(msgDataHash);
            await (0, chai_1.expect)(tx).emit(inboxMock, 'InboxMessageDelivered').withArgs(1, msgData);
            await (0, chai_1.expect)(tx)
                .emit(bridgeMock, 'MessageDelivered')
                .withArgs(expectedSeqNum, expectedInboxAccsEntry, inboxMock.address, 9, l1GraphTokenGateway.address, msgDataHash);
            const escrowBalance = await grt.balanceOf(bridgeEscrow.address);
            const senderBalance = await grt.balanceOf(tokenSender.address);
            (0, chai_1.expect)(escrowBalance).eq((0, sdk_1.toGRT)('10'));
            (0, chai_1.expect)(senderBalance).eq((0, sdk_1.toGRT)('990'));
        };
        before(async function () {
            await fixture.configureL1Bridge(governor, fixtureContracts, l2MockContracts);
        });
        describe('updateL2MintAllowance', function () {
            it('rejects calls that are not from the governor', async function () {
                const tx = l1GraphTokenGateway
                    .connect(pauseGuardian.address)
                    .updateL2MintAllowance((0, sdk_1.toGRT)('1'), await sdk_1.helpers.latestBlock());
                await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
            });
            it('does not allow using a future or current block number', async function () {
                const issuancePerBlock = (0, sdk_1.toGRT)('120');
                let issuanceUpdatedAtBlock = (await sdk_1.helpers.latestBlock()) + 2;
                const tx1 = l1GraphTokenGateway
                    .connect(governor)
                    .updateL2MintAllowance(issuancePerBlock, issuanceUpdatedAtBlock);
                await (0, chai_1.expect)(tx1).revertedWith('BLOCK_MUST_BE_PAST');
                issuanceUpdatedAtBlock = (await sdk_1.helpers.latestBlock()) + 1; // This will be block.number in our next tx
                const tx2 = l1GraphTokenGateway
                    .connect(governor)
                    .updateL2MintAllowance(issuancePerBlock, issuanceUpdatedAtBlock);
                await (0, chai_1.expect)(tx2).revertedWith('BLOCK_MUST_BE_PAST');
                issuanceUpdatedAtBlock = await sdk_1.helpers.latestBlock(); // This will be block.number-1 in our next tx
                const tx3 = l1GraphTokenGateway
                    .connect(governor)
                    .updateL2MintAllowance(issuancePerBlock, issuanceUpdatedAtBlock);
                await (0, chai_1.expect)(tx3)
                    .emit(l1GraphTokenGateway, 'L2MintAllowanceUpdated')
                    .withArgs((0, sdk_1.toGRT)('0'), issuancePerBlock, issuanceUpdatedAtBlock);
            });
            it('does not allow using a block number lower than or equal to the previous one', async function () {
                const issuancePerBlock = (0, sdk_1.toGRT)('120');
                const issuanceUpdatedAtBlock = await sdk_1.helpers.latestBlock();
                const tx1 = l1GraphTokenGateway
                    .connect(governor)
                    .updateL2MintAllowance(issuancePerBlock, issuanceUpdatedAtBlock);
                await (0, chai_1.expect)(tx1)
                    .emit(l1GraphTokenGateway, 'L2MintAllowanceUpdated')
                    .withArgs((0, sdk_1.toGRT)('0'), issuancePerBlock, issuanceUpdatedAtBlock);
                const tx2 = l1GraphTokenGateway
                    .connect(governor)
                    .updateL2MintAllowance(issuancePerBlock, issuanceUpdatedAtBlock);
                await (0, chai_1.expect)(tx2).revertedWith('BLOCK_MUST_BE_INCREMENTING');
                const tx3 = l1GraphTokenGateway
                    .connect(governor)
                    .updateL2MintAllowance(issuancePerBlock, issuanceUpdatedAtBlock - 1);
                await (0, chai_1.expect)(tx3).revertedWith('BLOCK_MUST_BE_INCREMENTING');
                const tx4 = l1GraphTokenGateway
                    .connect(governor)
                    .updateL2MintAllowance(issuancePerBlock, issuanceUpdatedAtBlock + 1);
                await (0, chai_1.expect)(tx4)
                    .emit(l1GraphTokenGateway, 'L2MintAllowanceUpdated')
                    .withArgs(issuancePerBlock, issuancePerBlock, issuanceUpdatedAtBlock + 1);
            });
            it('updates the snapshot and issuance to follow a new linear function, accumulating up to the specified block', async function () {
                const issuancePerBlock = (0, sdk_1.toGRT)('120');
                const issuanceUpdatedAtBlock = (await sdk_1.helpers.latestBlock()) - 2;
                const tx1 = l1GraphTokenGateway
                    .connect(governor)
                    .updateL2MintAllowance(issuancePerBlock, issuanceUpdatedAtBlock);
                await (0, chai_1.expect)(tx1)
                    .emit(l1GraphTokenGateway, 'L2MintAllowanceUpdated')
                    .withArgs((0, sdk_1.toGRT)('0'), issuancePerBlock, issuanceUpdatedAtBlock);
                // Now the mint allowance should be issuancePerBlock * 3
                (0, chai_1.expect)(await l1GraphTokenGateway.accumulatedL2MintAllowanceAtBlock(await sdk_1.helpers.latestBlock())).to.eq(issuancePerBlock.mul(3));
                (0, chai_1.expect)(await l1GraphTokenGateway.accumulatedL2MintAllowanceSnapshot()).to.eq(0);
                (0, chai_1.expect)(await l1GraphTokenGateway.l2MintAllowancePerBlock()).to.eq(issuancePerBlock);
                (0, chai_1.expect)(await l1GraphTokenGateway.lastL2MintAllowanceUpdateBlock()).to.eq(issuanceUpdatedAtBlock);
                await sdk_1.helpers.mine(10);
                const newIssuancePerBlock = (0, sdk_1.toGRT)('200');
                const newIssuanceUpdatedAtBlock = (await sdk_1.helpers.latestBlock()) - 1;
                const expectedAccumulatedSnapshot = issuancePerBlock.mul(newIssuanceUpdatedAtBlock - issuanceUpdatedAtBlock);
                const tx2 = l1GraphTokenGateway
                    .connect(governor)
                    .updateL2MintAllowance(newIssuancePerBlock, newIssuanceUpdatedAtBlock);
                await (0, chai_1.expect)(tx2)
                    .emit(l1GraphTokenGateway, 'L2MintAllowanceUpdated')
                    .withArgs(expectedAccumulatedSnapshot, newIssuancePerBlock, newIssuanceUpdatedAtBlock);
                (0, chai_1.expect)(await l1GraphTokenGateway.accumulatedL2MintAllowanceAtBlock(await sdk_1.helpers.latestBlock())).to.eq(expectedAccumulatedSnapshot.add(newIssuancePerBlock.mul(2)));
                (0, chai_1.expect)(await l1GraphTokenGateway.accumulatedL2MintAllowanceSnapshot()).to.eq(expectedAccumulatedSnapshot);
                (0, chai_1.expect)(await l1GraphTokenGateway.l2MintAllowancePerBlock()).to.eq(newIssuancePerBlock);
                (0, chai_1.expect)(await l1GraphTokenGateway.lastL2MintAllowanceUpdateBlock()).to.eq(newIssuanceUpdatedAtBlock);
            });
        });
        describe('setL2MintAllowanceParametersManual', function () {
            it('rejects calls that are not from the governor', async function () {
                const tx = l1GraphTokenGateway
                    .connect(pauseGuardian.address)
                    .setL2MintAllowanceParametersManual((0, sdk_1.toGRT)('0'), (0, sdk_1.toGRT)('1'), await sdk_1.helpers.latestBlock());
                await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
            });
            it('does not allow using a future or current block number', async function () {
                const issuancePerBlock = (0, sdk_1.toGRT)('120');
                let issuanceUpdatedAtBlock = (await sdk_1.helpers.latestBlock()) + 2;
                const tx1 = l1GraphTokenGateway
                    .connect(governor)
                    .setL2MintAllowanceParametersManual((0, sdk_1.toGRT)('0'), issuancePerBlock, issuanceUpdatedAtBlock);
                await (0, chai_1.expect)(tx1).revertedWith('BLOCK_MUST_BE_PAST');
                issuanceUpdatedAtBlock = (await sdk_1.helpers.latestBlock()) + 1; // This will be block.number in our next tx
                const tx2 = l1GraphTokenGateway
                    .connect(governor)
                    .setL2MintAllowanceParametersManual((0, sdk_1.toGRT)('0'), issuancePerBlock, issuanceUpdatedAtBlock);
                await (0, chai_1.expect)(tx2).revertedWith('BLOCK_MUST_BE_PAST');
                issuanceUpdatedAtBlock = await sdk_1.helpers.latestBlock(); // This will be block.number-1 in our next tx
                const tx3 = l1GraphTokenGateway
                    .connect(governor)
                    .setL2MintAllowanceParametersManual((0, sdk_1.toGRT)('0'), issuancePerBlock, issuanceUpdatedAtBlock);
                await (0, chai_1.expect)(tx3)
                    .emit(l1GraphTokenGateway, 'L2MintAllowanceUpdated')
                    .withArgs((0, sdk_1.toGRT)('0'), issuancePerBlock, issuanceUpdatedAtBlock);
            });
            it('updates the snapshot and issuance to follow a new linear function, manually setting the snapshot value', async function () {
                const issuancePerBlock = (0, sdk_1.toGRT)('120');
                const issuanceUpdatedAtBlock = (await sdk_1.helpers.latestBlock()) - 2;
                const snapshotValue = (0, sdk_1.toGRT)('10');
                const tx1 = l1GraphTokenGateway
                    .connect(governor)
                    .setL2MintAllowanceParametersManual(snapshotValue, issuancePerBlock, issuanceUpdatedAtBlock);
                await (0, chai_1.expect)(tx1)
                    .emit(l1GraphTokenGateway, 'L2MintAllowanceUpdated')
                    .withArgs(snapshotValue, issuancePerBlock, issuanceUpdatedAtBlock);
                // Now the mint allowance should be 10 + issuancePerBlock * 3
                (0, chai_1.expect)(await l1GraphTokenGateway.accumulatedL2MintAllowanceAtBlock(await sdk_1.helpers.latestBlock())).to.eq(snapshotValue.add(issuancePerBlock.mul(3)));
                (0, chai_1.expect)(await l1GraphTokenGateway.accumulatedL2MintAllowanceSnapshot()).to.eq(snapshotValue);
                (0, chai_1.expect)(await l1GraphTokenGateway.l2MintAllowancePerBlock()).to.eq(issuancePerBlock);
                (0, chai_1.expect)(await l1GraphTokenGateway.lastL2MintAllowanceUpdateBlock()).to.eq(issuanceUpdatedAtBlock);
                await sdk_1.helpers.mine(10);
                const newIssuancePerBlock = (0, sdk_1.toGRT)('200');
                const newIssuanceUpdatedAtBlock = (await sdk_1.helpers.latestBlock()) - 1;
                const newSnapshotValue = (0, sdk_1.toGRT)('10');
                const tx2 = l1GraphTokenGateway
                    .connect(governor)
                    .setL2MintAllowanceParametersManual(newSnapshotValue, newIssuancePerBlock, newIssuanceUpdatedAtBlock);
                await (0, chai_1.expect)(tx2)
                    .emit(l1GraphTokenGateway, 'L2MintAllowanceUpdated')
                    .withArgs(newSnapshotValue, newIssuancePerBlock, newIssuanceUpdatedAtBlock);
                (0, chai_1.expect)(await l1GraphTokenGateway.accumulatedL2MintAllowanceAtBlock(await sdk_1.helpers.latestBlock())).to.eq(newSnapshotValue.add(newIssuancePerBlock.mul(2)));
                (0, chai_1.expect)(await l1GraphTokenGateway.accumulatedL2MintAllowanceSnapshot()).to.eq(newSnapshotValue);
                (0, chai_1.expect)(await l1GraphTokenGateway.l2MintAllowancePerBlock()).to.eq(newIssuancePerBlock);
                (0, chai_1.expect)(await l1GraphTokenGateway.lastL2MintAllowanceUpdateBlock()).to.eq(newIssuanceUpdatedAtBlock);
            });
        });
        describe('calculateL2TokenAddress', function () {
            it('returns the L2 token address', async function () {
                (0, chai_1.expect)(await l1GraphTokenGateway.calculateL2TokenAddress(grt.address)).eq(l2GRTMock.address);
            });
            it('returns the zero address if the input is any other address', async function () {
                (0, chai_1.expect)(await l1GraphTokenGateway.calculateL2TokenAddress(tokenSender.address)).eq(AddressZero);
            });
        });
        describe('outboundTransfer', function () {
            it('reverts when called with the wrong token address', async function () {
                const tx = l1GraphTokenGateway
                    .connect(tokenSender)
                    .outboundTransfer(tokenSender.address, l2Receiver.address, (0, sdk_1.toGRT)('10'), maxGas, gasPriceBid, defaultData, {
                    value: defaultEthValue,
                });
                await (0, chai_1.expect)(tx).revertedWith('TOKEN_NOT_GRT');
            });
            it('puts tokens in escrow and creates a retryable ticket', async function () {
                await grt.connect(tokenSender).approve(l1GraphTokenGateway.address, (0, sdk_1.toGRT)('10'));
                await testValidOutboundTransfer(tokenSender, defaultData, emptyCallHookData);
            });
            it('decodes the sender address from messages sent by the router', async function () {
                await grt.connect(tokenSender).approve(l1GraphTokenGateway.address, (0, sdk_1.toGRT)('10'));
                const routerEncodedData = ethers_1.utils.defaultAbiCoder.encode(['address', 'bytes'], [tokenSender.address, defaultData]);
                await testValidOutboundTransfer(routerMock, routerEncodedData, emptyCallHookData);
            });
            it('reverts when called with no submission cost', async function () {
                await grt.connect(tokenSender).approve(l1GraphTokenGateway.address, (0, sdk_1.toGRT)('10'));
                const tx = l1GraphTokenGateway
                    .connect(tokenSender)
                    .outboundTransfer(grt.address, l2Receiver.address, (0, sdk_1.toGRT)('10'), maxGas, gasPriceBid, defaultDataNoSubmissionCost, {
                    value: defaultEthValue,
                });
                await (0, chai_1.expect)(tx).revertedWith('NO_SUBMISSION_COST');
            });
            it('reverts when called with nonempty calldata, if the sender is not allowlisted', async function () {
                await grt.connect(tokenSender).approve(l1GraphTokenGateway.address, (0, sdk_1.toGRT)('10'));
                const tx = l1GraphTokenGateway
                    .connect(tokenSender)
                    .outboundTransfer(grt.address, l2Receiver.address, (0, sdk_1.toGRT)('10'), maxGas, gasPriceBid, defaultDataWithNotEmptyCallHookData, {
                    value: defaultEthValue,
                });
                await (0, chai_1.expect)(tx).revertedWith('CALL_HOOK_DATA_NOT_ALLOWED');
            });
            it('allows sending nonempty calldata, if the sender is allowlisted', async function () {
                // Make the sender a contract so that it can be allowed to send callhooks
                await sdk_1.helpers.setCode(tokenSender.address, '0x1234');
                await l1GraphTokenGateway.connect(governor).addToCallhookAllowlist(tokenSender.address);
                await grt.connect(tokenSender).approve(l1GraphTokenGateway.address, (0, sdk_1.toGRT)('10'));
                await testValidOutboundTransfer(tokenSender, defaultDataWithNotEmptyCallHookData, notEmptyCallHookData);
            });
            it('reverts when the sender does not have enough GRT', async function () {
                await grt.connect(tokenSender).approve(l1GraphTokenGateway.address, (0, sdk_1.toGRT)('1001'));
                const tx = l1GraphTokenGateway
                    .connect(tokenSender)
                    .outboundTransfer(grt.address, l2Receiver.address, (0, sdk_1.toGRT)('1001'), maxGas, gasPriceBid, defaultData, {
                    value: defaultEthValue,
                });
                await (0, chai_1.expect)(tx).revertedWith('ERC20: transfer amount exceeds balance');
            });
        });
        describe('finalizeInboundTransfer', function () {
            it('reverts when called by an account that is not the bridge', async function () {
                const tx = l1GraphTokenGateway
                    .connect(tokenSender)
                    .finalizeInboundTransfer(grt.address, l2Receiver.address, tokenSender.address, (0, sdk_1.toGRT)('10'), defaultData);
                await (0, chai_1.expect)(tx).revertedWith('NOT_FROM_BRIDGE');
            });
            it('reverts when called by the bridge, but the tx was not started by the L2 gateway', async function () {
                const encodedCalldata = l1GraphTokenGateway.interface.encodeFunctionData('finalizeInboundTransfer', [
                    grt.address,
                    l2Receiver.address,
                    tokenSender.address,
                    (0, sdk_1.toGRT)('10'),
                    ethers_1.utils.defaultAbiCoder.encode(['uint256', 'bytes'], [0, []]),
                ]);
                // The real outbox would require a proof, which would
                // validate that the tx was initiated by the L2 gateway but our mock
                // just executes unconditionally
                const tx = outboxMock.connect(tokenSender).executeTransaction((0, sdk_1.toBN)('0'), [], (0, sdk_1.toBN)('0'), l2Receiver.address, // Note this is not l2GRTGatewayMock
                l1GraphTokenGateway.address, (0, sdk_1.toBN)('1337'), await sdk_1.helpers.latestBlock(), (0, sdk_1.toBN)('133701337'), (0, sdk_1.toBN)('0'), encodedCalldata);
                await (0, chai_1.expect)(tx).revertedWith('ONLY_COUNTERPART_GATEWAY');
            });
            it('reverts if the gateway does not have tokens or allowance', async function () {
                // This scenario should never really happen, but we still
                // test that the gateway reverts in this case
                const encodedCalldata = l1GraphTokenGateway.interface.encodeFunctionData('finalizeInboundTransfer', [
                    grt.address,
                    l2Receiver.address,
                    tokenSender.address,
                    (0, sdk_1.toGRT)('10'),
                    ethers_1.utils.defaultAbiCoder.encode(['uint256', 'bytes'], [0, []]),
                ]);
                // The real outbox would require a proof, which would
                // validate that the tx was initiated by the L2 gateway but our mock
                // just executes unconditionally
                const tx = outboxMock
                    .connect(tokenSender)
                    .executeTransaction((0, sdk_1.toBN)('0'), [], (0, sdk_1.toBN)('0'), l2GRTGatewayMock.address, l1GraphTokenGateway.address, (0, sdk_1.toBN)('1337'), await sdk_1.helpers.latestBlock(), (0, sdk_1.toBN)('133701337'), (0, sdk_1.toBN)('0'), encodedCalldata);
                await (0, chai_1.expect)(tx).revertedWith('INVALID_L2_MINT_AMOUNT');
            });
            it('reverts if the gateway is revoked from escrow', async function () {
                await grt.connect(tokenSender).approve(l1GraphTokenGateway.address, (0, sdk_1.toGRT)('10'));
                await testValidOutboundTransfer(tokenSender, defaultData, emptyCallHookData);
                // At this point, the gateway holds 10 GRT in escrow
                // But we revoke the gateway's permission to move the funds:
                await bridgeEscrow.connect(governor).revokeAll(l1GraphTokenGateway.address);
                const encodedCalldata = l1GraphTokenGateway.interface.encodeFunctionData('finalizeInboundTransfer', [
                    grt.address,
                    l2Receiver.address,
                    tokenSender.address,
                    (0, sdk_1.toGRT)('8'),
                    ethers_1.utils.defaultAbiCoder.encode(['uint256', 'bytes'], [0, []]),
                ]);
                // The real outbox would require a proof, which would
                // validate that the tx was initiated by the L2 gateway but our mock
                // just executes unconditionally
                const tx = outboxMock
                    .connect(tokenSender)
                    .executeTransaction((0, sdk_1.toBN)('0'), [], (0, sdk_1.toBN)('0'), l2GRTGatewayMock.address, l1GraphTokenGateway.address, (0, sdk_1.toBN)('1337'), await sdk_1.helpers.latestBlock(), (0, sdk_1.toBN)('133701337'), (0, sdk_1.toBN)('0'), encodedCalldata);
                await (0, chai_1.expect)(tx).revertedWith('ERC20: transfer amount exceeds allowance');
            });
            it('sends tokens out of escrow', async function () {
                await grt.connect(tokenSender).approve(l1GraphTokenGateway.address, (0, sdk_1.toGRT)('10'));
                await testValidOutboundTransfer(tokenSender, defaultData, emptyCallHookData);
                // At this point, the gateway holds 10 GRT in escrow
                const encodedCalldata = l1GraphTokenGateway.interface.encodeFunctionData('finalizeInboundTransfer', [
                    grt.address,
                    l2Receiver.address,
                    tokenSender.address,
                    (0, sdk_1.toGRT)('8'),
                    ethers_1.utils.defaultAbiCoder.encode(['uint256', 'bytes'], [0, []]),
                ]);
                // The real outbox would require a proof, which would
                // validate that the tx was initiated by the L2 gateway but our mock
                // just executes unconditionally
                const tx = outboxMock
                    .connect(tokenSender)
                    .executeTransaction((0, sdk_1.toBN)('0'), [], (0, sdk_1.toBN)('0'), l2GRTGatewayMock.address, l1GraphTokenGateway.address, (0, sdk_1.toBN)('1337'), await sdk_1.helpers.latestBlock(), (0, sdk_1.toBN)('133701337'), (0, sdk_1.toBN)('0'), encodedCalldata);
                await (0, chai_1.expect)(tx)
                    .emit(l1GraphTokenGateway, 'WithdrawalFinalized')
                    .withArgs(grt.address, l2Receiver.address, tokenSender.address, (0, sdk_1.toBN)('0'), (0, sdk_1.toGRT)('8'));
                const escrowBalance = await grt.balanceOf(bridgeEscrow.address);
                const senderBalance = await grt.balanceOf(tokenSender.address);
                (0, chai_1.expect)(escrowBalance).eq((0, sdk_1.toGRT)('2'));
                (0, chai_1.expect)(senderBalance).eq((0, sdk_1.toGRT)('998'));
            });
            it('mints tokens up to the L2 mint allowance', async function () {
                await grt.connect(tokenSender).approve(l1GraphTokenGateway.address, (0, sdk_1.toGRT)('10'));
                await testValidOutboundTransfer(tokenSender, defaultData, emptyCallHookData);
                // Start accruing L2 mint allowance at 2 GRT per block
                await l1GraphTokenGateway
                    .connect(governor)
                    .updateL2MintAllowance((0, sdk_1.toGRT)('2'), await sdk_1.helpers.latestBlock());
                await sdk_1.helpers.mine(2);
                // Now it's been three blocks since the lastL2MintAllowanceUpdateBlock, so
                // there should be 8 GRT allowed to be minted from L2 in the next block.
                // At this point, the gateway holds 10 GRT in escrow
                const encodedCalldata = l1GraphTokenGateway.interface.encodeFunctionData('finalizeInboundTransfer', [
                    grt.address,
                    l2Receiver.address,
                    tokenSender.address,
                    (0, sdk_1.toGRT)('18'),
                    ethers_1.utils.defaultAbiCoder.encode(['uint256', 'bytes'], [0, []]),
                ]);
                // The real outbox would require a proof, which would
                // validate that the tx was initiated by the L2 gateway but our mock
                // just executes unconditionally
                const tx = outboxMock
                    .connect(tokenSender)
                    .executeTransaction((0, sdk_1.toBN)('0'), [], (0, sdk_1.toBN)('0'), l2GRTGatewayMock.address, l1GraphTokenGateway.address, (0, sdk_1.toBN)('1337'), await sdk_1.helpers.latestBlock(), (0, sdk_1.toBN)('133701337'), (0, sdk_1.toBN)('0'), encodedCalldata);
                await (0, chai_1.expect)(tx)
                    .emit(l1GraphTokenGateway, 'WithdrawalFinalized')
                    .withArgs(grt.address, l2Receiver.address, tokenSender.address, (0, sdk_1.toBN)('0'), (0, sdk_1.toGRT)('18'))
                    .emit(l1GraphTokenGateway, 'TokensMintedFromL2')
                    .withArgs((0, sdk_1.toGRT)('8'));
                (0, chai_1.expect)(await l1GraphTokenGateway.totalMintedFromL2()).to.eq((0, sdk_1.toGRT)('8'));
                (0, chai_1.expect)(await l1GraphTokenGateway.accumulatedL2MintAllowanceAtBlock(await sdk_1.helpers.latestBlock())).to.eq((0, sdk_1.toGRT)('8'));
                const escrowBalance = await grt.balanceOf(bridgeEscrow.address);
                const senderBalance = await grt.balanceOf(tokenSender.address);
                (0, chai_1.expect)(escrowBalance).eq((0, sdk_1.toGRT)('0'));
                (0, chai_1.expect)(senderBalance).eq((0, sdk_1.toGRT)('1008'));
            });
            it('reverts if the amount to mint is over the allowance', async function () {
                await grt.connect(tokenSender).approve(l1GraphTokenGateway.address, (0, sdk_1.toGRT)('10'));
                await testValidOutboundTransfer(tokenSender, defaultData, emptyCallHookData);
                // Start accruing L2 mint allowance at 2 GRT per block
                await l1GraphTokenGateway
                    .connect(governor)
                    .updateL2MintAllowance((0, sdk_1.toGRT)('2'), await sdk_1.helpers.latestBlock());
                await sdk_1.helpers.mine(2);
                // Now it's been three blocks since the lastL2MintAllowanceUpdateBlock, so
                // there should be 8 GRT allowed to be minted from L2 in the next block.
                // At this point, the gateway holds 10 GRT in escrow
                const encodedCalldata = l1GraphTokenGateway.interface.encodeFunctionData('finalizeInboundTransfer', [
                    grt.address,
                    l2Receiver.address,
                    tokenSender.address,
                    (0, sdk_1.toGRT)('18.001'),
                    ethers_1.utils.defaultAbiCoder.encode(['uint256', 'bytes'], [0, []]),
                ]);
                // The real outbox would require a proof, which would
                // validate that the tx was initiated by the L2 gateway but our mock
                // just executes unconditionally
                const tx = outboxMock
                    .connect(tokenSender)
                    .executeTransaction((0, sdk_1.toBN)('0'), [], (0, sdk_1.toBN)('0'), l2GRTGatewayMock.address, l1GraphTokenGateway.address, (0, sdk_1.toBN)('1337'), await sdk_1.helpers.latestBlock(), (0, sdk_1.toBN)('133701337'), (0, sdk_1.toBN)('0'), encodedCalldata);
                await (0, chai_1.expect)(tx).revertedWith('INVALID_L2_MINT_AMOUNT');
            });
        });
    });
});
