"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const hardhat_1 = __importDefault(require("hardhat"));
const chai_1 = require("chai");
const ethers_1 = require("ethers");
const fixtures_1 = require("../lib/fixtures");
const { AddressZero } = ethers_1.constants;
describe('Managed', () => {
    const graph = hardhat_1.default.graph();
    let me;
    let governor;
    let mockController;
    let newMockEpochManager;
    let fixture;
    let contracts;
    let epochManager;
    let controller;
    before(async function () {
        [me, mockController, newMockEpochManager] = await graph.getTestAccounts();
        ({ governor } = await graph.getNamedAccounts());
        // We just run the fixures to set up a contract with  Managed, as this
        // is cleaner and easier for us to test
        fixture = new fixtures_1.NetworkFixture(graph.provider);
        contracts = await fixture.load(governor);
        epochManager = contracts.EpochManager;
        controller = contracts.Controller;
    });
    beforeEach(async function () {
        await fixture.setUp();
    });
    afterEach(async function () {
        await fixture.tearDown();
    });
    describe('setContractProxy()', function () {
        it('should set contract proxy and test get contract proxy', async function () {
            // Set right in the constructor
            (0, chai_1.expect)(await epochManager.controller()).eq(controller.address);
            // Test the controller
            const id = ethers_1.utils.id('EpochManager');
            const tx = controller.connect(governor).setContractProxy(id, newMockEpochManager.address);
            await (0, chai_1.expect)(tx)
                .emit(controller, 'SetContractProxy')
                .withArgs(id, newMockEpochManager.address);
            (0, chai_1.expect)(await controller.getContractProxy(id)).eq(newMockEpochManager.address);
        });
        it('reject set contract proxy to address zero', async function () {
            const id = ethers_1.utils.id('EpochManager');
            const tx = controller.connect(governor).setContractProxy(id, AddressZero);
            await (0, chai_1.expect)(tx).revertedWith('Contract address must be set');
        });
        it('reject set contract proxy if not governor', async function () {
            const id = ethers_1.utils.id('EpochManager');
            const tx = controller.connect(me).setContractProxy(id, newMockEpochManager.address);
            await (0, chai_1.expect)(tx).revertedWith('Only Governor can call');
        });
    });
    describe('unsetContractProxy()', function () {
        it('should unset contract proxy', async function () {
            // Set contract
            const id = ethers_1.utils.id('EpochManager');
            await controller.connect(governor).setContractProxy(id, newMockEpochManager.address);
            (0, chai_1.expect)(await controller.getContractProxy(id)).eq(newMockEpochManager.address);
            // Unset contract
            const tx = controller.connect(governor).unsetContractProxy(id);
            await (0, chai_1.expect)(tx).emit(controller, 'SetContractProxy').withArgs(id, AddressZero);
            (0, chai_1.expect)(await controller.getContractProxy(id)).eq(AddressZero);
        });
        it('reject to call if not authorized caller', async function () {
            // Unset contract
            const id = ethers_1.utils.id('EpochManager');
            const tx = controller.connect(me).unsetContractProxy(id);
            await (0, chai_1.expect)(tx).revertedWith('Only Governor can call');
        });
    });
    describe('updateController()', function () {
        it('should update controller on a manager', async function () {
            const tx = controller
                .connect(governor)
                .updateController(ethers_1.utils.id('EpochManager'), mockController.address);
            await (0, chai_1.expect)(tx).emit(epochManager, 'SetController').withArgs(mockController.address);
            (0, chai_1.expect)(await epochManager.controller()).eq(mockController.address);
        });
        it('should fail updating controller when not called by governor', async function () {
            const tx = controller
                .connect(me)
                .updateController(ethers_1.utils.id('EpochManager'), mockController.address);
            await (0, chai_1.expect)(tx).revertedWith('Only Governor can call');
        });
        it('reject update controller to address zero', async function () {
            const tx = controller
                .connect(governor)
                .updateController(ethers_1.utils.id('EpochManager'), AddressZero);
            await (0, chai_1.expect)(tx).revertedWith('Controller must be set');
        });
    });
    describe('setController()', function () {
        it('should fail setting controller when not called from Controller', async function () {
            const tx = epochManager.connect(me).setController(mockController.address);
            await (0, chai_1.expect)(tx).revertedWith('Caller must be Controller');
        });
    });
    describe('setPauseGuardian()', function () {
        it('should set the pause guardian', async function () {
            const tx = controller.connect(governor).setPauseGuardian(me.address);
            const currentPauseGuardian = await controller.pauseGuardian();
            await (0, chai_1.expect)(tx)
                .emit(controller, 'NewPauseGuardian')
                .withArgs(currentPauseGuardian, me.address);
            (0, chai_1.expect)(await controller.pauseGuardian()).eq(me.address);
        });
        it('reject to call if not authorized caller', async function () {
            const tx = controller.connect(me).setPauseGuardian(me.address);
            await (0, chai_1.expect)(tx).revertedWith('Only Governor can call');
        });
        it('reject set pause guardian to address zero', async function () {
            const tx = controller.connect(governor).setPauseGuardian(AddressZero);
            await (0, chai_1.expect)(tx).revertedWith('PauseGuardian must be set');
        });
    });
});
