"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const hardhat_1 = __importDefault(require("hardhat"));
const chai_1 = require("chai");
const fixtures_1 = require("../lib/fixtures");
const sdk_1 = require("@graphprotocol/sdk");
describe('Pausing', () => {
    const graph = hardhat_1.default.graph();
    let me;
    let governor;
    let guardian;
    let fixture;
    let contracts;
    let staking;
    let controller;
    const setPartialPause = async (account, setValue) => {
        const tx = controller.connect(account).setPartialPaused(setValue);
        await (0, chai_1.expect)(tx).emit(controller, 'PartialPauseChanged').withArgs(setValue);
        (0, chai_1.expect)(await controller.partialPaused()).eq(setValue);
    };
    const setPause = async (account, setValue) => {
        const tx = controller.connect(account).setPaused(setValue);
        await (0, chai_1.expect)(tx).emit(controller, 'PauseChanged').withArgs(setValue);
        (0, chai_1.expect)(await controller.paused()).eq(setValue);
    };
    before(async function () {
        [me] = await graph.getTestAccounts();
        ({ governor, pauseGuardian: guardian } = await graph.getNamedAccounts());
        fixture = new fixtures_1.NetworkFixture(graph.provider);
        contracts = await fixture.load(governor);
        staking = contracts.Staking;
        controller = contracts.Controller;
    });
    beforeEach(async function () {
        await fixture.setUp();
    });
    afterEach(async function () {
        await fixture.tearDown();
    });
    it('should set pause guardian', async function () {
        const currentGuardian = await controller.pauseGuardian();
        (0, chai_1.expect)(await controller.pauseGuardian()).eq(currentGuardian);
        const tx = controller.connect(governor).setPauseGuardian(guardian.address);
        await (0, chai_1.expect)(tx)
            .emit(controller, 'NewPauseGuardian')
            .withArgs(currentGuardian, guardian.address);
        (0, chai_1.expect)(await controller.pauseGuardian()).eq(guardian.address);
    });
    it('should fail pause guardian when not governor', async function () {
        const tx = controller.connect(me).setPauseGuardian(guardian.address);
        await (0, chai_1.expect)(tx).revertedWith('Only Governor can call');
    });
    it('should set partialPaused and unset from governor and guardian', async function () {
        (0, chai_1.expect)(await controller.partialPaused()).eq(false);
        // Governor set
        await setPartialPause(governor, true);
        // Governor unset
        await setPartialPause(governor, false);
        await controller.connect(governor).setPauseGuardian(guardian.address);
        // Guardian set
        await setPartialPause(guardian, true);
        // Guardian unset
        await setPartialPause(guardian, false);
    });
    it('should fail partial pause if not guardian or governor', async function () {
        const tx = controller.connect(me).setPartialPaused(true);
        await (0, chai_1.expect)(tx).revertedWith('Only Governor or Guardian can call');
    });
    it('should check that a function fails when partialPause is set', async function () {
        await setPartialPause(governor, true);
        const tokensToStake = (0, sdk_1.toGRT)('100');
        const tx = staking.connect(me).stake(tokensToStake);
        await (0, chai_1.expect)(tx).revertedWith('Partial-paused');
    });
    it('should set pause and unset from governor and guardian', async function () {
        (0, chai_1.expect)(await controller.paused()).eq(false);
        // Governor set
        await setPause(governor, true);
        // Governor unset
        await setPause(governor, false);
        await controller.connect(governor).setPauseGuardian(guardian.address);
        // Guardian set
        await setPause(guardian, true);
        // Guardian unset
        await setPause(guardian, false);
    });
    it('should fail pause if not guardian or governor', async function () {
        const tx = controller.connect(me).setPaused(true);
        await (0, chai_1.expect)(tx).revertedWith('Only Governor or Guardian can call');
    });
    it('should check that a function fails when pause is set', async function () {
        await setPause(governor, true);
        const tokensToStake = (0, sdk_1.toGRT)('100');
        const tx1 = staking.connect(me).stake(tokensToStake);
        await (0, chai_1.expect)(tx1).revertedWith('Paused');
        const tx2 = staking.connect(me).withdraw();
        await (0, chai_1.expect)(tx2).revertedWith('Paused');
    });
});
