"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
/* eslint-disable no-secrets/no-secrets */
const hardhat_1 = __importDefault(require("hardhat"));
const chai_1 = require("chai");
const ethers_1 = require("ethers");
const utils_1 = require("ethers/lib/utils");
const fixtures_1 = require("../lib/fixtures");
const gnsUtils_1 = require("../lib/gnsUtils");
const sdk_1 = require("@graphprotocol/sdk");
const { HashZero } = ethers_1.ethers.constants;
describe('L2GNS', () => {
    const graph = hardhat_1.default.graph();
    let me;
    let attacker;
    let other;
    let governor;
    let fixture;
    let fixtureContracts;
    let l1MockContracts;
    let l1GRTMock;
    let l1GNSMock;
    let l1GRTGatewayMock;
    let l2GraphTokenGateway;
    let gns;
    let curation;
    let grt;
    let staking;
    let newSubgraph0;
    let newSubgraph1;
    const tokens1000 = (0, sdk_1.toGRT)('1000');
    const tokens10000 = (0, sdk_1.toGRT)('10000');
    const tokens100000 = (0, sdk_1.toGRT)('100000');
    const curationTaxPercentage = 50000;
    const gatewayFinalizeTransfer = async function (from, to, amount, callhookData) {
        const l1GRTGatewayMockL2Alias = await sdk_1.helpers.getL2SignerFromL1(l1GRTGatewayMock.address);
        // Eth for gas:
        await sdk_1.helpers.setBalance(await l1GRTGatewayMockL2Alias.getAddress(), (0, utils_1.parseEther)('1'));
        const tx = l2GraphTokenGateway
            .connect(l1GRTGatewayMockL2Alias)
            .finalizeInboundTransfer(l1GRTMock.address, from, to, amount, callhookData);
        return tx;
    };
    const defaultL1SubgraphParams = async function () {
        return {
            l1SubgraphId: await (0, sdk_1.buildSubgraphId)(me.address, (0, sdk_1.toBN)('1'), graph.chainId),
            curatedTokens: (0, sdk_1.toGRT)('1337'),
            subgraphMetadata: (0, sdk_1.randomHexBytes)(),
            versionMetadata: (0, sdk_1.randomHexBytes)(),
        };
    };
    const transferMockSubgraphFromL1 = async function (l1SubgraphId, curatedTokens, subgraphMetadata, versionMetadata) {
        const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'uint256', 'address'], [(0, sdk_1.toBN)(0), l1SubgraphId, me.address]);
        await gatewayFinalizeTransfer(l1GNSMock.address, gns.address, curatedTokens, callhookData);
        const l2SubgraphId = await gns.getAliasedL2SubgraphID(l1SubgraphId);
        await gns
            .connect(me)
            .finishSubgraphTransferFromL1(l2SubgraphId, newSubgraph0.subgraphDeploymentID, subgraphMetadata, versionMetadata);
    };
    before(async function () {
        newSubgraph0 = (0, sdk_1.buildSubgraph)();
        [me, attacker, other] = await graph.getTestAccounts();
        ({ governor } = await graph.getNamedAccounts());
        fixture = new fixtures_1.NetworkFixture(graph.provider);
        // Deploy L2
        fixtureContracts = await fixture.load(governor, true);
        l2GraphTokenGateway = fixtureContracts.L2GraphTokenGateway;
        gns = fixtureContracts.L2GNS;
        staking = fixtureContracts.L2Staking;
        curation = fixtureContracts.L2Curation;
        grt = fixtureContracts.GraphToken;
        // Deploy L1 mock
        l1MockContracts = await fixture.loadMock(false);
        l1GRTMock = l1MockContracts.GraphToken;
        l1GNSMock = l1MockContracts.L1GNS;
        l1GRTGatewayMock = l1MockContracts.L1GraphTokenGateway;
        // Deploy L2 arbitrum bridge
        await fixture.loadL2ArbitrumBridge(governor);
        // Configure L2 bridge
        await fixture.configureL2Bridge(governor, fixtureContracts, l1MockContracts);
        await grt.connect(governor).mint(me.address, (0, sdk_1.toGRT)('10000'));
    });
    beforeEach(async function () {
        await fixture.setUp();
    });
    afterEach(async function () {
        await fixture.tearDown();
    });
    // Adapted from the L1 GNS tests but allowing curating to a pre-curated subgraph deployment
    describe('publishNewVersion', function () {
        let subgraph;
        beforeEach(async () => {
            newSubgraph0 = (0, sdk_1.buildSubgraph)();
            newSubgraph1 = (0, sdk_1.buildSubgraph)();
            // Give some funds to the signers and approve gns contract to use funds on signers behalf
            await grt.connect(governor).mint(me.address, tokens100000);
            await grt.connect(governor).mint(other.address, tokens100000);
            await grt.connect(me).approve(gns.address, tokens100000);
            await grt.connect(me).approve(curation.address, tokens100000);
            await grt.connect(other).approve(gns.address, tokens100000);
            await grt.connect(other).approve(curation.address, tokens100000);
            // Update curation tax to test the functionality of it in disableNameSignal()
            await curation.connect(governor).setCurationTaxPercentage(curationTaxPercentage);
            subgraph = await (0, gnsUtils_1.publishNewSubgraph)(me, newSubgraph0, gns, graph.chainId);
            await (0, gnsUtils_1.mintSignal)(me, subgraph.id, tokens10000, gns, curation);
        });
        it('should publish a new version on an existing subgraph', async function () {
            await (0, gnsUtils_1.publishNewVersion)(me, subgraph.id, newSubgraph1, gns, curation);
        });
        it('should publish a new version on an existing subgraph when owner tax is zero', async function () {
            await gns.connect(governor).setOwnerTaxPercentage(0);
            await (0, gnsUtils_1.publishNewVersion)(me, subgraph.id, newSubgraph1, gns, curation);
        });
        it('should publish a new version on an existing subgraph with no current signal', async function () {
            const emptySignalSubgraph = await (0, gnsUtils_1.publishNewSubgraph)(me, (0, sdk_1.buildSubgraph)(), gns, graph.chainId);
            await (0, gnsUtils_1.publishNewVersion)(me, emptySignalSubgraph.id, newSubgraph1, gns, curation);
        });
        it('should reject a new version with the same subgraph deployment ID', async function () {
            const tx = gns
                .connect(me)
                .publishNewVersion(subgraph.id, newSubgraph0.subgraphDeploymentID, newSubgraph0.versionMetadata);
            await (0, chai_1.expect)(tx).revertedWith('GNS: Cannot publish a new version with the same subgraph deployment ID');
        });
        it('should reject publishing a version to a subgraph that does not exist', async function () {
            const tx = gns
                .connect(me)
                .publishNewVersion((0, sdk_1.randomHexBytes)(32), newSubgraph1.subgraphDeploymentID, newSubgraph1.versionMetadata);
            await (0, chai_1.expect)(tx).revertedWith('ERC721: owner query for nonexistent token');
        });
        it('reject if not the owner', async function () {
            const tx = gns
                .connect(other)
                .publishNewVersion(subgraph.id, newSubgraph1.subgraphDeploymentID, newSubgraph1.versionMetadata);
            await (0, chai_1.expect)(tx).revertedWith('GNS: Must be authorized');
        });
        it('should NOT fail when upgrade tries to point to a pre-curated', async function () {
            // Curate directly to the deployment
            await curation.connect(me).mint(newSubgraph1.subgraphDeploymentID, tokens1000, 0);
            await (0, gnsUtils_1.publishNewVersion)(me, subgraph.id, newSubgraph1, gns, curation);
        });
        it('should upgrade version when there is no signal with no signal migration', async function () {
            await (0, gnsUtils_1.burnSignal)(me, subgraph.id, gns, curation);
            const tx = gns
                .connect(me)
                .publishNewVersion(subgraph.id, newSubgraph1.subgraphDeploymentID, newSubgraph1.versionMetadata);
            await (0, chai_1.expect)(tx)
                .emit(gns, 'SubgraphVersionUpdated')
                .withArgs(subgraph.id, newSubgraph1.subgraphDeploymentID, newSubgraph1.versionMetadata);
        });
        it('should fail when subgraph is deprecated', async function () {
            await (0, gnsUtils_1.deprecateSubgraph)(me, subgraph.id, gns, curation, grt);
            const tx = gns
                .connect(me)
                .publishNewVersion(subgraph.id, newSubgraph1.subgraphDeploymentID, newSubgraph1.versionMetadata);
            // NOTE: deprecate burns the Subgraph NFT, when someone wants to publish a new version it won't find it
            await (0, chai_1.expect)(tx).revertedWith('ERC721: owner query for nonexistent token');
        });
    });
    describe('receiving a subgraph from L1 (onTokenTransfer)', function () {
        it('cannot be called by someone other than the L2GraphTokenGateway', async function () {
            const { l1SubgraphId, curatedTokens } = await defaultL1SubgraphParams();
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'uint256', 'address'], [(0, sdk_1.toBN)(0), l1SubgraphId, me.address]);
            const tx = gns.connect(me).onTokenTransfer(l1GNSMock.address, curatedTokens, callhookData);
            await (0, chai_1.expect)(tx).revertedWith('ONLY_GATEWAY');
        });
        it('rejects calls if the L1 sender is not the L1GNS', async function () {
            const { l1SubgraphId, curatedTokens } = await defaultL1SubgraphParams();
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'uint256', 'address'], [(0, sdk_1.toBN)(0), l1SubgraphId, me.address]);
            const tx = gatewayFinalizeTransfer(me.address, gns.address, curatedTokens, callhookData);
            await (0, chai_1.expect)(tx).revertedWith('ONLY_L1_GNS_THROUGH_BRIDGE');
        });
        it('creates a subgraph in a disabled state', async function () {
            const { l1SubgraphId, curatedTokens } = await defaultL1SubgraphParams();
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'uint256', 'address'], [(0, sdk_1.toBN)(0), l1SubgraphId, me.address]);
            const tx = gatewayFinalizeTransfer(l1GNSMock.address, gns.address, curatedTokens, callhookData);
            const l2SubgraphId = await gns.getAliasedL2SubgraphID(l1SubgraphId);
            await (0, chai_1.expect)(tx)
                .emit(l2GraphTokenGateway, 'DepositFinalized')
                .withArgs(l1GRTMock.address, l1GNSMock.address, gns.address, curatedTokens);
            await (0, chai_1.expect)(tx)
                .emit(gns, 'SubgraphReceivedFromL1')
                .withArgs(l1SubgraphId, l2SubgraphId, me.address, curatedTokens);
            const transferData = await gns.subgraphL2TransferData(l2SubgraphId);
            const subgraphData = await gns.subgraphs(l2SubgraphId);
            (0, chai_1.expect)(transferData.tokens).eq(curatedTokens);
            (0, chai_1.expect)(transferData.l2Done).eq(false);
            (0, chai_1.expect)(transferData.subgraphReceivedOnL2BlockNumber).eq(await sdk_1.helpers.latestBlock());
            (0, chai_1.expect)(subgraphData.vSignal).eq(0);
            (0, chai_1.expect)(subgraphData.nSignal).eq(0);
            (0, chai_1.expect)(subgraphData.subgraphDeploymentID).eq(HashZero);
            (0, chai_1.expect)(subgraphData.__DEPRECATED_reserveRatio).eq(gnsUtils_1.DEFAULT_RESERVE_RATIO);
            (0, chai_1.expect)(subgraphData.disabled).eq(true);
            (0, chai_1.expect)(subgraphData.withdrawableGRT).eq(0); // Important so that it's not the same as a deprecated subgraph!
            (0, chai_1.expect)(await gns.ownerOf(l2SubgraphId)).eq(me.address);
        });
        it('does not conflict with a locally created subgraph', async function () {
            const l2Subgraph = await (0, gnsUtils_1.publishNewSubgraph)(me, newSubgraph0, gns, graph.chainId);
            const { l1SubgraphId, curatedTokens } = await defaultL1SubgraphParams();
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'uint256', 'address'], [(0, sdk_1.toBN)(0), l1SubgraphId, me.address]);
            const tx = gatewayFinalizeTransfer(l1GNSMock.address, gns.address, curatedTokens, callhookData);
            const l2SubgraphId = await gns.getAliasedL2SubgraphID(l1SubgraphId);
            await (0, chai_1.expect)(tx)
                .emit(l2GraphTokenGateway, 'DepositFinalized')
                .withArgs(l1GRTMock.address, l1GNSMock.address, gns.address, curatedTokens);
            await (0, chai_1.expect)(tx)
                .emit(gns, 'SubgraphReceivedFromL1')
                .withArgs(l1SubgraphId, l2SubgraphId, me.address, curatedTokens);
            const transferData = await gns.subgraphL2TransferData(l2SubgraphId);
            const subgraphData = await gns.subgraphs(l2SubgraphId);
            (0, chai_1.expect)(transferData.tokens).eq(curatedTokens);
            (0, chai_1.expect)(transferData.l2Done).eq(false);
            (0, chai_1.expect)(transferData.subgraphReceivedOnL2BlockNumber).eq(await sdk_1.helpers.latestBlock());
            (0, chai_1.expect)(subgraphData.vSignal).eq(0);
            (0, chai_1.expect)(subgraphData.nSignal).eq(0);
            (0, chai_1.expect)(subgraphData.subgraphDeploymentID).eq(HashZero);
            (0, chai_1.expect)(subgraphData.__DEPRECATED_reserveRatio).eq(gnsUtils_1.DEFAULT_RESERVE_RATIO);
            (0, chai_1.expect)(subgraphData.disabled).eq(true);
            (0, chai_1.expect)(subgraphData.withdrawableGRT).eq(0); // Important so that it's not the same as a deprecated subgraph!
            (0, chai_1.expect)(await gns.ownerOf(l2SubgraphId)).eq(me.address);
            (0, chai_1.expect)(l2Subgraph.id).not.eq(l2SubgraphId);
            const l2SubgraphData = await gns.subgraphs(l2Subgraph.id);
            (0, chai_1.expect)(l2SubgraphData.vSignal).eq(0);
            (0, chai_1.expect)(l2SubgraphData.nSignal).eq(0);
            (0, chai_1.expect)(l2SubgraphData.subgraphDeploymentID).eq(l2Subgraph.subgraphDeploymentID);
            (0, chai_1.expect)(l2SubgraphData.__DEPRECATED_reserveRatio).eq(gnsUtils_1.DEFAULT_RESERVE_RATIO);
            (0, chai_1.expect)(l2SubgraphData.disabled).eq(false);
            (0, chai_1.expect)(l2SubgraphData.withdrawableGRT).eq(0);
        });
    });
    describe('finishing a subgraph transfer from L1', function () {
        it('publishes the transferred subgraph and mints signal with no tax', async function () {
            const { l1SubgraphId, curatedTokens, subgraphMetadata, versionMetadata } = await defaultL1SubgraphParams();
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'uint256', 'address'], [(0, sdk_1.toBN)(0), l1SubgraphId, me.address]);
            await gatewayFinalizeTransfer(l1GNSMock.address, gns.address, curatedTokens, callhookData);
            // Calculate expected signal before minting
            const expectedSignal = await curation.tokensToSignalNoTax(newSubgraph0.subgraphDeploymentID, curatedTokens);
            const l2SubgraphId = await gns.getAliasedL2SubgraphID(l1SubgraphId);
            const tx = gns
                .connect(me)
                .finishSubgraphTransferFromL1(l2SubgraphId, newSubgraph0.subgraphDeploymentID, subgraphMetadata, versionMetadata);
            await (0, chai_1.expect)(tx)
                .emit(gns, 'SubgraphPublished')
                .withArgs(l2SubgraphId, newSubgraph0.subgraphDeploymentID, gnsUtils_1.DEFAULT_RESERVE_RATIO);
            await (0, chai_1.expect)(tx).emit(gns, 'SubgraphMetadataUpdated').withArgs(l2SubgraphId, subgraphMetadata);
            await (0, chai_1.expect)(tx)
                .emit(gns, 'SubgraphUpgraded')
                .withArgs(l2SubgraphId, expectedSignal, curatedTokens, newSubgraph0.subgraphDeploymentID);
            await (0, chai_1.expect)(tx)
                .emit(gns, 'SubgraphVersionUpdated')
                .withArgs(l2SubgraphId, newSubgraph0.subgraphDeploymentID, versionMetadata);
            await (0, chai_1.expect)(tx).emit(gns, 'SubgraphL2TransferFinalized').withArgs(l2SubgraphId);
            const subgraphAfter = await gns.subgraphs(l2SubgraphId);
            const transferDataAfter = await gns.subgraphL2TransferData(l2SubgraphId);
            (0, chai_1.expect)(subgraphAfter.vSignal).eq(expectedSignal);
            (0, chai_1.expect)(transferDataAfter.l2Done).eq(true);
            (0, chai_1.expect)(subgraphAfter.disabled).eq(false);
            (0, chai_1.expect)(subgraphAfter.subgraphDeploymentID).eq(newSubgraph0.subgraphDeploymentID);
            const expectedNSignal = await gns.vSignalToNSignal(l2SubgraphId, expectedSignal);
            (0, chai_1.expect)(await gns.getCuratorSignal(l2SubgraphId, me.address)).eq(expectedNSignal);
            await (0, chai_1.expect)(tx)
                .emit(gns, 'SignalMinted')
                .withArgs(l2SubgraphId, me.address, expectedNSignal, expectedSignal, curatedTokens);
        });
        it('protects the owner against a rounding attack', async function () {
            const { l1SubgraphId, curatedTokens, subgraphMetadata, versionMetadata } = await defaultL1SubgraphParams();
            const collectTokens = curatedTokens.mul(20);
            await staking.connect(governor).setCurationPercentage(100000);
            // Set up an indexer account with some stake
            await grt.connect(governor).mint(attacker.address, (0, sdk_1.toGRT)('1000000'));
            // Curate 1 wei GRT by minting 1 GRT and burning most of it
            await grt.connect(attacker).approve(curation.address, (0, sdk_1.toBN)(1));
            await curation.connect(attacker).mint(newSubgraph0.subgraphDeploymentID, (0, sdk_1.toBN)(1), 0);
            // Check this actually gave us 1 wei signal
            (0, chai_1.expect)(await curation.getCurationPoolTokens(newSubgraph0.subgraphDeploymentID)).eq(1);
            await grt.connect(attacker).approve(staking.address, (0, sdk_1.toGRT)('1000000'));
            await staking.connect(attacker).stake((0, sdk_1.toGRT)('100000'));
            const channelKey = (0, sdk_1.deriveChannelKey)();
            // Allocate to the same deployment ID
            await staking
                .connect(attacker)
                .allocateFrom(attacker.address, newSubgraph0.subgraphDeploymentID, (0, sdk_1.toGRT)('100000'), channelKey.address, (0, sdk_1.randomHexBytes)(32), await channelKey.generateProof(attacker.address));
            // Spoof some query fees, 10% of which will go to the Curation pool
            await staking.connect(attacker).collect(collectTokens, channelKey.address);
            // The curation pool now has 1 wei shares and a lot of tokens, so the rounding attack is prepared
            // But L2GNS will protect the owner by sending the tokens
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'uint256', 'address'], [(0, sdk_1.toBN)(0), l1SubgraphId, me.address]);
            await gatewayFinalizeTransfer(l1GNSMock.address, gns.address, curatedTokens, callhookData);
            const l2SubgraphId = await gns.getAliasedL2SubgraphID(l1SubgraphId);
            const tx = gns
                .connect(me)
                .finishSubgraphTransferFromL1(l2SubgraphId, newSubgraph0.subgraphDeploymentID, subgraphMetadata, versionMetadata);
            await (0, chai_1.expect)(tx)
                .emit(gns, 'SubgraphPublished')
                .withArgs(l2SubgraphId, newSubgraph0.subgraphDeploymentID, gnsUtils_1.DEFAULT_RESERVE_RATIO);
            await (0, chai_1.expect)(tx).emit(gns, 'SubgraphMetadataUpdated').withArgs(l2SubgraphId, subgraphMetadata);
            await (0, chai_1.expect)(tx).emit(gns, 'CuratorBalanceReturnedToBeneficiary');
            await (0, chai_1.expect)(tx)
                .emit(gns, 'SubgraphUpgraded')
                .withArgs(l2SubgraphId, 0, 0, newSubgraph0.subgraphDeploymentID);
            await (0, chai_1.expect)(tx)
                .emit(gns, 'SubgraphVersionUpdated')
                .withArgs(l2SubgraphId, newSubgraph0.subgraphDeploymentID, versionMetadata);
            await (0, chai_1.expect)(tx).emit(gns, 'SubgraphL2TransferFinalized').withArgs(l2SubgraphId);
        });
        it('cannot be called by someone other than the subgraph owner', async function () {
            const { l1SubgraphId, curatedTokens, subgraphMetadata, versionMetadata } = await defaultL1SubgraphParams();
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'uint256', 'address'], [(0, sdk_1.toBN)(0), l1SubgraphId, me.address]);
            await gatewayFinalizeTransfer(l1GNSMock.address, gns.address, curatedTokens, callhookData);
            const l2SubgraphId = await gns.getAliasedL2SubgraphID(l1SubgraphId);
            const tx = gns
                .connect(other)
                .finishSubgraphTransferFromL1(l2SubgraphId, newSubgraph0.subgraphDeploymentID, subgraphMetadata, versionMetadata);
            await (0, chai_1.expect)(tx).revertedWith('GNS: Must be authorized');
        });
        it('rejects calls for a subgraph that does not exist', async function () {
            const l1SubgraphId = await (0, sdk_1.buildSubgraphId)(me.address, (0, sdk_1.toBN)('1'), graph.chainId);
            const metadata = (0, sdk_1.randomHexBytes)();
            const l2SubgraphId = await gns.getAliasedL2SubgraphID(l1SubgraphId);
            const tx = gns
                .connect(me)
                .finishSubgraphTransferFromL1(l2SubgraphId, newSubgraph0.subgraphDeploymentID, metadata, metadata);
            await (0, chai_1.expect)(tx).revertedWith('ERC721: owner query for nonexistent token');
        });
        it('rejects calls for a subgraph that was not transferred from L1', async function () {
            const l2Subgraph = await (0, gnsUtils_1.publishNewSubgraph)(me, newSubgraph0, gns, graph.chainId);
            const metadata = (0, sdk_1.randomHexBytes)();
            const tx = gns
                .connect(me)
                .finishSubgraphTransferFromL1(l2Subgraph.id, newSubgraph0.subgraphDeploymentID, metadata, metadata);
            await (0, chai_1.expect)(tx).revertedWith('INVALID_SUBGRAPH');
        });
        it('accepts calls to a pre-curated subgraph deployment', async function () {
            const { l1SubgraphId, curatedTokens, subgraphMetadata, versionMetadata } = await defaultL1SubgraphParams();
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'uint256', 'address'], [(0, sdk_1.toBN)(0), l1SubgraphId, me.address]);
            await gatewayFinalizeTransfer(l1GNSMock.address, gns.address, curatedTokens, callhookData);
            const l2SubgraphId = await gns.getAliasedL2SubgraphID(l1SubgraphId);
            // Calculate expected signal before minting
            const expectedSignal = await curation.tokensToSignalNoTax(newSubgraph0.subgraphDeploymentID, curatedTokens);
            await grt.connect(me).approve(curation.address, (0, sdk_1.toGRT)('100'));
            await curation.connect(me).mint(newSubgraph0.subgraphDeploymentID, (0, sdk_1.toGRT)('100'), (0, sdk_1.toBN)('0'));
            (0, chai_1.expect)(await curation.getCurationPoolTokens(newSubgraph0.subgraphDeploymentID)).eq((0, sdk_1.toGRT)('100'));
            const tx = gns
                .connect(me)
                .finishSubgraphTransferFromL1(l2SubgraphId, newSubgraph0.subgraphDeploymentID, subgraphMetadata, versionMetadata);
            await (0, chai_1.expect)(tx)
                .emit(gns, 'SubgraphPublished')
                .withArgs(l2SubgraphId, newSubgraph0.subgraphDeploymentID, gnsUtils_1.DEFAULT_RESERVE_RATIO);
            await (0, chai_1.expect)(tx).emit(gns, 'SubgraphMetadataUpdated').withArgs(l2SubgraphId, subgraphMetadata);
            await (0, chai_1.expect)(tx)
                .emit(gns, 'SubgraphUpgraded')
                .withArgs(l2SubgraphId, expectedSignal, curatedTokens, newSubgraph0.subgraphDeploymentID);
            await (0, chai_1.expect)(tx)
                .emit(gns, 'SubgraphVersionUpdated')
                .withArgs(l2SubgraphId, newSubgraph0.subgraphDeploymentID, versionMetadata);
            await (0, chai_1.expect)(tx).emit(gns, 'SubgraphL2TransferFinalized').withArgs(l2SubgraphId);
            const subgraphAfter = await gns.subgraphs(l2SubgraphId);
            const transferDataAfter = await gns.subgraphL2TransferData(l2SubgraphId);
            (0, chai_1.expect)(subgraphAfter.vSignal).eq(expectedSignal);
            (0, chai_1.expect)(transferDataAfter.l2Done).eq(true);
            (0, chai_1.expect)(subgraphAfter.disabled).eq(false);
            (0, chai_1.expect)(subgraphAfter.subgraphDeploymentID).eq(newSubgraph0.subgraphDeploymentID);
            (0, chai_1.expect)(await curation.getCurationPoolTokens(newSubgraph0.subgraphDeploymentID)).eq((0, sdk_1.toGRT)('100').add(curatedTokens));
        });
        it('rejects calls if the subgraph deployment ID is zero', async function () {
            const metadata = (0, sdk_1.randomHexBytes)();
            const { l1SubgraphId, curatedTokens } = await defaultL1SubgraphParams();
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'uint256', 'address'], [(0, sdk_1.toBN)(0), l1SubgraphId, me.address]);
            await gatewayFinalizeTransfer(l1GNSMock.address, gns.address, curatedTokens, callhookData);
            const l2SubgraphId = await gns.getAliasedL2SubgraphID(l1SubgraphId);
            const tx = gns
                .connect(me)
                .finishSubgraphTransferFromL1(l2SubgraphId, HashZero, metadata, metadata);
            await (0, chai_1.expect)(tx).revertedWith('GNS: deploymentID != 0');
        });
        it('rejects calls if the subgraph transfer was already finished', async function () {
            const metadata = (0, sdk_1.randomHexBytes)();
            const { l1SubgraphId, curatedTokens } = await defaultL1SubgraphParams();
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'uint256', 'address'], [(0, sdk_1.toBN)(0), l1SubgraphId, me.address]);
            await gatewayFinalizeTransfer(l1GNSMock.address, gns.address, curatedTokens, callhookData);
            const l2SubgraphId = await gns.getAliasedL2SubgraphID(l1SubgraphId);
            await gns
                .connect(me)
                .finishSubgraphTransferFromL1(l2SubgraphId, newSubgraph0.subgraphDeploymentID, metadata, metadata);
            const tx = gns
                .connect(me)
                .finishSubgraphTransferFromL1(l2SubgraphId, newSubgraph0.subgraphDeploymentID, metadata, metadata);
            await (0, chai_1.expect)(tx).revertedWith('ALREADY_DONE');
        });
    });
    describe('claiming a curator balance with a message from L1 (onTokenTransfer)', function () {
        it('assigns a curator balance to a beneficiary', async function () {
            const l1GNSMockL2Alias = await sdk_1.helpers.getL2SignerFromL1(l1GNSMock.address);
            // Eth for gas:
            await sdk_1.helpers.setBalance(await l1GNSMockL2Alias.getAddress(), (0, utils_1.parseEther)('1'));
            const { l1SubgraphId, curatedTokens, subgraphMetadata, versionMetadata } = await defaultL1SubgraphParams();
            await transferMockSubgraphFromL1(l1SubgraphId, curatedTokens, subgraphMetadata, versionMetadata);
            const l2SubgraphId = await gns.getAliasedL2SubgraphID(l1SubgraphId);
            const l2OwnerSignalBefore = await gns.getCuratorSignal(l2SubgraphId, me.address);
            const newCuratorTokens = (0, sdk_1.toGRT)('10');
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'uint256', 'address'], [(0, sdk_1.toBN)(1), l1SubgraphId, other.address]);
            const tx = await gatewayFinalizeTransfer(l1GNSMock.address, gns.address, newCuratorTokens, callhookData);
            await (0, chai_1.expect)(tx)
                .emit(gns, 'CuratorBalanceReceived')
                .withArgs(l1SubgraphId, l2SubgraphId, other.address, newCuratorTokens);
            const l2NewCuratorSignal = await gns.getCuratorSignal(l2SubgraphId, other.address);
            const expectedNewCuratorSignal = await gns.vSignalToNSignal(l2SubgraphId, await curation.tokensToSignalNoTax(newSubgraph0.subgraphDeploymentID, newCuratorTokens));
            const l2OwnerSignalAfter = await gns.getCuratorSignal(l2SubgraphId, me.address);
            (0, chai_1.expect)(l2OwnerSignalAfter).eq(l2OwnerSignalBefore);
            (0, chai_1.expect)(l2NewCuratorSignal).eq(expectedNewCuratorSignal);
        });
        it('adds the signal to any existing signal for the beneficiary', async function () {
            const l1GNSMockL2Alias = await sdk_1.helpers.getL2SignerFromL1(l1GNSMock.address);
            // Eth for gas:
            await sdk_1.helpers.setBalance(await l1GNSMockL2Alias.getAddress(), (0, utils_1.parseEther)('1'));
            const { l1SubgraphId, curatedTokens, subgraphMetadata, versionMetadata } = await defaultL1SubgraphParams();
            await transferMockSubgraphFromL1(l1SubgraphId, curatedTokens, subgraphMetadata, versionMetadata);
            const l2SubgraphId = await gns.getAliasedL2SubgraphID(l1SubgraphId);
            await grt.connect(governor).mint(other.address, (0, sdk_1.toGRT)('10'));
            await grt.connect(other).approve(gns.address, (0, sdk_1.toGRT)('10'));
            await gns.connect(other).mintSignal(l2SubgraphId, (0, sdk_1.toGRT)('10'), (0, sdk_1.toBN)(0));
            const prevSignal = await gns.getCuratorSignal(l2SubgraphId, other.address);
            const newCuratorTokens = (0, sdk_1.toGRT)('10');
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'uint256', 'address'], [(0, sdk_1.toBN)(1), l1SubgraphId, other.address]);
            const tx = await gatewayFinalizeTransfer(l1GNSMock.address, gns.address, newCuratorTokens, callhookData);
            await (0, chai_1.expect)(tx)
                .emit(gns, 'CuratorBalanceReceived')
                .withArgs(l1SubgraphId, l2SubgraphId, other.address, newCuratorTokens);
            const expectedNewCuratorSignal = await gns.vSignalToNSignal(l2SubgraphId, await curation.tokensToSignalNoTax(newSubgraph0.subgraphDeploymentID, newCuratorTokens));
            const l2CuratorBalance = await gns.getCuratorSignal(l2SubgraphId, other.address);
            (0, chai_1.expect)(l2CuratorBalance).eq(prevSignal.add(expectedNewCuratorSignal));
        });
        it('cannot be called by someone other than the L2GraphTokenGateway', async function () {
            const { l1SubgraphId, curatedTokens, subgraphMetadata, versionMetadata } = await defaultL1SubgraphParams();
            await transferMockSubgraphFromL1(l1SubgraphId, curatedTokens, subgraphMetadata, versionMetadata);
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'uint256', 'address'], [(0, sdk_1.toBN)(1), l1SubgraphId, me.address]);
            const tx = gns.connect(me).onTokenTransfer(l1GNSMock.address, (0, sdk_1.toGRT)('1'), callhookData);
            await (0, chai_1.expect)(tx).revertedWith('ONLY_GATEWAY');
        });
        it('rejects calls if the L1 sender is not the L1GNS', async function () {
            const { l1SubgraphId, curatedTokens, subgraphMetadata, versionMetadata } = await defaultL1SubgraphParams();
            await transferMockSubgraphFromL1(l1SubgraphId, curatedTokens, subgraphMetadata, versionMetadata);
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'uint256', 'address'], [(0, sdk_1.toBN)(1), l1SubgraphId, me.address]);
            const tx = gatewayFinalizeTransfer(me.address, gns.address, (0, sdk_1.toGRT)('1'), callhookData);
            await (0, chai_1.expect)(tx).revertedWith('ONLY_L1_GNS_THROUGH_BRIDGE');
        });
        it('if a subgraph does not exist, it returns the tokens to the beneficiary', async function () {
            const l1GNSMockL2Alias = await sdk_1.helpers.getL2SignerFromL1(l1GNSMock.address);
            // Eth for gas:
            await sdk_1.helpers.setBalance(await l1GNSMockL2Alias.getAddress(), (0, utils_1.parseEther)('1'));
            const { l1SubgraphId } = await defaultL1SubgraphParams();
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'uint256', 'address'], [(0, sdk_1.toBN)(1), l1SubgraphId, me.address]);
            const curatorTokensBefore = await grt.balanceOf(me.address);
            const gnsBalanceBefore = await grt.balanceOf(gns.address);
            const tx = gatewayFinalizeTransfer(l1GNSMock.address, gns.address, (0, sdk_1.toGRT)('1'), callhookData);
            await (0, chai_1.expect)(tx)
                .emit(gns, 'CuratorBalanceReturnedToBeneficiary')
                .withArgs(l1SubgraphId, me.address, (0, sdk_1.toGRT)('1'));
            const curatorTokensAfter = await grt.balanceOf(me.address);
            (0, chai_1.expect)(curatorTokensAfter).eq(curatorTokensBefore.add((0, sdk_1.toGRT)('1')));
            const gnsBalanceAfter = await grt.balanceOf(gns.address);
            // gatewayFinalizeTransfer will mint the tokens that are sent to the curator,
            // so the GNS balance should be the same
            (0, chai_1.expect)(gnsBalanceAfter).eq(gnsBalanceBefore);
        });
        it('for an L2-native subgraph, it sends the tokens to the beneficiary', async function () {
            // This should never really happen unless there's a clash in subgraph IDs (which should
            // also never happen), but we test it anyway to ensure it's a well-defined behavior
            const l1GNSMockL2Alias = await sdk_1.helpers.getL2SignerFromL1(l1GNSMock.address);
            // Eth for gas:
            await sdk_1.helpers.setBalance(await l1GNSMockL2Alias.getAddress(), (0, utils_1.parseEther)('1'));
            const l2Subgraph = await (0, gnsUtils_1.publishNewSubgraph)(me, newSubgraph0, gns, graph.chainId);
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'uint256', 'address'], [(0, sdk_1.toBN)(1), l2Subgraph.id, me.address]);
            const curatorTokensBefore = await grt.balanceOf(me.address);
            const gnsBalanceBefore = await grt.balanceOf(gns.address);
            const tx = gatewayFinalizeTransfer(l1GNSMock.address, gns.address, (0, sdk_1.toGRT)('1'), callhookData);
            await (0, chai_1.expect)(tx)
                .emit(gns, 'CuratorBalanceReturnedToBeneficiary')
                .withArgs(l2Subgraph.id, me.address, (0, sdk_1.toGRT)('1'));
            const curatorTokensAfter = await grt.balanceOf(me.address);
            (0, chai_1.expect)(curatorTokensAfter).eq(curatorTokensBefore.add((0, sdk_1.toGRT)('1')));
            const gnsBalanceAfter = await grt.balanceOf(gns.address);
            // gatewayFinalizeTransfer will mint the tokens that are sent to the curator,
            // so the GNS balance should be the same
            (0, chai_1.expect)(gnsBalanceAfter).eq(gnsBalanceBefore);
        });
        it('if a subgraph transfer was not finished, it returns the tokens to the beneficiary', async function () {
            const l1GNSMockL2Alias = await sdk_1.helpers.getL2SignerFromL1(l1GNSMock.address);
            // Eth for gas:
            await sdk_1.helpers.setBalance(await l1GNSMockL2Alias.getAddress(), (0, utils_1.parseEther)('1'));
            const { l1SubgraphId, curatedTokens } = await defaultL1SubgraphParams();
            const callhookDataSG = utils_1.defaultAbiCoder.encode(['uint8', 'uint256', 'address'], [(0, sdk_1.toBN)(0), l1SubgraphId, me.address]);
            await gatewayFinalizeTransfer(l1GNSMock.address, gns.address, curatedTokens, callhookDataSG);
            // At this point the SG exists, but transfer is not finished
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'uint256', 'address'], [(0, sdk_1.toBN)(1), l1SubgraphId, me.address]);
            const curatorTokensBefore = await grt.balanceOf(me.address);
            const gnsBalanceBefore = await grt.balanceOf(gns.address);
            const tx = gatewayFinalizeTransfer(l1GNSMock.address, gns.address, (0, sdk_1.toGRT)('1'), callhookData);
            await (0, chai_1.expect)(tx)
                .emit(gns, 'CuratorBalanceReturnedToBeneficiary')
                .withArgs(l1SubgraphId, me.address, (0, sdk_1.toGRT)('1'));
            const curatorTokensAfter = await grt.balanceOf(me.address);
            (0, chai_1.expect)(curatorTokensAfter).eq(curatorTokensBefore.add((0, sdk_1.toGRT)('1')));
            const gnsBalanceAfter = await grt.balanceOf(gns.address);
            // gatewayFinalizeTransfer will mint the tokens that are sent to the curator,
            // so the GNS balance should be the same
            (0, chai_1.expect)(gnsBalanceAfter).eq(gnsBalanceBefore);
        });
        it('protects the curator against a rounding attack', async function () {
            // Transfer a subgraph from L1 with only 1 wei GRT of curated signal
            const { l1SubgraphId, subgraphMetadata, versionMetadata } = await defaultL1SubgraphParams();
            const curatedTokens = (0, sdk_1.toBN)('1');
            await transferMockSubgraphFromL1(l1SubgraphId, curatedTokens, subgraphMetadata, versionMetadata);
            // Prepare the rounding attack by setting up an indexer and collecting a lot of query fees
            const curatorTokens = (0, sdk_1.toGRT)('10000');
            const collectTokens = curatorTokens.mul(20);
            await staking.connect(governor).setCurationPercentage(100000);
            // Set up an indexer account with some stake
            await grt.connect(governor).mint(attacker.address, (0, sdk_1.toGRT)('1000000'));
            await grt.connect(attacker).approve(staking.address, (0, sdk_1.toGRT)('1000000'));
            await staking.connect(attacker).stake((0, sdk_1.toGRT)('100000'));
            const channelKey = (0, sdk_1.deriveChannelKey)();
            // Allocate to the same deployment ID
            await staking
                .connect(attacker)
                .allocateFrom(attacker.address, newSubgraph0.subgraphDeploymentID, (0, sdk_1.toGRT)('100000'), channelKey.address, (0, sdk_1.randomHexBytes)(32), await channelKey.generateProof(attacker.address));
            // Spoof some query fees, 10% of which will go to the Curation pool
            await staking.connect(attacker).collect(collectTokens, channelKey.address);
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'uint256', 'address'], [(0, sdk_1.toBN)(1), l1SubgraphId, me.address]);
            const curatorTokensBefore = await grt.balanceOf(me.address);
            const gnsBalanceBefore = await grt.balanceOf(gns.address);
            const tx = gatewayFinalizeTransfer(l1GNSMock.address, gns.address, curatorTokens, callhookData);
            await (0, chai_1.expect)(tx)
                .emit(gns, 'CuratorBalanceReturnedToBeneficiary')
                .withArgs(l1SubgraphId, me.address, curatorTokens);
            const curatorTokensAfter = await grt.balanceOf(me.address);
            (0, chai_1.expect)(curatorTokensAfter).eq(curatorTokensBefore.add(curatorTokens));
            const gnsBalanceAfter = await grt.balanceOf(gns.address);
            // gatewayFinalizeTransfer will mint the tokens that are sent to the curator,
            // so the GNS balance should be the same
            (0, chai_1.expect)(gnsBalanceAfter).eq(gnsBalanceBefore);
        });
        it('if a subgraph was deprecated after transfer, it returns the tokens to the beneficiary', async function () {
            const l1GNSMockL2Alias = await sdk_1.helpers.getL2SignerFromL1(l1GNSMock.address);
            // Eth for gas:
            await sdk_1.helpers.setBalance(await l1GNSMockL2Alias.getAddress(), (0, utils_1.parseEther)('1'));
            const { l1SubgraphId, curatedTokens, subgraphMetadata, versionMetadata } = await defaultL1SubgraphParams();
            await transferMockSubgraphFromL1(l1SubgraphId, curatedTokens, subgraphMetadata, versionMetadata);
            const l2SubgraphId = await gns.getAliasedL2SubgraphID(l1SubgraphId);
            await gns.connect(me).deprecateSubgraph(l2SubgraphId);
            // SG was transferred, but is deprecated now!
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'uint256', 'address'], [(0, sdk_1.toBN)(1), l1SubgraphId, me.address]);
            const curatorTokensBefore = await grt.balanceOf(me.address);
            const gnsBalanceBefore = await grt.balanceOf(gns.address);
            const tx = gatewayFinalizeTransfer(l1GNSMock.address, gns.address, (0, sdk_1.toGRT)('1'), callhookData);
            await (0, chai_1.expect)(tx)
                .emit(gns, 'CuratorBalanceReturnedToBeneficiary')
                .withArgs(l1SubgraphId, me.address, (0, sdk_1.toGRT)('1'));
            const curatorTokensAfter = await grt.balanceOf(me.address);
            (0, chai_1.expect)(curatorTokensAfter).eq(curatorTokensBefore.add((0, sdk_1.toGRT)('1')));
            const gnsBalanceAfter = await grt.balanceOf(gns.address);
            // gatewayFinalizeTransfer will mint the tokens that are sent to the curator,
            // so the GNS balance should be the same
            (0, chai_1.expect)(gnsBalanceAfter).eq(gnsBalanceBefore);
        });
    });
    describe('onTokenTransfer with invalid codes', function () {
        it('reverts', async function () {
            // This should never really happen unless the Arbitrum bridge is compromised,
            // so we test it anyway to ensure it's a well-defined behavior.
            // code 2 does not exist:
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'uint256', 'address'], [(0, sdk_1.toBN)(2), (0, sdk_1.toBN)(1337), me.address]);
            const tx = gatewayFinalizeTransfer(l1GNSMock.address, gns.address, (0, sdk_1.toGRT)('1'), callhookData);
            await (0, chai_1.expect)(tx).revertedWith('INVALID_CODE');
        });
    });
    describe('getAliasedL2SubgraphID', function () {
        it('returns the L2 subgraph ID that is the L1 subgraph ID with an offset', async function () {
            const l1SubgraphId = ethers_1.ethers.BigNumber.from('68799548758199140224151701590582019137924969401915573086349306511960790045480');
            const l2SubgraphId = await gns.getAliasedL2SubgraphID(l1SubgraphId);
            const offset = ethers_1.ethers.BigNumber.from('0x1111000000000000000000000000000000000000000000000000000000001111');
            const base = ethers_1.ethers.constants.MaxUint256.add(1);
            const expectedL2SubgraphId = l1SubgraphId.add(offset).mod(base);
            (0, chai_1.expect)(l2SubgraphId).eq(expectedL2SubgraphId);
        });
        it('wraps around MAX_UINT256 in case of overflow', async function () {
            const l1SubgraphId = ethers_1.ethers.constants.MaxUint256;
            const l2SubgraphId = await gns.getAliasedL2SubgraphID(l1SubgraphId);
            const offset = ethers_1.ethers.BigNumber.from('0x1111000000000000000000000000000000000000000000000000000000001111');
            const base = ethers_1.ethers.constants.MaxUint256.add(1);
            const expectedL2SubgraphId = l1SubgraphId.add(offset).mod(base);
            (0, chai_1.expect)(l2SubgraphId).eq(expectedL2SubgraphId);
        });
    });
});
