"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const hardhat_1 = __importDefault(require("hardhat"));
const chai_1 = require("chai");
const graphTokenTests_1 = require("../lib/graphTokenTests");
const fixtures_1 = require("../lib/fixtures");
const sdk_1 = require("@graphprotocol/sdk");
describe('L2GraphToken', () => {
    describe('Base GRT behavior', () => {
        graphTokenTests_1.grtTests.bind(this)(true);
    });
    describe('Extended L2 behavior', () => {
        const graph = hardhat_1.default.graph();
        let mockL2Gateway;
        let mockL1GRT;
        let governor;
        let user;
        let contracts;
        let fixture;
        let grt;
        before(async function () {
            [mockL1GRT, mockL2Gateway, user] = await graph.getTestAccounts();
            ({ governor } = await graph.getNamedAccounts());
            fixture = new fixtures_1.NetworkFixture(graph.provider);
            contracts = await fixture.load(governor, true);
            grt = contracts.L2GraphToken;
        });
        beforeEach(async function () {
            await fixture.setUp();
        });
        afterEach(async function () {
            await fixture.tearDown();
        });
        describe('setGateway', function () {
            it('cannot be called by someone who is not the governor', async function () {
                const tx = grt.connect(mockL2Gateway).setGateway(mockL2Gateway.address);
                await (0, chai_1.expect)(tx).revertedWith('Only Governor can call');
            });
            it('sets the L2 Gateway address when called by the governor', async function () {
                const tx = grt.connect(governor).setGateway(mockL2Gateway.address);
                await (0, chai_1.expect)(tx).emit(grt, 'GatewaySet').withArgs(mockL2Gateway.address);
                (0, chai_1.expect)(await grt.gateway()).eq(mockL2Gateway.address);
            });
        });
        describe('setL1Address', function () {
            it('cannot be called by someone who is not the governor', async function () {
                const tx = grt.connect(mockL2Gateway).setL1Address(mockL1GRT.address);
                await (0, chai_1.expect)(tx).revertedWith('Only Governor can call');
            });
            it('sets the L1 GRT address when called by the governor', async function () {
                const tx = grt.connect(governor).setL1Address(mockL1GRT.address);
                await (0, chai_1.expect)(tx).emit(grt, 'L1AddressSet').withArgs(mockL1GRT.address);
                (0, chai_1.expect)(await grt.l1Address()).eq(mockL1GRT.address);
            });
        });
        describe('bridge minting and burning', function () {
            beforeEach(async function () {
                // Configure the l1Address and gateway
                await grt.connect(governor).setL1Address(mockL1GRT.address);
                await grt.connect(governor).setGateway(mockL2Gateway.address);
            });
            describe('bridgeMint', function () {
                it('cannot be called by someone who is not the gateway', async function () {
                    const tx = grt.connect(governor).bridgeMint(user.address, (0, sdk_1.toGRT)('100'));
                    await (0, chai_1.expect)(tx).revertedWith('NOT_GATEWAY');
                });
                it('mints GRT into a destination account', async function () {
                    const tx = grt.connect(mockL2Gateway).bridgeMint(user.address, (0, sdk_1.toGRT)('100'));
                    await (0, chai_1.expect)(tx).emit(grt, 'BridgeMinted').withArgs(user.address, (0, sdk_1.toGRT)('100'));
                    (0, chai_1.expect)(await grt.balanceOf(user.address)).eq((0, sdk_1.toGRT)('100'));
                });
            });
            describe('bridgeBurn', function () {
                it('cannot be called by someone who is not the gateway', async function () {
                    const tx = grt.connect(governor).bridgeBurn(user.address, (0, sdk_1.toGRT)('100'));
                    await (0, chai_1.expect)(tx).revertedWith('NOT_GATEWAY');
                });
                it('requires approval for burning', async function () {
                    await grt.connect(mockL2Gateway).bridgeMint(user.address, (0, sdk_1.toGRT)('100'));
                    const tx = grt.connect(mockL2Gateway).bridgeBurn(user.address, (0, sdk_1.toGRT)('20'));
                    await (0, chai_1.expect)(tx).revertedWith('ERC20: burn amount exceeds allowance');
                });
                it('fails if the user does not have enough funds', async function () {
                    await grt.connect(mockL2Gateway).bridgeMint(user.address, (0, sdk_1.toGRT)('10'));
                    await grt.connect(user).approve(mockL2Gateway.address, (0, sdk_1.toGRT)('20'));
                    const tx = grt.connect(mockL2Gateway).bridgeBurn(user.address, (0, sdk_1.toGRT)('20'));
                    await (0, chai_1.expect)(tx).revertedWith('ERC20: burn amount exceeds balance');
                });
                it('burns GRT from an account when approved', async function () {
                    await grt.connect(mockL2Gateway).bridgeMint(user.address, (0, sdk_1.toGRT)('100'));
                    await grt.connect(user).approve(mockL2Gateway.address, (0, sdk_1.toGRT)('20'));
                    const tx = grt.connect(mockL2Gateway).bridgeBurn(user.address, (0, sdk_1.toGRT)('20'));
                    await (0, chai_1.expect)(tx).emit(grt, 'BridgeBurned').withArgs(user.address, (0, sdk_1.toGRT)('20'));
                    (0, chai_1.expect)(await grt.balanceOf(user.address)).eq((0, sdk_1.toGRT)('80'));
                });
            });
            it('does not allow the bridge to mint as a regular minter', async function () {
                const tx = grt.connect(mockL2Gateway).mint(user.address, (0, sdk_1.toGRT)('100'));
                await (0, chai_1.expect)(tx).revertedWith('Only minter can call');
            });
        });
    });
});
