"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const hardhat_1 = __importDefault(require("hardhat"));
const chai_1 = require("chai");
const ethers_1 = require("ethers");
const fixtures_1 = require("../lib/fixtures");
const smock_1 = require("@defi-wonderland/smock");
(0, chai_1.use)(smock_1.smock.matchers);
const sdk_1 = require("@graphprotocol/sdk");
const { AddressZero } = ethers_1.constants;
describe('L2GraphTokenGateway', () => {
    const graph = hardhat_1.default.graph();
    let me;
    let governor;
    let tokenSender;
    let l1Receiver;
    let l2Receiver;
    let pauseGuardian;
    let fixture;
    let arbSysMock;
    let fixtureContracts;
    let l1MockContracts;
    let l1GRTMock;
    let l1GRTGatewayMock;
    let routerMock;
    let grt;
    let l2GraphTokenGateway;
    let callhookReceiverMock;
    let rewardsManager;
    const senderTokens = (0, sdk_1.toGRT)('1000');
    const defaultData = '0x';
    const defaultDataWithNotEmptyCallHookData = ethers_1.utils.defaultAbiCoder.encode(['uint256', 'uint256'], [(0, sdk_1.toBN)('1337'), (0, sdk_1.toBN)('42')]);
    before(async function () {
        [me, governor, tokenSender, l1Receiver, l2Receiver, pauseGuardian]
            = await graph.getTestAccounts();
        fixture = new fixtures_1.NetworkFixture(graph.provider);
        // Deploy L2
        fixtureContracts = await fixture.load(governor, true);
        grt = fixtureContracts.GraphToken;
        l2GraphTokenGateway = fixtureContracts.L2GraphTokenGateway;
        rewardsManager = fixtureContracts.RewardsManager;
        ({ routerMock } = await fixture.loadL2ArbitrumBridge(governor));
        // Deploy L1 mock
        l1MockContracts = await fixture.loadMock(false);
        l1GRTMock = l1MockContracts.GraphToken;
        l1GRTGatewayMock = l1MockContracts.L1GraphTokenGateway;
        callhookReceiverMock = (await (0, sdk_1.deploy)(sdk_1.DeployType.Deploy, governor, {
            name: 'CallhookReceiverMock',
        })).contract;
        // Give some funds to the token sender and router mock
        await grt.connect(governor).mint(tokenSender.address, senderTokens);
        await sdk_1.helpers.setBalance(routerMock.address, ethers_1.utils.parseEther('1'));
    });
    beforeEach(async function () {
        await fixture.setUp();
        // Thanks to Livepeer: https://github.com/livepeer/arbitrum-lpt-bridge/blob/main/test/unit/L2/l2LPTGateway.test.ts#L86
        arbSysMock = await smock_1.smock.fake('ArbSys', {
            address: '0x0000000000000000000000000000000000000064',
        });
        arbSysMock.sendTxToL1.returns(1);
    });
    afterEach(async function () {
        await fixture.tearDown();
    });
    context('> immediately after deploy', function () {
        describe('calculateL2TokenAddress', function () {
            it('should return the zero address', async function () {
                (0, chai_1.expect)(await l2GraphTokenGateway.calculateL2TokenAddress(grt.address)).eq(AddressZero);
            });
        });
        describe('outboundTransfer', function () {
            it('reverts because it is paused', async function () {
                const tx = l2GraphTokenGateway
                    .connect(tokenSender)['outboundTransfer(address,address,uint256,bytes)'](grt.address, l1Receiver.address, (0, sdk_1.toGRT)('10'), defaultData);
                await (0, chai_1.expect)(tx).revertedWith('Paused (contract)');
            });
        });
        describe('finalizeInboundTransfer', function () {
            it('revert because it is paused', async function () {
                const tx = l2GraphTokenGateway
                    .connect(tokenSender)
                    .finalizeInboundTransfer(grt.address, tokenSender.address, l1Receiver.address, (0, sdk_1.toGRT)('10'), defaultData);
                await (0, chai_1.expect)(tx).revertedWith('Paused (contract)');
            });
        });
        describe('setL2Router', function () {
            it('is not callable by addreses that are not the governor', async function () {
                const tx = l2GraphTokenGateway.connect(tokenSender).setL2Router(routerMock.address);
                await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
            });
            it('sets router address', async function () {
                const tx = l2GraphTokenGateway.connect(governor).setL2Router(routerMock.address);
                await (0, chai_1.expect)(tx).emit(l2GraphTokenGateway, 'L2RouterSet').withArgs(routerMock.address);
                (0, chai_1.expect)(await l2GraphTokenGateway.l2Router()).eq(routerMock.address);
            });
        });
        describe('setL1TokenAddress', function () {
            it('is not callable by addreses that are not the governor', async function () {
                const tx = l2GraphTokenGateway.connect(tokenSender).setL1TokenAddress(l1GRTMock.address);
                await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
            });
            it('sets l2GRT', async function () {
                const tx = l2GraphTokenGateway.connect(governor).setL1TokenAddress(l1GRTMock.address);
                await (0, chai_1.expect)(tx).emit(l2GraphTokenGateway, 'L1TokenAddressSet').withArgs(l1GRTMock.address);
                (0, chai_1.expect)(await l2GraphTokenGateway.l1GRT()).eq(l1GRTMock.address);
            });
        });
        describe('setL1CounterpartAddress', function () {
            it('is not callable by addreses that are not the governor', async function () {
                const tx = l2GraphTokenGateway
                    .connect(tokenSender)
                    .setL1CounterpartAddress(l1GRTGatewayMock.address);
                await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
            });
            it('sets L1Counterpart', async function () {
                const tx = l2GraphTokenGateway
                    .connect(governor)
                    .setL1CounterpartAddress(l1GRTGatewayMock.address);
                await (0, chai_1.expect)(tx)
                    .emit(l2GraphTokenGateway, 'L1CounterpartAddressSet')
                    .withArgs(l1GRTGatewayMock.address);
                (0, chai_1.expect)(await l2GraphTokenGateway.l1Counterpart()).eq(l1GRTGatewayMock.address);
            });
        });
        describe('Pausable behavior', () => {
            it('cannot be paused or unpaused by someone other than governor or pauseGuardian', async () => {
                let tx = l2GraphTokenGateway.connect(tokenSender).setPaused(false);
                await (0, chai_1.expect)(tx).revertedWith('Only Governor or Guardian');
                tx = l2GraphTokenGateway.connect(tokenSender).setPaused(true);
                await (0, chai_1.expect)(tx).revertedWith('Only Governor or Guardian');
            });
            it('cannot be paused if some state variables are not set', async function () {
                let tx = l2GraphTokenGateway.connect(governor).setPaused(false);
                await (0, chai_1.expect)(tx).revertedWith('L2_ROUTER_NOT_SET');
                await l2GraphTokenGateway.connect(governor).setL2Router(routerMock.address);
                tx = l2GraphTokenGateway.connect(governor).setPaused(false);
                await (0, chai_1.expect)(tx).revertedWith('L1_COUNTERPART_NOT_SET');
                await l2GraphTokenGateway
                    .connect(governor)
                    .setL1CounterpartAddress(l1GRTGatewayMock.address);
                tx = l2GraphTokenGateway.connect(governor).setPaused(false);
                await (0, chai_1.expect)(tx).revertedWith('L1_GRT_NOT_SET');
            });
            it('can be paused and unpaused by the governor', async function () {
                await fixture.configureL2Bridge(governor, fixtureContracts, l1MockContracts);
                let tx = l2GraphTokenGateway.connect(governor).setPaused(true);
                await (0, chai_1.expect)(tx).emit(l2GraphTokenGateway, 'PauseChanged').withArgs(true);
                (0, chai_1.expect)(await l2GraphTokenGateway.paused()).eq(true);
                tx = l2GraphTokenGateway.connect(governor).setPaused(false);
                await (0, chai_1.expect)(tx).emit(l2GraphTokenGateway, 'PauseChanged').withArgs(false);
                (0, chai_1.expect)(await l2GraphTokenGateway.paused()).eq(false);
            });
            describe('setPauseGuardian', function () {
                it('cannot be called by someone other than governor', async function () {
                    const tx = l2GraphTokenGateway
                        .connect(tokenSender)
                        .setPauseGuardian(pauseGuardian.address);
                    await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
                });
                it('sets a new pause guardian', async function () {
                    const currentPauseGuardian = await l2GraphTokenGateway.pauseGuardian();
                    const tx = l2GraphTokenGateway.connect(governor).setPauseGuardian(pauseGuardian.address);
                    await (0, chai_1.expect)(tx)
                        .emit(l2GraphTokenGateway, 'NewPauseGuardian')
                        .withArgs(currentPauseGuardian, pauseGuardian.address);
                });
                it('allows a pause guardian to pause and unpause', async function () {
                    await fixture.configureL2Bridge(governor, fixtureContracts, l1MockContracts);
                    await l2GraphTokenGateway.connect(governor).setPauseGuardian(pauseGuardian.address);
                    let tx = l2GraphTokenGateway.connect(pauseGuardian).setPaused(true);
                    await (0, chai_1.expect)(tx).emit(l2GraphTokenGateway, 'PauseChanged').withArgs(true);
                    (0, chai_1.expect)(await l2GraphTokenGateway.paused()).eq(true);
                    tx = l2GraphTokenGateway.connect(pauseGuardian).setPaused(false);
                    await (0, chai_1.expect)(tx).emit(l2GraphTokenGateway, 'PauseChanged').withArgs(false);
                    (0, chai_1.expect)(await l2GraphTokenGateway.paused()).eq(false);
                });
            });
        });
    });
    context('> after configuring and unpausing', function () {
        const testValidOutboundTransfer = async function (signer, data) {
            const tx = l2GraphTokenGateway
                .connect(signer)['outboundTransfer(address,address,uint256,bytes)'](l1GRTMock.address, l1Receiver.address, (0, sdk_1.toGRT)('10'), data);
            const expectedId = 1;
            await (0, chai_1.expect)(tx)
                .emit(l2GraphTokenGateway, 'WithdrawalInitiated')
                .withArgs(l1GRTMock.address, tokenSender.address, l1Receiver.address, expectedId, 0, (0, sdk_1.toGRT)('10'));
            // Should use the L1 Gateway's interface, but both come from ITokenGateway
            const calldata = l2GraphTokenGateway.interface.encodeFunctionData('finalizeInboundTransfer', [
                l1GRTMock.address,
                tokenSender.address,
                l1Receiver.address,
                (0, sdk_1.toGRT)('10'),
                ethers_1.utils.defaultAbiCoder.encode(['uint256', 'bytes'], [0, []]),
            ]);
            await (0, chai_1.expect)(tx)
                .emit(l2GraphTokenGateway, 'TxToL1')
                .withArgs(tokenSender.address, l1GRTGatewayMock.address, 1, calldata);
            // For some reason the call count doesn't work properly,
            // and each function call is counted 12 times.
            // Possibly related to https://github.com/defi-wonderland/smock/issues/85 ?
            // expect(arbSysMock.sendTxToL1).to.have.been.calledOnce
            (0, chai_1.expect)(arbSysMock.sendTxToL1).to.have.been.calledWith(l1GRTGatewayMock.address, calldata);
            const senderBalance = await grt.balanceOf(tokenSender.address);
            (0, chai_1.expect)(senderBalance).eq((0, sdk_1.toGRT)('990'));
        };
        before(async function () {
            await fixture.configureL2Bridge(governor, fixtureContracts, l1MockContracts);
        });
        describe('calculateL2TokenAddress', function () {
            it('returns the L2 token address', async function () {
                (0, chai_1.expect)(await l2GraphTokenGateway.calculateL2TokenAddress(l1GRTMock.address)).eq(grt.address);
            });
            it('returns the zero address if the input is any other address', async function () {
                (0, chai_1.expect)(await l2GraphTokenGateway.calculateL2TokenAddress(tokenSender.address)).eq(AddressZero);
            });
        });
        describe('outboundTransfer', function () {
            it('reverts when called with the wrong token address', async function () {
                const tx = l2GraphTokenGateway
                    .connect(tokenSender)['outboundTransfer(address,address,uint256,bytes)'](tokenSender.address, l1Receiver.address, (0, sdk_1.toGRT)('10'), defaultData);
                await (0, chai_1.expect)(tx).revertedWith('TOKEN_NOT_GRT');
            });
            it('burns tokens and triggers an L1 call', async function () {
                await grt.connect(tokenSender).approve(l2GraphTokenGateway.address, (0, sdk_1.toGRT)('10'));
                await testValidOutboundTransfer(tokenSender, defaultData);
            });
            it('decodes the sender address from messages sent by the router', async function () {
                await grt.connect(tokenSender).approve(l2GraphTokenGateway.address, (0, sdk_1.toGRT)('10'));
                const routerEncodedData = ethers_1.utils.defaultAbiCoder.encode(['address', 'bytes'], [tokenSender.address, defaultData]);
                await testValidOutboundTransfer(routerMock, routerEncodedData);
            });
            it('reverts when called with nonempty calldata', async function () {
                await grt.connect(tokenSender).approve(l2GraphTokenGateway.address, (0, sdk_1.toGRT)('10'));
                const tx = l2GraphTokenGateway
                    .connect(tokenSender)['outboundTransfer(address,address,uint256,bytes)'](l1GRTMock.address, l1Receiver.address, (0, sdk_1.toGRT)('10'), defaultDataWithNotEmptyCallHookData);
                await (0, chai_1.expect)(tx).revertedWith('CALL_HOOK_DATA_NOT_ALLOWED');
            });
            it('reverts when the sender does not have enough GRT', async function () {
                await grt.connect(tokenSender).approve(l2GraphTokenGateway.address, (0, sdk_1.toGRT)('1001'));
                const tx = l2GraphTokenGateway
                    .connect(tokenSender)['outboundTransfer(address,address,uint256,bytes)'](l1GRTMock.address, l1Receiver.address, (0, sdk_1.toGRT)('1001'), defaultData);
                await (0, chai_1.expect)(tx).revertedWith('ERC20: burn amount exceeds balance');
            });
        });
        describe('finalizeInboundTransfer', function () {
            const testValidFinalizeTransfer = async function (data, to) {
                to = to ?? l2Receiver.address;
                const l1GRTGatewayMockL2Alias = await sdk_1.helpers.getL2SignerFromL1(l1GRTGatewayMock.address);
                await me.sendTransaction({
                    to: await l1GRTGatewayMockL2Alias.getAddress(),
                    value: ethers_1.utils.parseUnits('1', 'ether'),
                });
                const tx = l2GraphTokenGateway
                    .connect(l1GRTGatewayMockL2Alias)
                    .finalizeInboundTransfer(l1GRTMock.address, tokenSender.address, to, (0, sdk_1.toGRT)('10'), data);
                await (0, chai_1.expect)(tx)
                    .emit(l2GraphTokenGateway, 'DepositFinalized')
                    .withArgs(l1GRTMock.address, tokenSender.address, to, (0, sdk_1.toGRT)('10'));
                await (0, chai_1.expect)(tx).emit(grt, 'BridgeMinted').withArgs(to, (0, sdk_1.toGRT)('10'));
                // Unchanged
                const senderBalance = await grt.balanceOf(tokenSender.address);
                (0, chai_1.expect)(senderBalance).eq((0, sdk_1.toGRT)('1000'));
                // 10 newly minted GRT
                const receiverBalance = await grt.balanceOf(to);
                (0, chai_1.expect)(receiverBalance).eq((0, sdk_1.toGRT)('10'));
                return tx;
            };
            it('reverts when called by an account that is not the gateway', async function () {
                const tx = l2GraphTokenGateway
                    .connect(tokenSender)
                    .finalizeInboundTransfer(l1GRTMock.address, tokenSender.address, l2Receiver.address, (0, sdk_1.toGRT)('10'), defaultData);
                await (0, chai_1.expect)(tx).revertedWith('ONLY_COUNTERPART_GATEWAY');
            });
            it('reverts when called by an account that is the gateway but without the L2 alias', async function () {
                const impersonatedGateway = await sdk_1.helpers.impersonateAccount(l1GRTGatewayMock.address);
                const tx = l2GraphTokenGateway
                    .connect(impersonatedGateway)
                    .finalizeInboundTransfer(l1GRTMock.address, tokenSender.address, l2Receiver.address, (0, sdk_1.toGRT)('10'), defaultData);
                await (0, chai_1.expect)(tx).revertedWith('ONLY_COUNTERPART_GATEWAY');
            });
            it('mints and sends tokens when called by the aliased gateway', async function () {
                await testValidFinalizeTransfer(defaultData);
            });
            it('calls a callhook if the transfer includes calldata', async function () {
                const tx = await testValidFinalizeTransfer(defaultDataWithNotEmptyCallHookData, callhookReceiverMock.address);
                // Emitted by the callhook:
                await (0, chai_1.expect)(tx)
                    .emit(callhookReceiverMock, 'TransferReceived')
                    .withArgs(tokenSender.address, (0, sdk_1.toGRT)('10'), (0, sdk_1.toBN)('1337'), (0, sdk_1.toBN)('42'));
            });
            it('reverts if a callhook reverts', async function () {
                // The 0 will make the callhook revert (see CallhookReceiverMock.sol)
                const callHookData = ethers_1.utils.defaultAbiCoder.encode(['uint256', 'uint256'], [(0, sdk_1.toBN)('0'), (0, sdk_1.toBN)('42')]);
                const l1GRTGatewayMockL2Alias = await sdk_1.helpers.getL2SignerFromL1(l1GRTGatewayMock.address);
                await me.sendTransaction({
                    to: await l1GRTGatewayMockL2Alias.getAddress(),
                    value: ethers_1.utils.parseUnits('1', 'ether'),
                });
                const tx = l2GraphTokenGateway
                    .connect(l1GRTGatewayMockL2Alias)
                    .finalizeInboundTransfer(l1GRTMock.address, tokenSender.address, callhookReceiverMock.address, (0, sdk_1.toGRT)('10'), callHookData);
                await (0, chai_1.expect)(tx).revertedWith('FOO_IS_ZERO');
            });
            it('reverts if trying to call a callhook in a contract that does not implement onTokenTransfer', async function () {
                const callHookData = ethers_1.utils.defaultAbiCoder.encode(['uint256'], [(0, sdk_1.toBN)('0')]);
                const l1GRTGatewayMockL2Alias = await sdk_1.helpers.getL2SignerFromL1(l1GRTGatewayMock.address);
                await me.sendTransaction({
                    to: await l1GRTGatewayMockL2Alias.getAddress(),
                    value: ethers_1.utils.parseUnits('1', 'ether'),
                });
                // RewardsManager does not implement onTokenTransfer, so this will fail
                const tx = l2GraphTokenGateway
                    .connect(l1GRTGatewayMockL2Alias)
                    .finalizeInboundTransfer(l1GRTMock.address, tokenSender.address, rewardsManager.address, (0, sdk_1.toGRT)('10'), callHookData);
                await (0, chai_1.expect)(tx).revertedWith('function selector was not recognized and there\'s no fallback function');
            });
        });
    });
});
