"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const hardhat_1 = __importDefault(require("hardhat"));
const chai_1 = require("chai");
const ethers_1 = require("ethers");
const utils_1 = require("ethers/lib/utils");
const fixtures_1 = require("../lib/fixtures");
const sdk_1 = require("@graphprotocol/sdk");
const { AddressZero } = ethers_1.ethers.constants;
const subgraphDeploymentID = (0, sdk_1.randomHexBytes)();
const channelKey = (0, sdk_1.deriveChannelKey)();
const allocationID = channelKey.address;
const metadata = (0, sdk_1.randomHexBytes)(32);
describe('L2Staking', () => {
    const graph = hardhat_1.default.graph();
    let me;
    let other;
    let governor;
    let fixture;
    let fixtureContracts;
    let l1MockContracts;
    let l1GRTMock;
    let l1StakingMock;
    let l1GNSMock;
    let l1GRTGatewayMock;
    let l2GraphTokenGateway;
    let staking;
    let grt;
    const tokens10k = (0, sdk_1.toGRT)('10000');
    const tokens100k = (0, sdk_1.toGRT)('100000');
    const tokens1m = (0, sdk_1.toGRT)('1000000');
    // Allocate with test values
    const allocate = async (tokens) => {
        return staking
            .connect(me)
            .allocateFrom(me.address, subgraphDeploymentID, tokens, allocationID, metadata, await channelKey.generateProof(me.address));
    };
    const gatewayFinalizeTransfer = async function (from, to, amount, callhookData) {
        const mockL1GatewayL2Alias = await sdk_1.helpers.getL2SignerFromL1(l1GRTGatewayMock.address);
        // Eth for gas:
        await sdk_1.helpers.setBalance(await mockL1GatewayL2Alias.getAddress(), (0, utils_1.parseEther)('1'));
        const tx = l2GraphTokenGateway
            .connect(mockL1GatewayL2Alias)
            .finalizeInboundTransfer(l1GRTMock.address, from, to, amount, callhookData);
        return tx;
    };
    before(async function () {
        [me, other] = await graph.getTestAccounts();
        ({ governor } = await graph.getNamedAccounts());
        fixture = new fixtures_1.NetworkFixture(graph.provider);
        // Deploy L2
        fixtureContracts = await fixture.load(governor, true);
        grt = fixtureContracts.GraphToken;
        staking = fixtureContracts.Staking;
        l2GraphTokenGateway = fixtureContracts.L2GraphTokenGateway;
        // Deploy L1 mock
        l1MockContracts = await fixture.loadMock(false);
        l1GRTMock = l1MockContracts.GraphToken;
        l1StakingMock = l1MockContracts.L1Staking;
        l1GNSMock = l1MockContracts.L1GNS;
        l1GRTGatewayMock = l1MockContracts.L1GraphTokenGateway;
        // Deploy L2 arbitrum bridge
        await fixture.loadL2ArbitrumBridge(governor);
        // Configure L2 bridge
        await fixture.configureL2Bridge(governor, fixtureContracts, l1MockContracts);
        await grt.connect(governor).mint(me.address, tokens1m);
        await grt.connect(me).approve(staking.address, tokens1m);
        await grt.connect(governor).mint(other.address, tokens1m);
        await grt.connect(other).approve(staking.address, tokens1m);
    });
    beforeEach(async function () {
        await fixture.setUp();
    });
    afterEach(async function () {
        await fixture.tearDown();
    });
    describe('receive()', function () {
        it('should not allow receiving ETH', async function () {
            const tx = me.sendTransaction({
                to: staking.address,
                value: (0, utils_1.parseEther)('1'),
            });
            await (0, chai_1.expect)(tx).revertedWith('RECEIVE_ETH_NOT_ALLOWED');
        });
    });
    describe('receiving indexer stake from L1 (onTokenTransfer)', function () {
        it('cannot be called by someone other than the L2GraphTokenGateway', async function () {
            const functionData = utils_1.defaultAbiCoder.encode(['tuple(address)'], [[me.address]]);
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'bytes'], [(0, sdk_1.toBN)(0), functionData]);
            const tx = staking.connect(me).onTokenTransfer(l1GNSMock.address, tokens100k, callhookData);
            await (0, chai_1.expect)(tx).revertedWith('ONLY_GATEWAY');
        });
        it('rejects calls if the L1 sender is not the L1Staking', async function () {
            const functionData = utils_1.defaultAbiCoder.encode(['tuple(address)'], [[me.address]]);
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'bytes'], [(0, sdk_1.toBN)(0), functionData]);
            const tx = gatewayFinalizeTransfer(me.address, staking.address, tokens100k, callhookData);
            await (0, chai_1.expect)(tx).revertedWith('ONLY_L1_STAKING_THROUGH_BRIDGE');
        });
        it('adds stake to a new indexer', async function () {
            const functionData = utils_1.defaultAbiCoder.encode(['tuple(address)'], [[me.address]]);
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'bytes'], [(0, sdk_1.toBN)(0), functionData]);
            const tx = gatewayFinalizeTransfer(l1StakingMock.address, staking.address, tokens100k, callhookData);
            await (0, chai_1.expect)(tx)
                .emit(l2GraphTokenGateway, 'DepositFinalized')
                .withArgs(l1GRTMock.address, l1StakingMock.address, staking.address, tokens100k);
            await (0, chai_1.expect)(tx).emit(staking, 'StakeDeposited').withArgs(me.address, tokens100k);
            (0, chai_1.expect)(await staking.getIndexerStakedTokens(me.address)).to.equal(tokens100k);
            const delegationPool = await staking.delegationPools(me.address);
            (0, chai_1.expect)(delegationPool.indexingRewardCut).eq((0, sdk_1.toBN)(1000000)); // 1 in PPM
            (0, chai_1.expect)(delegationPool.queryFeeCut).eq((0, sdk_1.toBN)(1000000)); // 1 in PPM
        });
        it('adds stake to an existing indexer that was already transferred', async function () {
            const functionData = utils_1.defaultAbiCoder.encode(['tuple(address)'], [[me.address]]);
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'bytes'], [(0, sdk_1.toBN)(0), functionData]);
            await gatewayFinalizeTransfer(l1StakingMock.address, staking.address, tokens100k, callhookData);
            const tx = gatewayFinalizeTransfer(l1StakingMock.address, staking.address, tokens100k, callhookData);
            await (0, chai_1.expect)(tx)
                .emit(l2GraphTokenGateway, 'DepositFinalized')
                .withArgs(l1GRTMock.address, l1StakingMock.address, staking.address, tokens100k);
            await (0, chai_1.expect)(tx).emit(staking, 'StakeDeposited').withArgs(me.address, tokens100k);
            (0, chai_1.expect)(await staking.getIndexerStakedTokens(me.address)).to.equal(tokens100k.add(tokens100k));
        });
        it('adds stake to an existing indexer that was staked in L2 (without changing delegation params)', async function () {
            const functionData = utils_1.defaultAbiCoder.encode(['tuple(address)'], [[me.address]]);
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'bytes'], [(0, sdk_1.toBN)(0), functionData]);
            await staking.connect(me).stake(tokens100k);
            await staking.connect(me).setDelegationParameters(1000, 1000, 1000);
            const tx = gatewayFinalizeTransfer(l1StakingMock.address, staking.address, tokens100k, callhookData);
            await (0, chai_1.expect)(tx)
                .emit(l2GraphTokenGateway, 'DepositFinalized')
                .withArgs(l1GRTMock.address, l1StakingMock.address, staking.address, tokens100k);
            await (0, chai_1.expect)(tx).emit(staking, 'StakeDeposited').withArgs(me.address, tokens100k);
            (0, chai_1.expect)(await staking.getIndexerStakedTokens(me.address)).to.equal(tokens100k.add(tokens100k));
            const delegationPool = await staking.delegationPools(me.address);
            (0, chai_1.expect)(delegationPool.indexingRewardCut).eq((0, sdk_1.toBN)(1000));
            (0, chai_1.expect)(delegationPool.queryFeeCut).eq((0, sdk_1.toBN)(1000));
        });
    });
    describe('receiving delegation from L1 (onTokenTransfer)', function () {
        it('adds delegation for a new delegator', async function () {
            await staking.connect(me).stake(tokens100k);
            const functionData = utils_1.defaultAbiCoder.encode(['tuple(address,address)'], [[me.address, other.address]]);
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'bytes'], [(0, sdk_1.toBN)(1), functionData]);
            const tx = gatewayFinalizeTransfer(l1StakingMock.address, staking.address, tokens10k, callhookData);
            await (0, chai_1.expect)(tx)
                .emit(l2GraphTokenGateway, 'DepositFinalized')
                .withArgs(l1GRTMock.address, l1StakingMock.address, staking.address, tokens10k);
            const expectedShares = tokens10k;
            await (0, chai_1.expect)(tx)
                .emit(staking, 'StakeDelegated')
                .withArgs(me.address, other.address, tokens10k, expectedShares);
            const delegation = await staking.getDelegation(me.address, other.address);
            (0, chai_1.expect)(delegation.shares).to.equal(expectedShares);
        });
        it('adds delegation for an existing delegator', async function () {
            await staking.connect(me).stake(tokens100k);
            await staking.connect(other).delegate(me.address, tokens10k);
            const functionData = utils_1.defaultAbiCoder.encode(['tuple(address,address)'], [[me.address, other.address]]);
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'bytes'], [(0, sdk_1.toBN)(1), functionData]);
            const tx = gatewayFinalizeTransfer(l1StakingMock.address, staking.address, tokens10k, callhookData);
            await (0, chai_1.expect)(tx)
                .emit(l2GraphTokenGateway, 'DepositFinalized')
                .withArgs(l1GRTMock.address, l1StakingMock.address, staking.address, tokens10k);
            const expectedNewShares = tokens10k;
            const expectedTotalShares = tokens10k.mul(2);
            await (0, chai_1.expect)(tx)
                .emit(staking, 'StakeDelegated')
                .withArgs(me.address, other.address, tokens10k, expectedNewShares);
            const delegation = await staking.getDelegation(me.address, other.address);
            (0, chai_1.expect)(delegation.shares).to.equal(expectedTotalShares);
        });
        it('returns delegation to the delegator if it would produce no shares', async function () {
            await fixtureContracts.RewardsManager.connect(governor).setIssuancePerBlock((0, sdk_1.toGRT)('114'));
            await staking.connect(me).stake(tokens100k);
            // Initialize the delegation pool to allow delegating less than 1 GRT
            await staking.connect(me).delegate(me.address, tokens10k);
            await staking.connect(me).setDelegationParameters(1000, 1000, 1000);
            await grt.connect(me).approve(fixtureContracts.Curation.address, tokens10k);
            await fixtureContracts.Curation.connect(me).mint(subgraphDeploymentID, tokens10k, 0);
            await allocate(tokens100k);
            await sdk_1.helpers.mineEpoch(fixtureContracts.EpochManager);
            await sdk_1.helpers.mineEpoch(fixtureContracts.EpochManager);
            await staking.connect(me).closeAllocation(allocationID, (0, sdk_1.randomHexBytes)(32));
            // Now there are some rewards sent to delegation pool, so 1 weiGRT is less than 1 share
            const functionData = utils_1.defaultAbiCoder.encode(['tuple(address,address)'], [[me.address, other.address]]);
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'bytes'], [(0, sdk_1.toBN)(1), functionData]);
            const delegatorGRTBalanceBefore = await grt.balanceOf(other.address);
            const tx = gatewayFinalizeTransfer(l1StakingMock.address, staking.address, (0, sdk_1.toBN)(1), // Less than 1 share!
            callhookData);
            await (0, chai_1.expect)(tx)
                .emit(l2GraphTokenGateway, 'DepositFinalized')
                .withArgs(l1GRTMock.address, l1StakingMock.address, staking.address, (0, sdk_1.toBN)(1));
            const delegation = await staking.getDelegation(me.address, other.address);
            await (0, chai_1.expect)(tx)
                .emit(staking, 'TransferredDelegationReturnedToDelegator')
                .withArgs(me.address, other.address, (0, sdk_1.toBN)(1));
            (0, chai_1.expect)(delegation.shares).to.equal(0);
            const delegatorGRTBalanceAfter = await grt.balanceOf(other.address);
            (0, chai_1.expect)(delegatorGRTBalanceAfter.sub(delegatorGRTBalanceBefore)).to.equal((0, sdk_1.toBN)(1));
        });
        it('returns delegation to the delegator if it initializes the pool with less than the minimum delegation', async function () {
            await fixtureContracts.RewardsManager.connect(governor).setIssuancePerBlock((0, sdk_1.toGRT)('114'));
            await staking.connect(me).stake(tokens100k);
            await staking.connect(me).setDelegationParameters(1000, 1000, 1000);
            await grt.connect(me).approve(fixtureContracts.Curation.address, tokens10k);
            await fixtureContracts.Curation.connect(me).mint(subgraphDeploymentID, tokens10k, 0);
            await allocate(tokens100k);
            await sdk_1.helpers.mineEpoch(fixtureContracts.EpochManager, 2);
            await staking.connect(me).closeAllocation(allocationID, (0, sdk_1.randomHexBytes)(32));
            // Now there are some rewards sent to delegation pool, so 1 weiGRT is less than 1 share
            const functionData = utils_1.defaultAbiCoder.encode(['tuple(address,address)'], [[me.address, other.address]]);
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'bytes'], [(0, sdk_1.toBN)(1), functionData]);
            const delegatorGRTBalanceBefore = await grt.balanceOf(other.address);
            const tx = gatewayFinalizeTransfer(l1StakingMock.address, staking.address, (0, sdk_1.toGRT)('0.1'), // Less than 1 GRT!
            callhookData);
            await (0, chai_1.expect)(tx)
                .emit(l2GraphTokenGateway, 'DepositFinalized')
                .withArgs(l1GRTMock.address, l1StakingMock.address, staking.address, (0, sdk_1.toGRT)('0.1'));
            const delegation = await staking.getDelegation(me.address, other.address);
            await (0, chai_1.expect)(tx)
                .emit(staking, 'TransferredDelegationReturnedToDelegator')
                .withArgs(me.address, other.address, (0, sdk_1.toGRT)('0.1'));
            (0, chai_1.expect)(delegation.shares).to.equal(0);
            const delegatorGRTBalanceAfter = await grt.balanceOf(other.address);
            (0, chai_1.expect)(delegatorGRTBalanceAfter.sub(delegatorGRTBalanceBefore)).to.equal((0, sdk_1.toGRT)('0.1'));
        });
        it('returns delegation under the minimum if the pool is initialized', async function () {
            await staking.connect(me).stake(tokens100k);
            // Initialize the delegation pool to allow delegating less than 1 GRT
            await staking.connect(me).delegate(me.address, tokens10k);
            const functionData = utils_1.defaultAbiCoder.encode(['tuple(address,address)'], [[me.address, other.address]]);
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'bytes'], [(0, sdk_1.toBN)(1), functionData]);
            const delegatorGRTBalanceBefore = await grt.balanceOf(other.address);
            const tx = gatewayFinalizeTransfer(l1StakingMock.address, staking.address, (0, sdk_1.toGRT)('0.1'), callhookData);
            await (0, chai_1.expect)(tx)
                .emit(l2GraphTokenGateway, 'DepositFinalized')
                .withArgs(l1GRTMock.address, l1StakingMock.address, staking.address, (0, sdk_1.toGRT)('0.1'));
            const delegation = await staking.getDelegation(me.address, other.address);
            await (0, chai_1.expect)(tx)
                .emit(staking, 'TransferredDelegationReturnedToDelegator')
                .withArgs(me.address, other.address, (0, sdk_1.toGRT)('0.1'));
            (0, chai_1.expect)(delegation.shares).to.equal(0);
            const delegatorGRTBalanceAfter = await grt.balanceOf(other.address);
            (0, chai_1.expect)(delegatorGRTBalanceAfter.sub(delegatorGRTBalanceBefore)).to.equal((0, sdk_1.toGRT)('0.1'));
        });
    });
    describe('onTokenTransfer with invalid messages', function () {
        it('reverts if the code is invalid', async function () {
            // This should never really happen unless the Arbitrum bridge is compromised,
            // so we test it anyway to ensure it's a well-defined behavior.
            // code 2 does not exist:
            const callhookData = utils_1.defaultAbiCoder.encode(['uint8', 'bytes'], [(0, sdk_1.toBN)(2), '0x12345678']);
            const tx = gatewayFinalizeTransfer(l1StakingMock.address, staking.address, (0, sdk_1.toGRT)('1'), callhookData);
            await (0, chai_1.expect)(tx).revertedWith('INVALID_CODE');
        });
        it('reverts if the message encoding is invalid', async function () {
            // This should never really happen unless the Arbitrum bridge is compromised,
            // so we test it anyway to ensure it's a well-defined behavior.
            const callhookData = utils_1.defaultAbiCoder.encode(['address', 'uint128'], [AddressZero, (0, sdk_1.toBN)(2)]);
            const tx = gatewayFinalizeTransfer(l1StakingMock.address, staking.address, (0, sdk_1.toGRT)('1'), callhookData);
            await (0, chai_1.expect)(tx).reverted; // abi.decode will fail with no reason
        });
    });
});
