"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.deprecateSubgraph = exports.burnSignal = exports.mintSignal = exports.publishNewVersion = exports.publishNewSubgraph = exports.getTokensAndVSignal = exports.createDefaultName = exports.DEFAULT_RESERVE_RATIO = void 0;
const utils_1 = require("ethers/lib/utils");
const chai_1 = require("chai");
const sdk_1 = require("@graphprotocol/sdk");
exports.DEFAULT_RESERVE_RATIO = 1000000;
const createDefaultName = (name) => {
    return {
        name: name,
        nameIdentifier: (0, utils_1.namehash)(name),
    };
};
exports.createDefaultName = createDefaultName;
const getTokensAndVSignal = async (subgraphDeploymentID, curation) => {
    const curationPool = await curation.pools(subgraphDeploymentID);
    const vSignal = await curation.getCurationPoolSignal(subgraphDeploymentID);
    return [curationPool.tokens, vSignal];
};
exports.getTokensAndVSignal = getTokensAndVSignal;
const publishNewSubgraph = async (account, newSubgraph, gns, chainId) => {
    const subgraphID = await (0, sdk_1.buildSubgraphId)(account.address, await gns.nextAccountSeqID(account.address), chainId);
    // Send tx
    const tx = gns
        .connect(account)
        .publishNewSubgraph(newSubgraph.subgraphDeploymentID, newSubgraph.versionMetadata, newSubgraph.subgraphMetadata);
    // Check events
    await (0, chai_1.expect)(tx)
        .emit(gns, 'SubgraphPublished')
        .withArgs(subgraphID, newSubgraph.subgraphDeploymentID, exports.DEFAULT_RESERVE_RATIO)
        .emit(gns, 'SubgraphMetadataUpdated')
        .withArgs(subgraphID, newSubgraph.subgraphMetadata)
        .emit(gns, 'SubgraphVersionUpdated')
        .withArgs(subgraphID, newSubgraph.subgraphDeploymentID, newSubgraph.versionMetadata);
    // Check state
    const subgraph = await gns.subgraphs(subgraphID);
    (0, chai_1.expect)(subgraph.vSignal).eq(0);
    (0, chai_1.expect)(subgraph.nSignal).eq(0);
    (0, chai_1.expect)(subgraph.subgraphDeploymentID).eq(newSubgraph.subgraphDeploymentID);
    (0, chai_1.expect)(subgraph.__DEPRECATED_reserveRatio).eq(exports.DEFAULT_RESERVE_RATIO);
    (0, chai_1.expect)(subgraph.disabled).eq(false);
    (0, chai_1.expect)(subgraph.withdrawableGRT).eq(0);
    // Check NFT issuance
    const owner = await gns.ownerOf(subgraphID);
    (0, chai_1.expect)(owner).eq(account.address);
    // TODO: properly type this
    return { ...subgraph, id: subgraphID };
};
exports.publishNewSubgraph = publishNewSubgraph;
const publishNewVersion = async (account, subgraphID, newSubgraph, gns, curation) => {
    // Before state
    const ownerTaxPercentage = await gns.ownerTaxPercentage();
    const curationTaxPercentage = await curation.curationTaxPercentage();
    const beforeSubgraph = await gns.subgraphs(subgraphID);
    // Check what selling all nSignal, which == selling all vSignal, should return for tokens
    // NOTE - no tax on burning on nSignal
    const tokensReceivedEstimate = beforeSubgraph.nSignal.gt(0)
        ? (await gns.nSignalToTokens(subgraphID, beforeSubgraph.nSignal))[1]
        : (0, sdk_1.toBN)(0);
    // Example:
    // Deposit 100, 5 is taxed, 95 GRT in curve
    // Upgrade - calculate 5% tax on 95 --> 4.75 GRT
    // Multiple by ownerPercentage --> 50% * 4.75 = 2.375 GRT
    // Owner adds 2.375 to 90.25, we deposit 92.625 GRT into the curve
    // Divide this by 0.95 to get exactly 97.5 total tokens to be deposited
    // nSignalToTokens returns the amount of tokens with tax removed
    // already. So we must add in the tokens removed
    const MAX_PPM = 1000000;
    const taxOnOriginal = tokensReceivedEstimate.mul(curationTaxPercentage).div(MAX_PPM);
    const totalWithoutOwnerTax = tokensReceivedEstimate.sub(taxOnOriginal);
    const ownerTax = taxOnOriginal.mul(ownerTaxPercentage).div(MAX_PPM);
    const totalWithOwnerTax = totalWithoutOwnerTax.add(ownerTax);
    const totalAdjustedUp = totalWithOwnerTax.mul(MAX_PPM).div(MAX_PPM - curationTaxPercentage);
    // Re-estimate amount of signal to get considering the owner tax paid by the owner
    const { 0: newVSignalEstimate, 1: newCurationTaxEstimate } = beforeSubgraph.nSignal.gt(0)
        ? await curation.tokensToSignal(newSubgraph.subgraphDeploymentID, totalAdjustedUp)
        : [(0, sdk_1.toBN)(0), (0, sdk_1.toBN)(0)];
    // Check the vSignal of the new curation curve, and tokens, before upgrading
    const [beforeTokensNewCurve, beforeVSignalNewCurve] = await (0, exports.getTokensAndVSignal)(newSubgraph.subgraphDeploymentID, curation);
    // Send tx
    const tx = gns
        .connect(account)
        .publishNewVersion(subgraphID, newSubgraph.subgraphDeploymentID, newSubgraph.versionMetadata);
    const txResult = (0, chai_1.expect)(tx)
        .emit(gns, 'SubgraphVersionUpdated')
        .withArgs(subgraphID, newSubgraph.subgraphDeploymentID, newSubgraph.versionMetadata);
    // Only emits this event if there was actual signal to upgrade
    if (beforeSubgraph.nSignal.gt(0)) {
        await txResult
            .emit(gns, 'SubgraphUpgraded')
            .withArgs(subgraphID, newVSignalEstimate, totalAdjustedUp, newSubgraph.subgraphDeploymentID);
    }
    await txResult;
    // Check curation vSignal old are set to zero
    const [afterTokensOldCuration, afterVSignalOldCuration] = await (0, exports.getTokensAndVSignal)(beforeSubgraph.subgraphDeploymentID, curation);
    (0, chai_1.expect)(afterTokensOldCuration).eq(0);
    (0, chai_1.expect)(afterVSignalOldCuration).eq(0);
    // Check the vSignal of the new curation curve, and tokens
    const [afterTokensNewCurve, afterVSignalNewCurve] = await (0, exports.getTokensAndVSignal)(newSubgraph.subgraphDeploymentID, curation);
    (0, chai_1.expect)(afterTokensNewCurve).eq(beforeTokensNewCurve.add(totalAdjustedUp).sub(newCurationTaxEstimate));
    (0, chai_1.expect)(afterVSignalNewCurve).eq(beforeVSignalNewCurve.add(newVSignalEstimate));
    // Check the nSignal pool
    const afterSubgraph = await gns.subgraphs(subgraphID);
    (0, chai_1.expect)(afterSubgraph.vSignal)
        .eq(afterVSignalNewCurve.sub(beforeVSignalNewCurve))
        .eq(newVSignalEstimate);
    (0, chai_1.expect)(afterSubgraph.nSignal).eq(beforeSubgraph.nSignal); // should not change
    (0, chai_1.expect)(afterSubgraph.subgraphDeploymentID).eq(newSubgraph.subgraphDeploymentID);
    // Check NFT should not change owner
    const owner = await gns.ownerOf(subgraphID);
    (0, chai_1.expect)(owner).eq(account.address);
    return tx;
};
exports.publishNewVersion = publishNewVersion;
const mintSignal = async (account, subgraphID, tokensIn, gns, curation) => {
    // Before state
    const beforeSubgraph = await gns.subgraphs(subgraphID);
    const [beforeTokens, beforeVSignal] = await (0, exports.getTokensAndVSignal)(beforeSubgraph.subgraphDeploymentID, curation);
    // Deposit
    const { 0: vSignalExpected, 1: nSignalExpected, 2: curationTax, } = await gns.tokensToNSignal(subgraphID, tokensIn);
    const tx = gns.connect(account).mintSignal(subgraphID, tokensIn, 0);
    await (0, chai_1.expect)(tx)
        .emit(gns, 'SignalMinted')
        .withArgs(subgraphID, account.address, nSignalExpected, vSignalExpected, tokensIn);
    // After state
    const afterSubgraph = await gns.subgraphs(subgraphID);
    const [afterTokens, afterVSignal] = await (0, exports.getTokensAndVSignal)(afterSubgraph.subgraphDeploymentID, curation);
    // Check state
    (0, chai_1.expect)(afterTokens).eq(beforeTokens.add(tokensIn.sub(curationTax)));
    (0, chai_1.expect)(afterVSignal).eq(beforeVSignal.add(vSignalExpected));
    (0, chai_1.expect)(afterSubgraph.nSignal).eq(beforeSubgraph.nSignal.add(nSignalExpected));
    (0, chai_1.expect)(afterSubgraph.vSignal).eq(beforeVSignal.add(vSignalExpected));
    return tx;
};
exports.mintSignal = mintSignal;
const burnSignal = async (account, subgraphID, gns, curation) => {
    // Before state
    const beforeSubgraph = await gns.subgraphs(subgraphID);
    const [beforeTokens, beforeVSignal] = await (0, exports.getTokensAndVSignal)(beforeSubgraph.subgraphDeploymentID, curation);
    const beforeUsersNSignal = await gns.getCuratorSignal(subgraphID, account.address);
    // Withdraw
    const { 0: vSignalExpected, 1: tokensExpected } = await gns.nSignalToTokens(subgraphID, beforeUsersNSignal);
    // Send tx
    const tx = gns.connect(account).burnSignal(subgraphID, beforeUsersNSignal, 0);
    await (0, chai_1.expect)(tx)
        .emit(gns, 'SignalBurned')
        .withArgs(subgraphID, account.address, beforeUsersNSignal, vSignalExpected, tokensExpected);
    // After state
    const afterSubgraph = await gns.subgraphs(subgraphID);
    const [afterTokens, afterVSignalCuration] = await (0, exports.getTokensAndVSignal)(afterSubgraph.subgraphDeploymentID, curation);
    // Check state
    (0, chai_1.expect)(afterTokens).eq(beforeTokens.sub(tokensExpected));
    (0, chai_1.expect)(afterVSignalCuration).eq(beforeVSignal.sub(vSignalExpected));
    (0, chai_1.expect)(afterSubgraph.nSignal).eq(beforeSubgraph.nSignal.sub(beforeUsersNSignal));
    return tx;
};
exports.burnSignal = burnSignal;
const deprecateSubgraph = async (account, subgraphID, gns, curation, grt) => {
    // Before state
    const beforeSubgraph = await gns.subgraphs(subgraphID);
    const [beforeTokens] = await (0, exports.getTokensAndVSignal)(beforeSubgraph.subgraphDeploymentID, curation);
    // We can use the whole amount, since in this test suite all vSignal is used to be staked on nSignal
    const ownerBalanceBefore = await grt.balanceOf(account.address);
    // Send tx
    const tx = gns.connect(account).deprecateSubgraph(subgraphID);
    await (0, chai_1.expect)(tx).emit(gns, 'SubgraphDeprecated').withArgs(subgraphID, beforeTokens);
    // After state
    const afterSubgraph = await gns.subgraphs(subgraphID);
    // Check marked as deprecated
    (0, chai_1.expect)(afterSubgraph.disabled).eq(true);
    // Signal for the deployment must be all burned
    (0, chai_1.expect)(afterSubgraph.vSignal.eq((0, sdk_1.toBN)('0')));
    // Cleanup reserve ratio
    (0, chai_1.expect)(afterSubgraph.__DEPRECATED_reserveRatio).eq(0);
    // Should be equal since owner pays curation tax
    (0, chai_1.expect)(afterSubgraph.withdrawableGRT).eq(beforeTokens);
    // Check balance of GNS increased by curation tax from owner being added
    const afterGNSBalance = await grt.balanceOf(gns.address);
    (0, chai_1.expect)(afterGNSBalance).eq(afterSubgraph.withdrawableGRT);
    // Check that the owner balance decreased by the curation tax
    const ownerBalanceAfter = await grt.balanceOf(account.address);
    (0, chai_1.expect)(ownerBalanceBefore.eq(ownerBalanceAfter));
    // Check NFT was burned
    await (0, chai_1.expect)(gns.ownerOf(subgraphID)).revertedWith('ERC721: owner query for nonexistent token');
    return tx;
};
exports.deprecateSubgraph = deprecateSubgraph;
