"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.grtTests = grtTests;
const hardhat_1 = __importDefault(require("hardhat"));
const chai_1 = require("chai");
const ethers_1 = require("ethers");
const sdk_1 = require("@graphprotocol/sdk");
const fixtures_1 = require("./fixtures");
const { AddressZero, MaxUint256 } = ethers_1.constants;
const L1SALT = '0x51f3d585afe6dfeb2af01bba0889a36c1db03beec88c6a4d0c53817069026afa';
const L2SALT = '0xe33842a7acd1d5a1d28f25a931703e5605152dc48d64dc4716efdae1f5659591';
function grtTests(isL2) {
    let me;
    let other;
    let governor;
    let salt;
    let fixture;
    let fixtureContracts;
    const graph = hardhat_1.default.graph();
    const mePrivateKey = '0x4f3edf983ac636a65a842ce7c78d9aa706d3b113bce9c46f30d7d21715b23b1d';
    const otherPrivateKey = '0x6cbed15c793ce57650b9877cf6fa156fbef513c4e6134f022a85b1ffdd59b2a1';
    let grt;
    async function permitMaxOK() {
        return permitOK(MaxUint256);
    }
    async function permitOK(value) {
        const nonce = await grt.nonces(me.address);
        return {
            owner: me.address,
            spender: other.address,
            value: value,
            nonce: nonce,
            deadline: (0, sdk_1.toBN)('0'),
        };
    }
    async function permitExpired() {
        const permit = await permitMaxOK();
        permit.deadline = (0, sdk_1.toBN)('1');
        return permit;
    }
    async function permitDeny() {
        const permit = await permitMaxOK();
        permit.value = (0, sdk_1.toBN)('0');
        return permit;
    }
    function createPermitTransaction(permit, signer, salt) {
        const signature = (0, sdk_1.signPermit)(signer, graph.chainId, grt.address, permit, salt);
        const wallet = new ethers_1.ethers.Wallet(signer, graph.provider);
        return grt
            .connect(wallet)
            .permit(permit.owner, permit.spender, permit.value, permit.deadline, signature.v, signature.r, signature.s);
    }
    before(async function () {
        ({ governor } = await graph.getNamedAccounts());
        me = new ethers_1.ethers.Wallet(mePrivateKey, graph.provider);
        other = new ethers_1.ethers.Wallet(otherPrivateKey, graph.provider);
        fixture = new fixtures_1.NetworkFixture(graph.provider);
        fixtureContracts = await fixture.load(governor, isL2);
        grt = fixtureContracts.GraphToken;
        salt = isL2 ? L2SALT : L1SALT;
        // Mint some tokens
        const tokens = (0, sdk_1.toGRT)('10000');
        await grt.connect(governor).mint(me.address, tokens);
    });
    describe('permit', function () {
        it('should permit max token allowance', async function () {
            // Allow to transfer tokens
            const tokensToApprove = (0, sdk_1.toGRT)('1000');
            const permit = await permitOK(tokensToApprove);
            const tx = createPermitTransaction(permit, mePrivateKey, salt);
            await (0, chai_1.expect)(tx).emit(grt, 'Approval').withArgs(permit.owner, permit.spender, tokensToApprove);
            // Allowance updated
            const allowance = await grt.allowance(me.address, other.address);
            (0, chai_1.expect)(allowance).eq(tokensToApprove);
            // Transfer tokens should work
            const tokens = (0, sdk_1.toGRT)('100');
            await grt.connect(other).transferFrom(me.address, other.address, tokens);
        });
        it('should permit max token allowance', async function () {
            // Allow to transfer tokens
            const permit = await permitMaxOK();
            const tx = createPermitTransaction(permit, mePrivateKey, salt);
            await (0, chai_1.expect)(tx).emit(grt, 'Approval').withArgs(permit.owner, permit.spender, MaxUint256);
            // Allowance updated
            const allowance = await grt.allowance(me.address, other.address);
            (0, chai_1.expect)(allowance).eq(MaxUint256);
            // Transfer tokens should work
            const tokens = (0, sdk_1.toGRT)('100');
            await grt.connect(other).transferFrom(me.address, other.address, tokens);
        });
        it('reject to transfer more tokens than approved by permit', async function () {
            // Allow to transfer tokens
            const tokensToApprove = (0, sdk_1.toGRT)('1000');
            const permit = await permitOK(tokensToApprove);
            await createPermitTransaction(permit, mePrivateKey, salt);
            // Should not transfer more than approved
            const tooManyTokens = (0, sdk_1.toGRT)('1001');
            const tx = grt.connect(other).transferFrom(me.address, other.address, tooManyTokens);
            await (0, chai_1.expect)(tx).revertedWith('ERC20: transfer amount exceeds allowance');
            // Should transfer up to the approved amount
            await grt.connect(other).transferFrom(me.address, other.address, tokensToApprove);
        });
        it('reject use two permits with same nonce', async function () {
            // Allow to transfer tokens
            const permit = await permitMaxOK();
            await createPermitTransaction(permit, mePrivateKey, salt);
            // Try to re-use the permit
            const tx = createPermitTransaction(permit, mePrivateKey, salt);
            await (0, chai_1.expect)(tx).revertedWith('GRT: invalid permit');
        });
        it('reject use expired permit', async function () {
            const permit = await permitExpired();
            const tx = createPermitTransaction(permit, mePrivateKey, salt);
            await (0, chai_1.expect)(tx).revertedWith('GRT: expired permit');
        });
        it('reject permit if holder address does not match', async function () {
            const permit = await permitMaxOK();
            const tx = createPermitTransaction(permit, otherPrivateKey, salt);
            await (0, chai_1.expect)(tx).revertedWith('GRT: invalid permit');
        });
        it('should deny transfer from if permit was denied', async function () {
            // Allow to transfer tokens
            const permit1 = await permitMaxOK();
            await createPermitTransaction(permit1, mePrivateKey, salt);
            // Deny transfer tokens
            const permit2 = await permitDeny();
            await createPermitTransaction(permit2, mePrivateKey, salt);
            // Allowance updated
            const allowance = await grt.allowance(me.address, other.address);
            (0, chai_1.expect)(allowance).eq((0, sdk_1.toBN)('0'));
            // Try to transfer without permit should fail
            const tokens = (0, sdk_1.toGRT)('100');
            const tx = grt.connect(other).transferFrom(me.address, other.address, tokens);
            await (0, chai_1.expect)(tx).revertedWith('ERC20: transfer amount exceeds allowance');
        });
    });
    describe('mint', function () {
        describe('mint', function () {
            it('reject mint if not minter', async function () {
                const tx = grt.connect(me).mint(me.address, (0, sdk_1.toGRT)('100'));
                await (0, chai_1.expect)(tx).revertedWith('Only minter can call');
            });
        });
        describe('addMinter', function () {
            it('reject add a new minter if not allowed', async function () {
                const tx = grt.connect(me).addMinter(me.address);
                await (0, chai_1.expect)(tx).revertedWith('Only Governor can call');
            });
            it('should add a new minter', async function () {
                (0, chai_1.expect)(await grt.isMinter(me.address)).eq(false);
                const tx = grt.connect(governor).addMinter(me.address);
                await (0, chai_1.expect)(tx).emit(grt, 'MinterAdded').withArgs(me.address);
                (0, chai_1.expect)(await grt.isMinter(me.address)).eq(true);
            });
        });
        context('> when is minter', function () {
            beforeEach(async function () {
                await grt.connect(governor).addMinter(me.address);
                (0, chai_1.expect)(await grt.isMinter(me.address)).eq(true);
            });
            describe('mint', function () {
                it('should mint', async function () {
                    const beforeTokens = await grt.balanceOf(me.address);
                    const tokensToMint = (0, sdk_1.toGRT)('100');
                    const tx = grt.connect(me).mint(me.address, tokensToMint);
                    await (0, chai_1.expect)(tx).emit(grt, 'Transfer').withArgs(AddressZero, me.address, tokensToMint);
                    const afterTokens = await grt.balanceOf(me.address);
                    (0, chai_1.expect)(afterTokens).eq(beforeTokens.add(tokensToMint));
                });
                it('should mint if governor', async function () {
                    const tokensToMint = (0, sdk_1.toGRT)('100');
                    await grt.connect(governor).mint(me.address, tokensToMint);
                });
            });
            describe('removeMinter', function () {
                it('reject remove a minter if not allowed', async function () {
                    const tx = grt.connect(me).removeMinter(me.address);
                    await (0, chai_1.expect)(tx).revertedWith('Only Governor can call');
                });
                it('should remove a minter', async function () {
                    const tx = grt.connect(governor).removeMinter(me.address);
                    await (0, chai_1.expect)(tx).emit(grt, 'MinterRemoved').withArgs(me.address);
                    (0, chai_1.expect)(await grt.isMinter(me.address)).eq(false);
                });
            });
            describe('renounceMinter', function () {
                it('should renounce to be a minter', async function () {
                    const tx = grt.connect(me).renounceMinter();
                    await (0, chai_1.expect)(tx).emit(grt, 'MinterRemoved').withArgs(me.address);
                    (0, chai_1.expect)(await grt.isMinter(me.address)).eq(false);
                });
            });
        });
    });
}
