"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const hardhat_1 = __importDefault(require("hardhat"));
const chai_1 = require("chai");
const ethers_1 = require("ethers");
const fixtures_1 = require("../lib/fixtures");
const utils_1 = require("ethers/lib/utils");
const sdk_1 = require("@graphprotocol/sdk");
const { AddressZero, MaxUint256 } = ethers_1.constants;
describe('AllocationExchange', () => {
    let governor;
    let allocationExchangeOwner;
    let indexer;
    let authority;
    let fixture;
    let contracts;
    let grt;
    let staking;
    let allocationExchange;
    let epochManager;
    const graph = hardhat_1.default.graph();
    function createVoucher(allocationID, amount, signerKey) {
        const messageHash = (0, utils_1.solidityKeccak256)(['address', 'uint256'], [allocationID, amount]);
        const messageHashBytes = (0, utils_1.arrayify)(messageHash);
        const key = new utils_1.SigningKey(signerKey);
        const signature = key.signDigest(messageHashBytes);
        return {
            allocationID,
            amount,
            signature: (0, utils_1.joinSignature)(signature),
        };
    }
    before(async function () {
        [indexer] = await graph.getTestAccounts();
        ({ governor, allocationExchangeOwner } = await graph.getNamedAccounts());
        authority = ethers_1.Wallet.createRandom();
        fixture = new fixtures_1.NetworkFixture(graph.provider);
        contracts = await fixture.load(governor);
        allocationExchange = contracts.AllocationExchange;
        grt = contracts.GraphToken;
        staking = contracts.Staking;
        epochManager = contracts.EpochManager;
        // Give some funds to the indexer and approve staking contract to use funds on indexer behalf
        const indexerTokens = (0, sdk_1.toGRT)('100000');
        await grt.connect(governor).mint(indexer.address, indexerTokens);
        await grt.connect(indexer).approve(staking.address, indexerTokens);
        // Give some funds to the AllocationExchange
        const exchangeTokens = (0, sdk_1.toGRT)('2000');
        await grt.connect(governor).mint(allocationExchange.address, exchangeTokens);
        // Ensure the exchange is correctly setup
        await allocationExchange.connect(allocationExchangeOwner).setAuthority(authority.address, true);
        await allocationExchange.connect(allocationExchangeOwner).approveAll();
    });
    beforeEach(async function () {
        await fixture.setUp();
    });
    afterEach(async function () {
        await fixture.tearDown();
    });
    async function setupAllocation() {
        // Generate test data
        const channelKey = (0, sdk_1.deriveChannelKey)();
        const allocationID = channelKey.address;
        const subgraphDeploymentID = (0, sdk_1.randomHexBytes)(32);
        const metadata = (0, sdk_1.randomHexBytes)(32);
        // Setup staking
        const stakeTokens = (0, sdk_1.toGRT)('100000');
        await staking.connect(indexer).stake(stakeTokens);
        await staking
            .connect(indexer)
            .allocateFrom(indexer.address, subgraphDeploymentID, stakeTokens, allocationID, metadata, await channelKey.generateProof(indexer.address));
        return allocationID;
    }
    describe('config', function () {
        it('should set an authority', async function () {
            // Set authority
            const newAuthority = (0, sdk_1.randomAddress)();
            const tx1 = allocationExchange
                .connect(allocationExchangeOwner)
                .setAuthority(newAuthority, true);
            await (0, chai_1.expect)(tx1).emit(allocationExchange, 'AuthoritySet').withArgs(newAuthority, true);
            (0, chai_1.expect)(await allocationExchange.authority(newAuthority)).eq(true);
            // Unset authority
            const tx2 = allocationExchange
                .connect(allocationExchangeOwner)
                .setAuthority(newAuthority, false);
            await (0, chai_1.expect)(tx2).emit(allocationExchange, 'AuthoritySet').withArgs(newAuthority, false);
            (0, chai_1.expect)(await allocationExchange.authority(newAuthority)).eq(false);
        });
        it('reject set an authority if not allowed', async function () {
            const newAuthority = (0, sdk_1.randomAddress)();
            const tx = allocationExchange.connect(indexer).setAuthority(newAuthority, true);
            await (0, chai_1.expect)(tx).revertedWith('Only Governor can call');
        });
        it('reject set an empty authority', async function () {
            const newAuthority = AddressZero;
            const tx = allocationExchange
                .connect(allocationExchangeOwner)
                .setAuthority(newAuthority, true);
            await (0, chai_1.expect)(tx).revertedWith('Exchange: empty authority');
        });
        it('should allow to approve all tokens to staking contract', async function () {
            await allocationExchange.connect(allocationExchangeOwner).approveAll();
            const allowance = await grt.allowance(allocationExchange.address, staking.address);
            (0, chai_1.expect)(allowance).eq(MaxUint256);
        });
    });
    describe('withdraw funds', function () {
        it('should withdraw to destination', async function () {
            const beforeExchangeBalance = await grt.balanceOf(allocationExchange.address);
            const destinationAddress = (0, sdk_1.randomAddress)();
            const amount = (0, sdk_1.toGRT)('1000');
            const tx = allocationExchange
                .connect(allocationExchangeOwner)
                .withdraw(destinationAddress, amount);
            await (0, chai_1.expect)(tx)
                .emit(allocationExchange, 'TokensWithdrawn')
                .withArgs(destinationAddress, amount);
            const afterExchangeBalance = await grt.balanceOf(allocationExchange.address);
            const afterDestinationBalance = await grt.balanceOf(destinationAddress);
            (0, chai_1.expect)(afterExchangeBalance).eq(beforeExchangeBalance.sub(amount));
            (0, chai_1.expect)(afterDestinationBalance).eq(amount);
        });
        it('reject withdraw zero amount', async function () {
            const destinationAddress = (0, sdk_1.randomAddress)();
            const amount = (0, sdk_1.toGRT)('0');
            const tx = allocationExchange
                .connect(allocationExchangeOwner)
                .withdraw(destinationAddress, amount);
            await (0, chai_1.expect)(tx).revertedWith('Exchange: empty amount');
        });
        it('reject withdraw to zero destination', async function () {
            const destinationAddress = AddressZero;
            const amount = (0, sdk_1.toGRT)('1000');
            const tx = allocationExchange
                .connect(allocationExchangeOwner)
                .withdraw(destinationAddress, amount);
            await (0, chai_1.expect)(tx).revertedWith('Exchange: empty destination');
        });
        it('reject withdraw if not allowed', async function () {
            const destinationAddress = (0, sdk_1.randomAddress)();
            const amount = (0, sdk_1.toGRT)('1000');
            const tx = allocationExchange.connect(indexer.address).withdraw(destinationAddress, amount);
            await (0, chai_1.expect)(tx).revertedWith('Only Governor can call');
        });
    });
    describe('redeem vouchers', function () {
        it('redeem a voucher', async function () {
            const beforeExchangeBalance = await grt.balanceOf(allocationExchange.address);
            // Setup an active allocation
            const allocationID = await setupAllocation();
            await sdk_1.helpers.mineEpoch(epochManager);
            // Initiate a withdrawal
            const actualAmount = (0, sdk_1.toGRT)('2000'); // <- withdraw amount
            const voucher = createVoucher(allocationID, actualAmount, authority.privateKey);
            const tx = allocationExchange.connect(allocationExchangeOwner).redeem(voucher);
            await (0, chai_1.expect)(tx)
                .emit(allocationExchange, 'AllocationRedeemed')
                .withArgs(allocationID, actualAmount);
            // Allocation must have collected the withdrawn tokens
            const allocation = await staking.allocations(allocationID);
            (0, chai_1.expect)(allocation.collectedFees).eq(actualAmount);
            // AllocationExchange should have less funds
            const afterExchangeBalance = await grt.balanceOf(allocationExchange.address);
            (0, chai_1.expect)(afterExchangeBalance).eq(beforeExchangeBalance.sub(actualAmount));
        });
        it('reject double spending of a voucher', async function () {
            // Setup an active allocation
            const allocationID = await setupAllocation();
            await sdk_1.helpers.mineEpoch(epochManager);
            // Initiate a withdrawal
            const actualAmount = (0, sdk_1.toGRT)('2000'); // <- withdraw amount
            const voucher = createVoucher(allocationID, actualAmount, authority.privateKey);
            // First redeem
            await allocationExchange.connect(allocationExchangeOwner).redeem(voucher);
            // Double spend the same voucher!
            await (0, chai_1.expect)(allocationExchange.connect(allocationExchangeOwner).redeem(voucher)).revertedWith('Exchange: allocation already redeemed');
        });
        it('reject redeem voucher for invalid allocation', async function () {
            // Use an invalid allocation
            const allocationID = '0xfefefefefefefefefefefefefefefefefefefefe';
            // Ensure the exchange is correctly setup
            await allocationExchange
                .connect(allocationExchangeOwner)
                .setAuthority(authority.address, true);
            await allocationExchange.connect(allocationExchangeOwner).approveAll();
            // Initiate a withdrawal
            const actualAmount = (0, sdk_1.toGRT)('2000'); // <- withdraw amount
            const voucher = createVoucher(allocationID, actualAmount, authority.privateKey);
            const tx = allocationExchange.connect(allocationExchangeOwner).redeem(voucher);
            await (0, chai_1.expect)(tx).revertedWith('!collect');
        });
        it('reject redeem voucher not signed by the authority', async function () {
            // Initiate a withdrawal
            const actualAmount = (0, sdk_1.toGRT)('2000'); // <- withdraw amount
            const voucher = createVoucher((0, sdk_1.randomAddress)(), actualAmount, ethers_1.Wallet.createRandom().privateKey);
            const tx = allocationExchange.connect(allocationExchangeOwner).redeem(voucher);
            await (0, chai_1.expect)(tx).revertedWith('Exchange: invalid signer');
        });
        it('reject redeem voucher with empty amount', async function () {
            // Initiate a withdrawal
            const actualAmount = (0, sdk_1.toGRT)('0'); // <- withdraw amount
            const voucher = createVoucher((0, sdk_1.randomAddress)(), actualAmount, authority.privateKey);
            const tx = allocationExchange.connect(allocationExchangeOwner).redeem(voucher);
            await (0, chai_1.expect)(tx).revertedWith('Exchange: zero tokens voucher');
        });
        it('reject redeem voucher with invalid signature', async function () {
            const actualAmount = (0, sdk_1.toGRT)('2000'); // <- withdraw amount
            const voucher = createVoucher((0, sdk_1.randomAddress)(), actualAmount, authority.privateKey);
            voucher.signature = '0x1234';
            const tx = allocationExchange.connect(allocationExchangeOwner).redeem(voucher);
            await (0, chai_1.expect)(tx).revertedWith('Exchange: invalid signature');
        });
    });
});
