"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const hardhat_1 = __importDefault(require("hardhat"));
const chai_1 = require("chai");
const ethers_1 = require("ethers");
const bignumber_js_1 = require("bignumber.js");
const fixtures_1 = require("../lib/fixtures");
const hardhat_network_helpers_1 = require("@nomicfoundation/hardhat-network-helpers");
const sdk_1 = require("@graphprotocol/sdk");
const MAX_PPM = 1000000;
const { HashZero, WeiPerEther } = ethers_1.constants;
const toRound = (n) => (0, sdk_1.formatGRT)(n.add((0, sdk_1.toGRT)('0.5'))).split('.')[0];
describe('Rewards', () => {
    const graph = hardhat_1.default.graph();
    let delegator;
    let governor;
    let curator1;
    let curator2;
    let indexer1;
    let indexer2;
    let oracle;
    let assetHolder;
    let fixture;
    let contracts;
    let grt;
    let curation;
    let epochManager;
    let staking;
    let rewardsManager;
    // Derive some channel keys for each indexer used to sign attestations
    const channelKey1 = (0, sdk_1.deriveChannelKey)();
    const channelKey2 = (0, sdk_1.deriveChannelKey)();
    const channelKeyNull = (0, sdk_1.deriveChannelKey)();
    const subgraphDeploymentID1 = (0, sdk_1.randomHexBytes)();
    const subgraphDeploymentID2 = (0, sdk_1.randomHexBytes)();
    const allocationID1 = channelKey1.address;
    const allocationID2 = channelKey2.address;
    const allocationIDNull = channelKeyNull.address;
    const metadata = HashZero;
    const ISSUANCE_RATE_PERIODS = 4; // blocks required to issue 800 GRT rewards
    const ISSUANCE_PER_BLOCK = (0, sdk_1.toBN)('200000000000000000000'); // 200 GRT every block
    // Core formula that gets accumulated rewards per signal for a period of time
    const getRewardsPerSignal = (k, t, s) => {
        if (s.eq(0)) {
            return '0';
        }
        return k.times(t).div(s).toPrecision(18).toString();
    };
    // Tracks the accumulated rewards as totalSignalled or supply changes across snapshots
    class RewardsTracker {
        constructor() {
            this.totalSignalled = ethers_1.BigNumber.from(0);
            this.lastUpdatedBlock = 0;
            this.accumulated = ethers_1.BigNumber.from(0);
        }
        static async create() {
            const tracker = new RewardsTracker();
            await tracker.snapshot();
            return tracker;
        }
        async snapshot() {
            this.accumulated = this.accumulated.add(await this.accrued());
            this.totalSignalled = await grt.balanceOf(curation.address);
            this.lastUpdatedBlock = await sdk_1.helpers.latestBlock();
            return this;
        }
        async elapsedBlocks() {
            const currentBlock = await sdk_1.helpers.latestBlock();
            return currentBlock - this.lastUpdatedBlock;
        }
        async accrued() {
            const nBlocks = await this.elapsedBlocks();
            return this.accruedByElapsed(nBlocks);
        }
        accruedByElapsed(nBlocks) {
            const n = getRewardsPerSignal(new bignumber_js_1.BigNumber(ISSUANCE_PER_BLOCK.toString()), new bignumber_js_1.BigNumber(nBlocks.toString()), new bignumber_js_1.BigNumber(this.totalSignalled.toString()));
            return (0, sdk_1.toGRT)(n);
        }
    }
    // Test accumulated rewards per signal
    const shouldGetNewRewardsPerSignal = async (nBlocks = ISSUANCE_RATE_PERIODS) => {
        // -- t0 --
        const tracker = await RewardsTracker.create();
        // Jump
        await sdk_1.helpers.mine(nBlocks);
        // -- t1 --
        // Contract calculation
        const contractAccrued = await rewardsManager.getNewRewardsPerSignal();
        // Local calculation
        const expectedAccrued = await tracker.accrued();
        // Check
        (0, chai_1.expect)(toRound(expectedAccrued)).eq(toRound(contractAccrued));
        return expectedAccrued;
    };
    before(async function () {
        [delegator, curator1, curator2, indexer1, indexer2, oracle, assetHolder]
            = await graph.getTestAccounts();
        ({ governor } = await graph.getNamedAccounts());
        fixture = new fixtures_1.NetworkFixture(graph.provider);
        contracts = await fixture.load(governor);
        grt = contracts.GraphToken;
        curation = contracts.Curation;
        epochManager = contracts.EpochManager;
        staking = contracts.Staking;
        rewardsManager = contracts.RewardsManager;
        // 200 GRT per block
        await rewardsManager.connect(governor).setIssuancePerBlock(ISSUANCE_PER_BLOCK);
        // Distribute test funds
        for (const wallet of [indexer1, indexer2, curator1, curator2, assetHolder]) {
            await grt.connect(governor).mint(wallet.address, (0, sdk_1.toGRT)('1000000'));
            await grt.connect(wallet).approve(staking.address, (0, sdk_1.toGRT)('1000000'));
            await grt.connect(wallet).approve(curation.address, (0, sdk_1.toGRT)('1000000'));
        }
    });
    beforeEach(async function () {
        await fixture.setUp();
    });
    afterEach(async function () {
        await fixture.tearDown();
    });
    describe('configuration', function () {
        describe('issuance per block update', function () {
            it('reject set issuance per block if unauthorized', async function () {
                const tx = rewardsManager.connect(indexer1).setIssuancePerBlock((0, sdk_1.toGRT)('1.025'));
                await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
            });
            it('should set issuance rate to minimum allowed (0)', async function () {
                const newIssuancePerBlock = (0, sdk_1.toGRT)('0');
                await rewardsManager.connect(governor).setIssuancePerBlock(newIssuancePerBlock);
                (0, chai_1.expect)(await rewardsManager.issuancePerBlock()).eq(newIssuancePerBlock);
            });
            it('should set issuance rate', async function () {
                const newIssuancePerBlock = (0, sdk_1.toGRT)('100.025');
                await rewardsManager.connect(governor).setIssuancePerBlock(newIssuancePerBlock);
                (0, chai_1.expect)(await rewardsManager.issuancePerBlock()).eq(newIssuancePerBlock);
                (0, chai_1.expect)(await rewardsManager.accRewardsPerSignalLastBlockUpdated()).eq(await sdk_1.helpers.latestBlock());
            });
        });
        describe('subgraph availability service', function () {
            it('reject set subgraph oracle if unauthorized', async function () {
                const tx = rewardsManager.connect(indexer1).setSubgraphAvailabilityOracle(oracle.address);
                await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
            });
            it('should set subgraph oracle if governor', async function () {
                await rewardsManager.connect(governor).setSubgraphAvailabilityOracle(oracle.address);
                (0, chai_1.expect)(await rewardsManager.subgraphAvailabilityOracle()).eq(oracle.address);
            });
            it('reject to deny subgraph if not the oracle', async function () {
                const tx = rewardsManager.connect(governor).setDenied(subgraphDeploymentID1, true);
                await (0, chai_1.expect)(tx).revertedWith('Caller must be the subgraph availability oracle');
            });
            it('should deny subgraph', async function () {
                await rewardsManager.connect(governor).setSubgraphAvailabilityOracle(oracle.address);
                const tx = rewardsManager.connect(oracle).setDenied(subgraphDeploymentID1, true);
                const blockNum = await sdk_1.helpers.latestBlock();
                await (0, chai_1.expect)(tx)
                    .emit(rewardsManager, 'RewardsDenylistUpdated')
                    .withArgs(subgraphDeploymentID1, blockNum + 1);
                (0, chai_1.expect)(await rewardsManager.isDenied(subgraphDeploymentID1)).eq(true);
            });
        });
    });
    context('issuing rewards', function () {
        beforeEach(async function () {
            // 5% minute rate (4 blocks)
            await rewardsManager.connect(governor).setIssuancePerBlock(ISSUANCE_PER_BLOCK);
        });
        describe('getNewRewardsPerSignal', function () {
            it('accrued per signal when no tokens signalled', async function () {
                // When there is no tokens signalled no rewards are accrued
                await sdk_1.helpers.mineEpoch(epochManager);
                const accrued = await rewardsManager.getNewRewardsPerSignal();
                (0, chai_1.expect)(accrued).eq(0);
            });
            it('accrued per signal when tokens signalled', async function () {
                // Update total signalled
                const tokensToSignal = (0, sdk_1.toGRT)('1000');
                await curation.connect(curator1).mint(subgraphDeploymentID1, tokensToSignal, 0);
                // Check
                await shouldGetNewRewardsPerSignal();
            });
            it('accrued per signal when signalled tokens w/ many subgraphs', async function () {
                // Update total signalled
                await curation.connect(curator1).mint(subgraphDeploymentID1, (0, sdk_1.toGRT)('1000'), 0);
                // Check
                await shouldGetNewRewardsPerSignal();
                // Update total signalled
                await curation.connect(curator2).mint(subgraphDeploymentID2, (0, sdk_1.toGRT)('250'), 0);
                // Check
                await shouldGetNewRewardsPerSignal();
            });
        });
        describe('updateAccRewardsPerSignal', function () {
            it('update the accumulated rewards per signal state', async function () {
                // Update total signalled
                await curation.connect(curator1).mint(subgraphDeploymentID1, (0, sdk_1.toGRT)('1000'), 0);
                // Snapshot
                const tracker = await RewardsTracker.create();
                // Update
                await rewardsManager.connect(governor).updateAccRewardsPerSignal();
                const contractAccrued = await rewardsManager.accRewardsPerSignal();
                // Check
                const expectedAccrued = await tracker.accrued();
                (0, chai_1.expect)(toRound(expectedAccrued)).eq(toRound(contractAccrued));
            });
            it('update the accumulated rewards per signal state after many blocks', async function () {
                // Update total signalled
                await curation.connect(curator1).mint(subgraphDeploymentID1, (0, sdk_1.toGRT)('1000'), 0);
                // Snapshot
                const tracker = await RewardsTracker.create();
                // Jump
                await sdk_1.helpers.mine(ISSUANCE_RATE_PERIODS);
                // Update
                await rewardsManager.connect(governor).updateAccRewardsPerSignal();
                const contractAccrued = await rewardsManager.accRewardsPerSignal();
                // Check
                const expectedAccrued = await tracker.accrued();
                (0, chai_1.expect)(toRound(expectedAccrued)).eq(toRound(contractAccrued));
            });
        });
        describe('getAccRewardsForSubgraph', function () {
            it('accrued for each subgraph', async function () {
                // Curator1 - Update total signalled
                const signalled1 = (0, sdk_1.toGRT)('1500');
                await curation.connect(curator1).mint(subgraphDeploymentID1, signalled1, 0);
                const tracker1 = await RewardsTracker.create();
                // Curator2 - Update total signalled
                const signalled2 = (0, sdk_1.toGRT)('500');
                await curation.connect(curator2).mint(subgraphDeploymentID2, signalled2, 0);
                // Snapshot
                const tracker2 = await RewardsTracker.create();
                await tracker1.snapshot();
                // Jump
                await sdk_1.helpers.mine(ISSUANCE_RATE_PERIODS);
                // Snapshot
                await tracker1.snapshot();
                await tracker2.snapshot();
                // Calculate rewards
                const rewardsPerSignal1 = tracker1.accumulated;
                const rewardsPerSignal2 = tracker2.accumulated;
                const expectedRewardsSG1 = rewardsPerSignal1.mul(signalled1).div(WeiPerEther);
                const expectedRewardsSG2 = rewardsPerSignal2.mul(signalled2).div(WeiPerEther);
                // Get rewards from contract
                const contractRewardsSG1 = await rewardsManager.getAccRewardsForSubgraph(subgraphDeploymentID1);
                const contractRewardsSG2 = await rewardsManager.getAccRewardsForSubgraph(subgraphDeploymentID2);
                // Check
                (0, chai_1.expect)(toRound(expectedRewardsSG1)).eq(toRound(contractRewardsSG1));
                (0, chai_1.expect)(toRound(expectedRewardsSG2)).eq(toRound(contractRewardsSG2));
            });
        });
        describe('onSubgraphSignalUpdate', function () {
            it('update the accumulated rewards for subgraph state', async function () {
                // Update total signalled
                const signalled1 = (0, sdk_1.toGRT)('1500');
                await curation.connect(curator1).mint(subgraphDeploymentID1, signalled1, 0);
                // Snapshot
                const tracker1 = await RewardsTracker.create();
                // Jump
                await sdk_1.helpers.mine(ISSUANCE_RATE_PERIODS);
                // Update
                await rewardsManager.connect(governor).onSubgraphSignalUpdate(subgraphDeploymentID1);
                // Check
                const contractRewardsSG1 = (await rewardsManager.subgraphs(subgraphDeploymentID1))
                    .accRewardsForSubgraph;
                const rewardsPerSignal1 = await tracker1.accrued();
                const expectedRewardsSG1 = rewardsPerSignal1.mul(signalled1).div(WeiPerEther);
                (0, chai_1.expect)(toRound(expectedRewardsSG1)).eq(toRound(contractRewardsSG1));
                const contractAccrued = await rewardsManager.accRewardsPerSignal();
                const expectedAccrued = await tracker1.accrued();
                (0, chai_1.expect)(toRound(expectedAccrued)).eq(toRound(contractAccrued));
                const contractBlockUpdated = await rewardsManager.accRewardsPerSignalLastBlockUpdated();
                const expectedBlockUpdated = await sdk_1.helpers.latestBlock();
                (0, chai_1.expect)(expectedBlockUpdated).eq(contractBlockUpdated);
            });
        });
        describe('getAccRewardsPerAllocatedToken', function () {
            it('accrued per allocated token', async function () {
                // Update total signalled
                const signalled1 = (0, sdk_1.toGRT)('1500');
                await curation.connect(curator1).mint(subgraphDeploymentID1, signalled1, 0);
                // Allocate
                const tokensToAllocate = (0, sdk_1.toGRT)('12500');
                await staking.connect(indexer1).stake(tokensToAllocate);
                await staking
                    .connect(indexer1)
                    .allocateFrom(indexer1.address, subgraphDeploymentID1, tokensToAllocate, allocationID1, metadata, await channelKey1.generateProof(indexer1.address));
                // Jump
                await sdk_1.helpers.mine(ISSUANCE_RATE_PERIODS);
                // Check
                const sg1 = await rewardsManager.subgraphs(subgraphDeploymentID1);
                // We trust this function because it was individually tested in previous test
                const accRewardsForSubgraphSG1 = await rewardsManager.getAccRewardsForSubgraph(subgraphDeploymentID1);
                const accruedRewardsSG1 = accRewardsForSubgraphSG1.sub(sg1.accRewardsForSubgraphSnapshot);
                const expectedRewardsAT1 = accruedRewardsSG1.mul(WeiPerEther).div(tokensToAllocate);
                const contractRewardsAT1 = (await rewardsManager.getAccRewardsPerAllocatedToken(subgraphDeploymentID1))[0];
                (0, chai_1.expect)(expectedRewardsAT1).eq(contractRewardsAT1);
            });
        });
        describe('onSubgraphAllocationUpdate', function () {
            it('update the accumulated rewards for allocated tokens state', async function () {
                // Update total signalled
                const signalled1 = (0, sdk_1.toGRT)('1500');
                await curation.connect(curator1).mint(subgraphDeploymentID1, signalled1, 0);
                // Allocate
                const tokensToAllocate = (0, sdk_1.toGRT)('12500');
                await staking.connect(indexer1).stake(tokensToAllocate);
                await staking
                    .connect(indexer1)
                    .allocateFrom(indexer1.address, subgraphDeploymentID1, tokensToAllocate, allocationID1, metadata, await channelKey1.generateProof(indexer1.address));
                // Jump
                await sdk_1.helpers.mine(ISSUANCE_RATE_PERIODS);
                // Prepare expected results
                const expectedSubgraphRewards = (0, sdk_1.toGRT)('1400'); // 7 blocks since signaling to when we do getAccRewardsForSubgraph
                const expectedRewardsAT = (0, sdk_1.toGRT)('0.08'); // allocated during 5 blocks: 1000 GRT, divided by 12500 allocated tokens
                // Update
                await rewardsManager.connect(governor).onSubgraphAllocationUpdate(subgraphDeploymentID1);
                // Check on demand results saved
                const subgraph = await rewardsManager.subgraphs(subgraphDeploymentID1);
                const contractSubgraphRewards = await rewardsManager.getAccRewardsForSubgraph(subgraphDeploymentID1);
                const contractRewardsAT = subgraph.accRewardsPerAllocatedToken;
                (0, chai_1.expect)(toRound(expectedSubgraphRewards)).eq(toRound(contractSubgraphRewards));
                (0, chai_1.expect)(toRound(expectedRewardsAT.mul(1000))).eq(toRound(contractRewardsAT.mul(1000)));
            });
        });
        describe('getRewards', function () {
            it('calculate rewards using the subgraph signalled + allocated tokens', async function () {
                // Update total signalled
                const signalled1 = (0, sdk_1.toGRT)('1500');
                await curation.connect(curator1).mint(subgraphDeploymentID1, signalled1, 0);
                // Allocate
                const tokensToAllocate = (0, sdk_1.toGRT)('12500');
                await staking.connect(indexer1).stake(tokensToAllocate);
                await staking
                    .connect(indexer1)
                    .allocateFrom(indexer1.address, subgraphDeploymentID1, tokensToAllocate, allocationID1, metadata, await channelKey1.generateProof(indexer1.address));
                // Jump
                await sdk_1.helpers.mine(ISSUANCE_RATE_PERIODS);
                // Rewards
                const contractRewards = await rewardsManager.getRewards(staking.address, allocationID1);
                // We trust using this function in the test because we tested it
                // standalone in a previous test
                const contractRewardsAT1 = (await rewardsManager.getAccRewardsPerAllocatedToken(subgraphDeploymentID1))[0];
                const expectedRewards = contractRewardsAT1.mul(tokensToAllocate).div(WeiPerEther);
                (0, chai_1.expect)(expectedRewards).eq(contractRewards);
            });
            it('rewards should be zero if the allocation is closed', async function () {
                // Update total signalled
                const signalled1 = (0, sdk_1.toGRT)('1500');
                await curation.connect(curator1).mint(subgraphDeploymentID1, signalled1, 0);
                // Allocate
                const tokensToAllocate = (0, sdk_1.toGRT)('12500');
                await staking.connect(indexer1).stake(tokensToAllocate);
                await staking
                    .connect(indexer1)
                    .allocateFrom(indexer1.address, subgraphDeploymentID1, tokensToAllocate, allocationID1, metadata, await channelKey1.generateProof(indexer1.address));
                // Jump
                await sdk_1.helpers.mine(ISSUANCE_RATE_PERIODS);
                await sdk_1.helpers.mineEpoch(epochManager);
                // Close allocation
                await staking.connect(indexer1).closeAllocation(allocationID1, (0, sdk_1.randomHexBytes)());
                // Rewards
                const contractRewards = await rewardsManager.getRewards(staking.address, allocationID1);
                (0, chai_1.expect)(contractRewards).eq(ethers_1.BigNumber.from(0));
            });
            it('rewards should be zero if the allocation does not exist', async function () {
                // Rewards
                const contractRewards = await rewardsManager.getRewards(staking.address, allocationIDNull);
                (0, chai_1.expect)(contractRewards).eq(ethers_1.BigNumber.from(0));
            });
        });
        describe('takeRewards', function () {
            async function setupIndexerAllocation() {
                // Setup
                await epochManager.connect(governor).setEpochLength(10);
                // Update total signalled
                const signalled1 = (0, sdk_1.toGRT)('1500');
                await curation.connect(curator1).mint(subgraphDeploymentID1, signalled1, 0);
                // Allocate
                const tokensToAllocate = (0, sdk_1.toGRT)('12500');
                await staking.connect(indexer1).stake(tokensToAllocate);
                await staking
                    .connect(indexer1)
                    .allocateFrom(indexer1.address, subgraphDeploymentID1, tokensToAllocate, allocationID1, metadata, await channelKey1.generateProof(indexer1.address));
            }
            async function setupIndexerAllocationSignalingAfter() {
                // Setup
                await epochManager.connect(governor).setEpochLength(10);
                // Allocate
                const tokensToAllocate = (0, sdk_1.toGRT)('12500');
                await staking.connect(indexer1).stake(tokensToAllocate);
                await staking
                    .connect(indexer1)
                    .allocateFrom(indexer1.address, subgraphDeploymentID1, tokensToAllocate, allocationID1, metadata, await channelKey1.generateProof(indexer1.address));
                // Update total signalled
                const signalled1 = (0, sdk_1.toGRT)('1500');
                await curation.connect(curator1).mint(subgraphDeploymentID1, signalled1, 0);
            }
            async function setupIndexerAllocationWithDelegation(tokensToDelegate, delegationParams) {
                const tokensToAllocate = (0, sdk_1.toGRT)('12500');
                // Setup
                await epochManager.connect(governor).setEpochLength(10);
                // Transfer some funds from the curator, I don't want to mint new tokens
                await grt.connect(curator1).transfer(delegator.address, tokensToDelegate);
                await grt.connect(delegator).approve(staking.address, tokensToDelegate);
                // Stake and set delegation parameters
                await staking.connect(indexer1).stake(tokensToAllocate);
                await staking
                    .connect(indexer1)
                    .setDelegationParameters(delegationParams.indexingRewardCut, delegationParams.queryFeeCut, 0);
                // Delegate
                await staking.connect(delegator).delegate(indexer1.address, tokensToDelegate);
                // Update total signalled
                const signalled1 = (0, sdk_1.toGRT)('1500');
                await curation.connect(curator1).mint(subgraphDeploymentID1, signalled1, 0);
                // Allocate
                await staking
                    .connect(indexer1)
                    .allocateFrom(indexer1.address, subgraphDeploymentID1, tokensToAllocate, allocationID1, metadata, await channelKey1.generateProof(indexer1.address));
            }
            it('should distribute rewards on closed allocation and stake', async function () {
                // Align with the epoch boundary
                await sdk_1.helpers.mineEpoch(epochManager);
                // Setup
                await setupIndexerAllocation();
                // Jump
                await sdk_1.helpers.mineEpoch(epochManager);
                // Before state
                const beforeTokenSupply = await grt.totalSupply();
                const beforeIndexer1Stake = await staking.getIndexerStakedTokens(indexer1.address);
                const beforeIndexer1Balance = await grt.balanceOf(indexer1.address);
                const beforeStakingBalance = await grt.balanceOf(staking.address);
                // All the rewards in this subgraph go to this allocation.
                // Rewards per token will be (issuancePerBlock * nBlocks) / allocatedTokens
                // The first snapshot is after allocating, that is 2 blocks after the signal is minted.
                // The final snapshot is when we close the allocation, that happens 9 blocks after signal is minted.
                // So the rewards will be ((issuancePerBlock * 7) / allocatedTokens) * allocatedTokens
                const expectedIndexingRewards = (0, sdk_1.toGRT)('1400');
                // Close allocation. At this point rewards should be collected for that indexer
                const tx = await staking.connect(indexer1).closeAllocation(allocationID1, (0, sdk_1.randomHexBytes)());
                const receipt = await tx.wait();
                const event = rewardsManager.interface.parseLog(receipt.logs[1]).args;
                (0, chai_1.expect)(event.indexer).eq(indexer1.address);
                (0, chai_1.expect)(event.allocationID).eq(allocationID1);
                (0, chai_1.expect)(toRound(event.amount)).eq(toRound(expectedIndexingRewards));
                // After state
                const afterTokenSupply = await grt.totalSupply();
                const afterIndexer1Stake = await staking.getIndexerStakedTokens(indexer1.address);
                const afterIndexer1Balance = await grt.balanceOf(indexer1.address);
                const afterStakingBalance = await grt.balanceOf(staking.address);
                // Check that rewards are put into indexer stake
                const expectedIndexerStake = beforeIndexer1Stake.add(expectedIndexingRewards);
                const expectedTokenSupply = beforeTokenSupply.add(expectedIndexingRewards);
                // Check stake should have increased with the rewards staked
                (0, chai_1.expect)(toRound(afterIndexer1Stake)).eq(toRound(expectedIndexerStake));
                // Check indexer balance remains the same
                (0, chai_1.expect)(afterIndexer1Balance).eq(beforeIndexer1Balance);
                // Check indexing rewards are kept in the staking contract
                (0, chai_1.expect)(toRound(afterStakingBalance)).eq(toRound(beforeStakingBalance.add(expectedIndexingRewards)));
                // Check that tokens have been minted
                (0, chai_1.expect)(toRound(afterTokenSupply)).eq(toRound(expectedTokenSupply));
            });
            it('does not revert with an underflow if the minimum signal changes', async function () {
                // Align with the epoch boundary
                await sdk_1.helpers.mineEpoch(epochManager);
                // Setup
                await setupIndexerAllocation();
                await rewardsManager.connect(governor).setMinimumSubgraphSignal((0, sdk_1.toGRT)(14000));
                // Jump
                await sdk_1.helpers.mineEpoch(epochManager);
                // Close allocation. At this point rewards should be collected for that indexer
                const tx = staking.connect(indexer1).closeAllocation(allocationID1, (0, sdk_1.randomHexBytes)());
                await (0, chai_1.expect)(tx)
                    .emit(rewardsManager, 'HorizonRewardsAssigned')
                    .withArgs(indexer1.address, allocationID1, (0, sdk_1.toBN)(0));
            });
            it('does not revert with an underflow if the minimum signal changes, and signal came after allocation', async function () {
                // Align with the epoch boundary
                await sdk_1.helpers.mineEpoch(epochManager);
                // Setup
                await setupIndexerAllocationSignalingAfter();
                await rewardsManager.connect(governor).setMinimumSubgraphSignal((0, sdk_1.toGRT)(14000));
                // Jump
                await sdk_1.helpers.mineEpoch(epochManager);
                // Close allocation. At this point rewards should be collected for that indexer
                const tx = staking.connect(indexer1).closeAllocation(allocationID1, (0, sdk_1.randomHexBytes)());
                await (0, chai_1.expect)(tx)
                    .emit(rewardsManager, 'HorizonRewardsAssigned')
                    .withArgs(indexer1.address, allocationID1, (0, sdk_1.toBN)(0));
            });
            it('does not revert if signal was already under minimum', async function () {
                await rewardsManager.connect(governor).setMinimumSubgraphSignal((0, sdk_1.toGRT)(2000));
                // Align with the epoch boundary
                await sdk_1.helpers.mineEpoch(epochManager);
                // Setup
                await setupIndexerAllocation();
                // Jump
                await sdk_1.helpers.mineEpoch(epochManager);
                // Close allocation. At this point rewards should be collected for that indexer
                const tx = staking.connect(indexer1).closeAllocation(allocationID1, (0, sdk_1.randomHexBytes)());
                await (0, chai_1.expect)(tx)
                    .emit(rewardsManager, 'HorizonRewardsAssigned')
                    .withArgs(indexer1.address, allocationID1, (0, sdk_1.toBN)(0));
            });
            it('should distribute rewards on closed allocation and send to destination', async function () {
                const destinationAddress = (0, sdk_1.randomHexBytes)(20);
                await staking.connect(indexer1).setRewardsDestination(destinationAddress);
                // Align with the epoch boundary
                await sdk_1.helpers.mineEpoch(epochManager);
                // Setup
                await setupIndexerAllocation();
                // Jump
                await sdk_1.helpers.mineEpoch(epochManager);
                // Before state
                const beforeTokenSupply = await grt.totalSupply();
                const beforeIndexer1Stake = await staking.getIndexerStakedTokens(indexer1.address);
                const beforeDestinationBalance = await grt.balanceOf(destinationAddress);
                const beforeStakingBalance = await grt.balanceOf(staking.address);
                // All the rewards in this subgraph go to this allocation.
                // Rewards per token will be (issuancePerBlock * nBlocks) / allocatedTokens
                // The first snapshot is after allocating, that is 2 blocks after the signal is minted.
                // The final snapshot is when we close the allocation, that happens 9 blocks after signal is minted.
                // So the rewards will be ((issuancePerBlock * 7) / allocatedTokens) * allocatedTokens
                const expectedIndexingRewards = (0, sdk_1.toGRT)('1400');
                // Close allocation. At this point rewards should be collected for that indexer
                const tx = await staking.connect(indexer1).closeAllocation(allocationID1, (0, sdk_1.randomHexBytes)());
                const receipt = await tx.wait();
                const event = rewardsManager.interface.parseLog(receipt.logs[1]).args;
                (0, chai_1.expect)(event.indexer).eq(indexer1.address);
                (0, chai_1.expect)(event.allocationID).eq(allocationID1);
                (0, chai_1.expect)(toRound(event.amount)).eq(toRound(expectedIndexingRewards));
                // After state
                const afterTokenSupply = await grt.totalSupply();
                const afterIndexer1Stake = await staking.getIndexerStakedTokens(indexer1.address);
                const afterDestinationBalance = await grt.balanceOf(destinationAddress);
                const afterStakingBalance = await grt.balanceOf(staking.address);
                // Check that rewards are properly assigned
                const expectedIndexerStake = beforeIndexer1Stake;
                const expectedTokenSupply = beforeTokenSupply.add(expectedIndexingRewards);
                // Check stake should not have changed
                (0, chai_1.expect)(toRound(afterIndexer1Stake)).eq(toRound(expectedIndexerStake));
                // Check indexing rewards are received by the rewards destination
                (0, chai_1.expect)(toRound(afterDestinationBalance)).eq(toRound(beforeDestinationBalance.add(expectedIndexingRewards)));
                // Check indexing rewards were not sent to the staking contract
                (0, chai_1.expect)(afterStakingBalance).eq(beforeStakingBalance);
                // Check that tokens have been minted
                (0, chai_1.expect)(toRound(afterTokenSupply)).eq(toRound(expectedTokenSupply));
            });
            it('should distribute rewards on closed allocation w/delegators', async function () {
                // Setup
                const delegationParams = {
                    indexingRewardCut: (0, sdk_1.toBN)('823000'), // 82.30%
                    queryFeeCut: (0, sdk_1.toBN)('80000'), // 8%
                    cooldownBlocks: 0,
                };
                const tokensToDelegate = (0, sdk_1.toGRT)('2000');
                // Align with the epoch boundary
                await sdk_1.helpers.mineEpoch(epochManager);
                // Setup the allocation and delegators
                await setupIndexerAllocationWithDelegation(tokensToDelegate, delegationParams);
                // Jump
                await sdk_1.helpers.mineEpoch(epochManager);
                // Before state
                const beforeTokenSupply = await grt.totalSupply();
                const beforeDelegationPool = await staking.delegationPools(indexer1.address);
                const beforeIndexer1Stake = await staking.getIndexerStakedTokens(indexer1.address);
                // Close allocation. At this point rewards should be collected for that indexer
                await staking.connect(indexer1).closeAllocation(allocationID1, (0, sdk_1.randomHexBytes)());
                // After state
                const afterTokenSupply = await grt.totalSupply();
                const afterDelegationPool = await staking.delegationPools(indexer1.address);
                const afterIndexer1Stake = await staking.getIndexerStakedTokens(indexer1.address);
                // Check that rewards are put into indexer stake (only indexer cut)
                // Check that rewards are put into delegators pool accordingly
                // All the rewards in this subgraph go to this allocation.
                // Rewards per token will be (issuancePerBlock * nBlocks) / allocatedTokens
                // The first snapshot is after allocating, that is 1 block after the signal is minted.
                // The final snapshot is when we close the allocation, that happens 4 blocks after signal is minted.
                // So the rewards will be ((issuancePerBlock * 3) / allocatedTokens) * allocatedTokens
                const expectedIndexingRewards = (0, sdk_1.toGRT)('600');
                // Calculate delegators cut
                const indexerRewards = delegationParams.indexingRewardCut
                    .mul(expectedIndexingRewards)
                    .div((0, sdk_1.toBN)(MAX_PPM));
                // Calculate indexer cut
                const delegatorsRewards = expectedIndexingRewards.sub(indexerRewards);
                // Check
                const expectedIndexerStake = beforeIndexer1Stake.add(indexerRewards);
                const expectedDelegatorsPoolTokens = beforeDelegationPool.tokens.add(delegatorsRewards);
                const expectedTokenSupply = beforeTokenSupply.add(expectedIndexingRewards);
                (0, chai_1.expect)(toRound(afterIndexer1Stake)).eq(toRound(expectedIndexerStake));
                (0, chai_1.expect)(toRound(afterDelegationPool.tokens)).eq(toRound(expectedDelegatorsPoolTokens));
                // Check that tokens have been minted
                (0, chai_1.expect)(toRound(afterTokenSupply)).eq(toRound(expectedTokenSupply));
            });
            it('should deny rewards if subgraph on denylist', async function () {
                // Setup
                await rewardsManager.connect(governor).setSubgraphAvailabilityOracle(governor.address);
                await rewardsManager.connect(governor).setDenied(subgraphDeploymentID1, true);
                await setupIndexerAllocation();
                // Jump
                await sdk_1.helpers.mineEpoch(epochManager);
                // Close allocation. At this point rewards should be collected for that indexer
                const tx = staking.connect(indexer1).closeAllocation(allocationID1, (0, sdk_1.randomHexBytes)());
                await (0, chai_1.expect)(tx)
                    .emit(rewardsManager, 'RewardsDenied')
                    .withArgs(indexer1.address, allocationID1);
            });
            it('should not distribute rewards if allocation is not active', async function () {
                // Setup
                await setupIndexerAllocation();
                // Jump and close allocation
                await sdk_1.helpers.mineEpoch(epochManager);
                await staking.connect(indexer1).closeAllocation(allocationID1, (0, sdk_1.randomHexBytes)());
                // Jump some more
                await sdk_1.helpers.mineEpoch(epochManager, 10);
                // Impersonate staking contract
                await (0, hardhat_network_helpers_1.impersonateAccount)(staking.address);
                const impersonatedStaking = await hardhat_1.default.ethers.getSigner(staking.address);
                await (0, hardhat_network_helpers_1.setBalance)(staking.address, (0, sdk_1.toGRT)('1000000'));
                // Distribute rewards
                const tx = await rewardsManager.connect(impersonatedStaking).takeRewards(allocationID1);
                const receipt = await tx.wait();
                const event = rewardsManager.interface.parseLog(receipt.logs[0]).args;
                (0, chai_1.expect)(event.indexer).eq(indexer1.address);
                (0, chai_1.expect)(event.allocationID).eq(allocationID1);
                (0, chai_1.expect)(toRound(event.amount)).eq(toRound(ethers_1.BigNumber.from(0)));
            });
        });
    });
    describe('edge scenarios', function () {
        it('close allocation on a subgraph that no longer have signal', async function () {
            // Update total signalled
            const signalled1 = (0, sdk_1.toGRT)('1500');
            await curation.connect(curator1).mint(subgraphDeploymentID1, signalled1, 0);
            // Allocate
            const tokensToAllocate = (0, sdk_1.toGRT)('12500');
            await staking.connect(indexer1).stake(tokensToAllocate);
            await staking
                .connect(indexer1)
                .allocateFrom(indexer1.address, subgraphDeploymentID1, tokensToAllocate, allocationID1, metadata, await channelKey1.generateProof(indexer1.address));
            // Jump
            await sdk_1.helpers.mineEpoch(epochManager);
            // Remove all signal from the subgraph
            const curatorShares = await curation.getCuratorSignal(curator1.address, subgraphDeploymentID1);
            await curation.connect(curator1).burn(subgraphDeploymentID1, curatorShares, 0);
            // Close allocation. At this point rewards should be collected for that indexer
            await staking.connect(indexer1).closeAllocation(allocationID1, (0, sdk_1.randomHexBytes)());
        });
    });
    describe('multiple allocations', function () {
        it('two allocations in the same block with a GRT burn in the middle should succeed', async function () {
            // If rewards are not monotonically increasing, this can trigger
            // a subtraction overflow error as seen in mainnet tx:
            // 0xb6bf7bbc446720a7409c482d714aebac239dd62e671c3c94f7e93dd3a61835ab
            await sdk_1.helpers.mineEpoch(epochManager);
            // Setup
            await epochManager.connect(governor).setEpochLength(10);
            // Update total signalled
            const signalled1 = (0, sdk_1.toGRT)('1500');
            await curation.connect(curator1).mint(subgraphDeploymentID1, signalled1, 0);
            // Stake
            const tokensToStake = (0, sdk_1.toGRT)('12500');
            await staking.connect(indexer1).stake(tokensToStake);
            // Allocate simultaneously, burning in the middle
            const tokensToAlloc = (0, sdk_1.toGRT)('5000');
            await sdk_1.helpers.setAutoMine(false);
            const tx1 = await staking
                .connect(indexer1)
                .allocateFrom(indexer1.address, subgraphDeploymentID1, tokensToAlloc, allocationID1, metadata, await channelKey1.generateProof(indexer1.address));
            const tx2 = await grt.connect(indexer1).burn((0, sdk_1.toGRT)(1));
            const tx3 = await staking
                .connect(indexer1)
                .allocateFrom(indexer1.address, subgraphDeploymentID1, tokensToAlloc, allocationID2, metadata, await channelKey2.generateProof(indexer1.address));
            await sdk_1.helpers.mine();
            await sdk_1.helpers.setAutoMine(true);
            await (0, chai_1.expect)(tx1).emit(staking, 'AllocationCreated');
            await (0, chai_1.expect)(tx2).emit(grt, 'Transfer');
            await (0, chai_1.expect)(tx3).emit(staking, 'AllocationCreated');
        });
        it('two simultanous-similar allocations should get same amount of rewards', async function () {
            await sdk_1.helpers.mineEpoch(epochManager);
            // Setup
            await epochManager.connect(governor).setEpochLength(10);
            // Update total signalled
            const signalled1 = (0, sdk_1.toGRT)('1500');
            await curation.connect(curator1).mint(subgraphDeploymentID1, signalled1, 0);
            // Stake
            const tokensToStake = (0, sdk_1.toGRT)('12500');
            await staking.connect(indexer1).stake(tokensToStake);
            // Allocate simultaneously
            const tokensToAlloc = (0, sdk_1.toGRT)('5000');
            const tx1 = await staking.populateTransaction.allocateFrom(indexer1.address, subgraphDeploymentID1, tokensToAlloc, allocationID1, metadata, await channelKey1.generateProof(indexer1.address));
            const tx2 = await staking.populateTransaction.allocateFrom(indexer1.address, subgraphDeploymentID1, tokensToAlloc, allocationID2, metadata, await channelKey2.generateProof(indexer1.address));
            await staking.connect(indexer1).multicall([tx1.data, tx2.data]);
            // Jump
            await sdk_1.helpers.mineEpoch(epochManager);
            // Close allocations simultaneously
            const tx3 = await staking.populateTransaction.closeAllocation(allocationID1, (0, sdk_1.randomHexBytes)());
            const tx4 = await staking.populateTransaction.closeAllocation(allocationID2, (0, sdk_1.randomHexBytes)());
            const tx5 = await staking.connect(indexer1).multicall([tx3.data, tx4.data]);
            // Both allocations should receive the same amount of rewards
            const receipt = await tx5.wait();
            const event1 = rewardsManager.interface.parseLog(receipt.logs[1]).args;
            const event2 = rewardsManager.interface.parseLog(receipt.logs[5]).args;
            (0, chai_1.expect)(event1.amount).eq(event2.amount);
        });
    });
    describe('rewards progression when collecting query fees', function () {
        it('collect query fees with two subgraphs and one allocation', async function () {
            async function getRewardsAccrual(subgraphs) {
                const [sg1, sg2] = await Promise.all(subgraphs.map(sg => rewardsManager.getAccRewardsForSubgraph(sg)));
                return {
                    sg1,
                    sg2,
                    all: sg1.add(sg2),
                };
            }
            // set curation percentage
            await staking.connect(governor).setCurationPercentage(100000);
            // allow the asset holder
            const tokensToCollect = (0, sdk_1.toGRT)('10000');
            // signal in two subgraphs in the same block
            const subgraphs = [subgraphDeploymentID1, subgraphDeploymentID2];
            for (const sub of subgraphs) {
                await curation.connect(curator1).mint(sub, (0, sdk_1.toGRT)('1500'), 0);
            }
            // snapshot block before any accrual (we substract 1 because accrual starts after the first mint happens)
            const b1 = await epochManager.blockNum().then(x => x.toNumber() - 1);
            // allocate
            const tokensToAllocate = (0, sdk_1.toGRT)('12500');
            await staking
                .connect(indexer1)
                .multicall([
                await staking.populateTransaction.stake(tokensToAllocate).then(tx => tx.data),
                await staking.populateTransaction
                    .allocateFrom(indexer1.address, subgraphDeploymentID1, tokensToAllocate, allocationID1, metadata, await channelKey1.generateProof(indexer1.address))
                    .then(tx => tx.data),
            ]);
            // move time fwd
            await sdk_1.helpers.mineEpoch(epochManager);
            // collect funds into staking for that sub
            await staking.connect(assetHolder).collect(tokensToCollect, allocationID1);
            // check rewards diff
            await rewardsManager.getRewards(staking.address, allocationID1).then(sdk_1.formatGRT);
            await sdk_1.helpers.mine();
            const accrual = await getRewardsAccrual(subgraphs);
            const b2 = await epochManager.blockNum().then(x => x.toNumber());
            // round comparison because there is a small precision error due to dividing and accrual per signal
            (0, chai_1.expect)(toRound(accrual.all)).eq(toRound(ISSUANCE_PER_BLOCK.mul(b2 - b1)));
        });
    });
});
