"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const hardhat_1 = __importDefault(require("hardhat"));
const chai_1 = require("chai");
const hardhat_2 = require("hardhat");
const ethers_1 = require("ethers");
const fixtures_1 = require("../lib/fixtures");
const sdk_1 = require("@graphprotocol/sdk");
const { AddressZero } = ethers_1.constants;
const MAX_PPM = (0, sdk_1.toBN)('1000000');
describe('Staking:Config', () => {
    const graph = hardhat_1.default.graph();
    let me;
    let other;
    let governor;
    let fixture;
    let contracts;
    let staking;
    let proxyAdmin;
    before(async function () {
        [me, other] = await graph.getTestAccounts();
        ({ governor } = await graph.getNamedAccounts());
        fixture = new fixtures_1.NetworkFixture(graph.provider);
        contracts = await fixture.load(governor);
        staking = contracts.Staking;
        proxyAdmin = contracts.GraphProxyAdmin;
    });
    beforeEach(async function () {
        await fixture.setUp();
    });
    afterEach(async function () {
        await fixture.tearDown();
    });
    describe('minimumIndexerStake', function () {
        it('should set `minimumIndexerStake`', async function () {
            const oldValue = (0, sdk_1.toGRT)('10');
            const newValue = (0, sdk_1.toGRT)('100');
            // Set right in the constructor
            (0, chai_1.expect)(await staking.minimumIndexerStake()).eq(oldValue);
            // Set new value
            await staking.connect(governor).setMinimumIndexerStake(newValue);
            (0, chai_1.expect)(await staking.minimumIndexerStake()).eq(newValue);
        });
        it('reject set `minimumIndexerStake` if not allowed', async function () {
            const newValue = (0, sdk_1.toGRT)('100');
            const tx = staking.connect(me).setMinimumIndexerStake(newValue);
            await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
        });
        it('reject set `minimumIndexerStake` to zero', async function () {
            const tx = staking.connect(governor).setMinimumIndexerStake(0);
            await (0, chai_1.expect)(tx).revertedWith('!minimumIndexerStake');
        });
    });
    describe('setSlasher', function () {
        it('should set `slasher`', async function () {
            (0, chai_1.expect)(await staking.slashers(me.address)).eq(false);
            await staking.connect(governor).setSlasher(me.address, true);
            (0, chai_1.expect)(await staking.slashers(me.address)).eq(true);
            await staking.connect(governor).setSlasher(me.address, false);
            (0, chai_1.expect)(await staking.slashers(me.address)).eq(false);
        });
        it('reject set `slasher` if not allowed', async function () {
            const tx = staking.connect(other).setSlasher(me.address, true);
            await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
        });
        it('reject set `slasher` for zero', async function () {
            const tx = staking.connect(governor).setSlasher(AddressZero, true);
            await (0, chai_1.expect)(tx).revertedWith('!slasher');
        });
    });
    describe('curationPercentage', function () {
        it('should set `curationPercentage`', async function () {
            const newValue = (0, sdk_1.toBN)('5');
            await staking.connect(governor).setCurationPercentage(newValue);
            (0, chai_1.expect)(await staking.curationPercentage()).eq(newValue);
        });
        it('reject set `curationPercentage` if out of bounds', async function () {
            const newValue = MAX_PPM.add((0, sdk_1.toBN)('1'));
            const tx = staking.connect(governor).setCurationPercentage(newValue);
            await (0, chai_1.expect)(tx).revertedWith('>percentage');
        });
        it('reject set `curationPercentage` if not allowed', async function () {
            const tx = staking.connect(other).setCurationPercentage(50);
            await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
        });
    });
    describe('protocolPercentage', function () {
        it('should set `protocolPercentage`', async function () {
            for (const newValue of [(0, sdk_1.toBN)('0'), (0, sdk_1.toBN)('5'), MAX_PPM]) {
                await staking.connect(governor).setProtocolPercentage(newValue);
                (0, chai_1.expect)(await staking.protocolPercentage()).eq(newValue);
            }
        });
        it('reject set `protocolPercentage` if out of bounds', async function () {
            const newValue = MAX_PPM.add((0, sdk_1.toBN)('1'));
            const tx = staking.connect(governor).setProtocolPercentage(newValue);
            await (0, chai_1.expect)(tx).revertedWith('>percentage');
        });
        it('reject set `protocolPercentage` if not allowed', async function () {
            const tx = staking.connect(other).setProtocolPercentage(50);
            await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
        });
    });
    describe('maxAllocationEpochs', function () {
        it('should set `maxAllocationEpochs`', async function () {
            const newValue = (0, sdk_1.toBN)('5');
            await staking.connect(governor).setMaxAllocationEpochs(newValue);
            (0, chai_1.expect)(await staking.maxAllocationEpochs()).eq(newValue);
        });
        it('reject set `maxAllocationEpochs` if not allowed', async function () {
            const newValue = (0, sdk_1.toBN)('5');
            const tx = staking.connect(other).setMaxAllocationEpochs(newValue);
            await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
        });
    });
    describe('thawingPeriod', function () {
        it('should set `thawingPeriod`', async function () {
            const newValue = (0, sdk_1.toBN)('5');
            await staking.connect(governor).setThawingPeriod(newValue);
            (0, chai_1.expect)(await staking.thawingPeriod()).eq(newValue);
        });
        it('reject set `thawingPeriod` if not allowed', async function () {
            const newValue = (0, sdk_1.toBN)('5');
            const tx = staking.connect(other).setThawingPeriod(newValue);
            await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
        });
        it('reject set `thawingPeriod` to zero', async function () {
            const tx = staking.connect(governor).setThawingPeriod(0);
            await (0, chai_1.expect)(tx).revertedWith('!thawingPeriod');
        });
    });
    describe('rebateParameters', function () {
        it('should be setup on init', async function () {
            (0, chai_1.expect)(await staking.alphaNumerator()).eq((0, sdk_1.toBN)(100));
            (0, chai_1.expect)(await staking.alphaDenominator()).eq((0, sdk_1.toBN)(100));
            (0, chai_1.expect)(await staking.lambdaNumerator()).eq((0, sdk_1.toBN)(60));
            (0, chai_1.expect)(await staking.lambdaDenominator()).eq((0, sdk_1.toBN)(100));
        });
        it('should set `rebateParameters`', async function () {
            await staking.connect(governor).setRebateParameters(5, 6, 7, 8);
            (0, chai_1.expect)(await staking.alphaNumerator()).eq((0, sdk_1.toBN)(5));
            (0, chai_1.expect)(await staking.alphaDenominator()).eq((0, sdk_1.toBN)(6));
            (0, chai_1.expect)(await staking.lambdaNumerator()).eq((0, sdk_1.toBN)(7));
            (0, chai_1.expect)(await staking.lambdaDenominator()).eq((0, sdk_1.toBN)(8));
        });
        it('reject set `rebateParameters` if out of bounds', async function () {
            const tx2 = staking.connect(governor).setRebateParameters(1, 0, 1, 1);
            await (0, chai_1.expect)(tx2).revertedWith('!alphaDenominator');
            const tx3 = staking.connect(governor).setRebateParameters(1, 1, 0, 1);
            await (0, chai_1.expect)(tx3).revertedWith('!lambdaNumerator');
            const tx4 = staking.connect(governor).setRebateParameters(1, 1, 1, 0);
            await (0, chai_1.expect)(tx4).revertedWith('!lambdaDenominator');
        });
        it('reject set `rebateParameters` if not allowed', async function () {
            const tx = staking.connect(other).setRebateParameters(1, 1, 1, 1);
            await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
        });
    });
    describe('Staking and StakingExtension', function () {
        it('does not allow calling the fallback from the Staking implementation', async function () {
            const impl = await proxyAdmin.getProxyImplementation(staking.address);
            const factory = await hardhat_2.ethers.getContractFactory('StakingExtension');
            const implAsStaking = factory.attach(impl);
            const tx = implAsStaking.connect(other).setDelegationRatio(50);
            await (0, chai_1.expect)(tx).revertedWith('only through proxy');
        });
        it('can set the staking extension implementation with setExtensionImpl', async function () {
            const newImpl = await (0, sdk_1.deploy)(sdk_1.DeployType.Deploy, governor, {
                name: 'StakingExtension',
            });
            const tx = await staking.connect(governor).setExtensionImpl(newImpl.contract.address);
            await (0, chai_1.expect)(tx)
                .emit(staking, 'ExtensionImplementationSet')
                .withArgs(newImpl.contract.address);
        });
        it('rejects calls to setExtensionImpl from non-governor', async function () {
            const newImpl = await (0, sdk_1.deploy)(sdk_1.DeployType.Deploy, governor, { name: 'StakingExtension' });
            const tx = staking.connect(other).setExtensionImpl(newImpl.contract.address);
            await (0, chai_1.expect)(tx).revertedWith('Only Controller governor');
        });
    });
});
