"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const hardhat_1 = __importDefault(require("hardhat"));
const chai_1 = require("chai");
const ethers_1 = require("ethers");
const utils_1 = require("ethers/lib/utils");
const fixtures_1 = require("../lib/fixtures");
const sdk_1 = require("@graphprotocol/sdk");
const { AddressZero } = ethers_1.constants;
describe('L1Staking:L2Transfer', () => {
    const graph = hardhat_1.default.graph();
    let governor;
    let indexer;
    let l2Indexer;
    let delegator;
    let l2Delegator;
    let fixture;
    let fixtureContracts;
    let l2MockContracts;
    let l2StakingMock;
    let l2GRTGatewayMock;
    let grt;
    let staking;
    let controller;
    let l1GraphTokenGateway;
    let l1GraphTokenLockTransferTool;
    let l1GraphTokenLockTransferToolBad;
    // Test values
    const indexerTokens = (0, sdk_1.toGRT)('10000000');
    const delegatorTokens = (0, sdk_1.toGRT)('1000000');
    const tokensToStake = (0, sdk_1.toGRT)('200000');
    const subgraphDeploymentID = (0, sdk_1.randomHexBytes)();
    const channelKey = (0, sdk_1.deriveChannelKey)();
    const allocationID = channelKey.address;
    const metadata = (0, sdk_1.randomHexBytes)(32);
    const minimumIndexerStake = (0, sdk_1.toGRT)('100000');
    const delegationTaxPPM = 10000; // 1%
    // Dummy L2 gas values
    const maxGas = (0, sdk_1.toBN)('1000000');
    const gasPriceBid = (0, sdk_1.toBN)('1000000000');
    const maxSubmissionCost = (0, sdk_1.toBN)('1000000000');
    // Allocate with test values
    const allocate = async (tokens) => {
        return staking
            .connect(indexer)
            .allocateFrom(indexer.address, subgraphDeploymentID, tokens, allocationID, metadata, await channelKey.generateProof(indexer.address));
    };
    before(async function () {
        [indexer, delegator, l2Indexer, l2Delegator] = await graph.getTestAccounts();
        ({ governor } = await graph.getNamedAccounts());
        fixture = new fixtures_1.NetworkFixture(graph.provider);
        // Deploy L1
        fixtureContracts = await fixture.load(governor);
        grt = fixtureContracts.GraphToken;
        staking = fixtureContracts.L1Staking;
        l1GraphTokenGateway = fixtureContracts.L1GraphTokenGateway;
        controller = fixtureContracts.Controller;
        // Deploy L1 arbitrum bridge
        await fixture.loadL1ArbitrumBridge(governor);
        // Deploy L2 mock
        l2MockContracts = await fixture.loadMock(true);
        l2StakingMock = l2MockContracts.L2Staking;
        l2GRTGatewayMock = l2MockContracts.L2GraphTokenGateway;
        // Configure graph bridge
        await fixture.configureL1Bridge(governor, fixtureContracts, l2MockContracts);
        l1GraphTokenLockTransferTool = (await (0, sdk_1.deploy)(sdk_1.DeployType.Deploy, governor, {
            name: 'L1GraphTokenLockTransferToolMock',
        })).contract;
        l1GraphTokenLockTransferToolBad = (await (0, sdk_1.deploy)(sdk_1.DeployType.Deploy, governor, { name: 'L1GraphTokenLockTransferToolBadMock' })).contract;
        await sdk_1.helpers.setBalances([
            { address: l1GraphTokenLockTransferTool.address, balance: (0, utils_1.parseEther)('1') },
            { address: l1GraphTokenLockTransferToolBad.address, balance: (0, utils_1.parseEther)('1') },
        ]);
        await staking
            .connect(governor)
            .setL1GraphTokenLockTransferTool(l1GraphTokenLockTransferTool.address);
        // Give some funds to the indexer and approve staking contract to use funds on indexer behalf
        await grt.connect(governor).mint(indexer.address, indexerTokens);
        await grt.connect(indexer).approve(staking.address, indexerTokens);
        await grt.connect(governor).mint(delegator.address, delegatorTokens);
        await grt.connect(delegator).approve(staking.address, delegatorTokens);
        await staking.connect(governor).setMinimumIndexerStake(minimumIndexerStake);
        await staking.connect(governor).setDelegationTaxPercentage(delegationTaxPPM); // 1%
    });
    beforeEach(async function () {
        await fixture.setUp();
    });
    afterEach(async function () {
        await fixture.tearDown();
    });
    context('> when not staked', function () {
        describe('transferStakeToL2', function () {
            it('should not allow transferring for someone who has not staked', async function () {
                const tx = staking
                    .connect(indexer)
                    .transferStakeToL2(l2Indexer.address, tokensToStake, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                await (0, chai_1.expect)(tx).revertedWith('tokensStaked == 0');
            });
        });
    });
    context('> when staked', function () {
        const shouldTransferIndexerStake = async (amountToSend, options = {}) => {
            const l2Beneficiary = options.l2Beneficiary ?? l2Indexer.address;
            const expectedSeqNum = options.expectedSeqNum ?? 1;
            const tx = staking
                .connect(indexer)
                .transferStakeToL2(l2Beneficiary, amountToSend, maxGas, gasPriceBid, maxSubmissionCost, {
                value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
            });
            const expectedFunctionData = utils_1.defaultAbiCoder.encode(['tuple(address)'], [[l2Indexer.address]]);
            const expectedCallhookData = utils_1.defaultAbiCoder.encode(['uint8', 'bytes'], [(0, sdk_1.toBN)(0), expectedFunctionData]);
            const expectedL2Data = await l1GraphTokenGateway.getOutboundCalldata(grt.address, staking.address, l2StakingMock.address, amountToSend, expectedCallhookData);
            await (0, chai_1.expect)(tx)
                .emit(l1GraphTokenGateway, 'TxToL2')
                .withArgs(staking.address, l2GRTGatewayMock.address, (0, sdk_1.toBN)(expectedSeqNum), expectedL2Data);
        };
        beforeEach(async function () {
            await staking.connect(indexer).stake(tokensToStake);
        });
        describe('receive()', function () {
            it('should not allow receiving funds from a random address', async function () {
                const tx = indexer.sendTransaction({
                    to: staking.address,
                    value: (0, utils_1.parseEther)('1'),
                });
                await (0, chai_1.expect)(tx).revertedWith('Only transfer tool can send ETH');
            });
            it('should allow receiving funds from the transfer tool', async function () {
                const impersonatedTransferTool = await sdk_1.helpers.impersonateAccount(l1GraphTokenLockTransferTool.address);
                const tx = impersonatedTransferTool.sendTransaction({
                    to: staking.address,
                    value: (0, utils_1.parseEther)('1'),
                });
                await (0, chai_1.expect)(tx).to.not.be.reverted;
            });
        });
        describe('transferStakeToL2', function () {
            it('should not allow transferring if the protocol is partially paused', async function () {
                await controller.connect(governor).setPartialPaused(true);
                const tx = staking
                    .connect(indexer)
                    .transferStakeToL2(l2Indexer.address, tokensToStake.sub(minimumIndexerStake), maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                await (0, chai_1.expect)(tx).revertedWith('Partial-paused');
            });
            it('should not allow transferring but leaving less than the minimum indexer stake', async function () {
                const tx = staking
                    .connect(indexer)
                    .transferStakeToL2(l2Indexer.address, tokensToStake.sub(minimumIndexerStake).add(1), maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                await (0, chai_1.expect)(tx).revertedWith('!minimumIndexerStake remaining');
            });
            it('should not allow transferring less than the minimum indexer stake the first time', async function () {
                const tx = staking
                    .connect(indexer)
                    .transferStakeToL2(l2Indexer.address, minimumIndexerStake.sub(1), maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                await (0, chai_1.expect)(tx).revertedWith('!minimumIndexerStake sent');
            });
            it('should not allow transferring if there are tokens locked for withdrawal', async function () {
                await staking.connect(indexer).unstake(tokensToStake);
                const tx = staking
                    .connect(indexer)
                    .transferStakeToL2(l2Indexer.address, tokensToStake, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                await (0, chai_1.expect)(tx).revertedWith('tokensLocked != 0');
            });
            it('should not allow transferring to a beneficiary that is address zero', async function () {
                const tx = staking
                    .connect(indexer)
                    .transferStakeToL2(AddressZero, tokensToStake, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                await (0, chai_1.expect)(tx).revertedWith('l2Beneficiary == 0');
            });
            it('should not allow transferring the whole stake if there are open allocations', async function () {
                await allocate((0, sdk_1.toGRT)('10'));
                const tx = staking
                    .connect(indexer)
                    .transferStakeToL2(l2Indexer.address, tokensToStake, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                await (0, chai_1.expect)(tx).revertedWith('allocated');
            });
            it('should not allow transferring partial stake if the remaining indexer capacity is insufficient for open allocations', async function () {
                // We set delegation ratio == 1 so an indexer can only use as much delegation as their own stake
                await staking.connect(governor).setDelegationRatio(1);
                const tokensToDelegate = (0, sdk_1.toGRT)('202100');
                await staking.connect(delegator).delegate(indexer.address, tokensToDelegate);
                // Now the indexer has 200k tokens staked and 200k tokens delegated
                await allocate((0, sdk_1.toGRT)('400000'));
                // But if we try to transfer even 100k, we will not have enough indexer capacity to cover the open allocation
                const tx = staking
                    .connect(indexer)
                    .transferStakeToL2(l2Indexer.address, (0, sdk_1.toGRT)('100000'), maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                await (0, chai_1.expect)(tx).revertedWith('! allocation capacity');
            });
            it('should not allow transferring if the ETH sent is more than required', async function () {
                const tx = staking
                    .connect(indexer)
                    .transferStakeToL2(indexer.address, tokensToStake, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)).add(1),
                });
                await (0, chai_1.expect)(tx).revertedWith('INVALID_ETH_AMOUNT');
            });
            it('sends the tokens and a message through the L1GraphTokenGateway', async function () {
                const amountToSend = minimumIndexerStake;
                await shouldTransferIndexerStake(amountToSend);
                // Check that the indexer stake was reduced by the sent amount
                (0, chai_1.expect)((await staking.stakes(indexer.address)).tokensStaked).to.equal(tokensToStake.sub(amountToSend));
            });
            it('should allow transferring the whole stake if there are no open allocations', async function () {
                await shouldTransferIndexerStake(tokensToStake);
                // Check that the indexer stake was reduced by the sent amount
                (0, chai_1.expect)((await staking.stakes(indexer.address)).tokensStaked).to.equal(0);
            });
            it('should allow transferring partial stake if the remaining capacity can cover the allocations', async function () {
                // We set delegation ratio == 1 so an indexer can only use as much delegation as their own stake
                await staking.connect(governor).setDelegationRatio(1);
                const tokensToDelegate = (0, sdk_1.toGRT)('200000');
                await staking.connect(delegator).delegate(indexer.address, tokensToDelegate);
                // Now the indexer has 200k tokens staked and 200k tokens delegated,
                // but they allocate 200k
                await allocate((0, sdk_1.toGRT)('200000'));
                // If we transfer 100k, we will still have enough indexer capacity to cover the open allocation
                const amountToSend = (0, sdk_1.toGRT)('100000');
                await shouldTransferIndexerStake(amountToSend);
                // Check that the indexer stake was reduced by the sent amount
                (0, chai_1.expect)((await staking.stakes(indexer.address)).tokensStaked).to.equal(tokensToStake.sub(amountToSend));
            });
            it('allows transferring several times to the same beneficiary', async function () {
                // Stake a bit more so we're still over the minimum stake after transferring twice
                await staking.connect(indexer).stake(tokensToStake);
                await shouldTransferIndexerStake(minimumIndexerStake);
                await shouldTransferIndexerStake((0, sdk_1.toGRT)('1000'), { expectedSeqNum: 2 });
                (0, chai_1.expect)((await staking.stakes(indexer.address)).tokensStaked).to.equal(tokensToStake.mul(2).sub(minimumIndexerStake).sub((0, sdk_1.toGRT)('1000')));
            });
            it('should not allow transferring to a different beneficiary the second time', async function () {
                await shouldTransferIndexerStake(minimumIndexerStake);
                const tx = staking.connect(indexer).transferStakeToL2(indexer.address, // Note this is different from l2Indexer used before
                minimumIndexerStake, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                await (0, chai_1.expect)(tx).revertedWith('l2Beneficiary != previous');
            });
        });
        describe('transferLockedStakeToL2', function () {
            it('should not allow transferring if the protocol is partially paused', async function () {
                await controller.connect(governor).setPartialPaused(true);
                const tx = staking
                    .connect(indexer)
                    .transferLockedStakeToL2(minimumIndexerStake, maxGas, gasPriceBid, maxSubmissionCost);
                await (0, chai_1.expect)(tx).revertedWith('Partial-paused');
            });
            it('sends a message through L1GraphTokenGateway like transferStakeToL2, but gets the beneficiary and ETH from a transfer tool contract', async function () {
                const amountToSend = minimumIndexerStake;
                await l1GraphTokenLockTransferTool.setL2WalletAddress(indexer.address, l2Indexer.address);
                const oldTransferToolEthBalance = await graph.provider.getBalance(l1GraphTokenLockTransferTool.address);
                const tx = staking
                    .connect(indexer)
                    .transferLockedStakeToL2(minimumIndexerStake, maxGas, gasPriceBid, maxSubmissionCost);
                const expectedFunctionData = utils_1.defaultAbiCoder.encode(['tuple(address)'], [[l2Indexer.address]]);
                const expectedCallhookData = utils_1.defaultAbiCoder.encode(['uint8', 'bytes'], [(0, sdk_1.toBN)(0), expectedFunctionData]);
                const expectedL2Data = await l1GraphTokenGateway.getOutboundCalldata(grt.address, staking.address, l2StakingMock.address, amountToSend, expectedCallhookData);
                await (0, chai_1.expect)(tx)
                    .emit(l1GraphTokenGateway, 'TxToL2')
                    .withArgs(staking.address, l2GRTGatewayMock.address, (0, sdk_1.toBN)(1), expectedL2Data);
                (0, chai_1.expect)(await graph.provider.getBalance(l1GraphTokenLockTransferTool.address)).to.equal(oldTransferToolEthBalance.sub(maxSubmissionCost).sub(gasPriceBid.mul(maxGas)));
            });
            it('should not allow transferring if the transfer tool contract returns a zero address beneficiary', async function () {
                const tx = staking
                    .connect(indexer)
                    .transferLockedStakeToL2(minimumIndexerStake, maxGas, gasPriceBid, maxSubmissionCost);
                await (0, chai_1.expect)(tx).revertedWith('LOCK NOT TRANSFERRED');
            });
            it('should not allow transferring if the transfer tool contract does not provide enough ETH', async function () {
                await staking
                    .connect(governor)
                    .setL1GraphTokenLockTransferTool(l1GraphTokenLockTransferToolBad.address);
                await l1GraphTokenLockTransferToolBad.setL2WalletAddress(indexer.address, l2Indexer.address);
                const tx = staking
                    .connect(indexer)
                    .transferLockedStakeToL2(minimumIndexerStake, maxGas, gasPriceBid, maxSubmissionCost);
                await (0, chai_1.expect)(tx).revertedWith('ETH TRANSFER FAILED');
            });
        });
        describe('unlockDelegationToTransferredIndexer', function () {
            beforeEach(async function () {
                await staking.connect(governor).setDelegationUnbondingPeriod(28); // epochs
            });
            it('allows a delegator to a transferred indexer to withdraw locked delegation before the unbonding period', async function () {
                const tokensToDelegate = (0, sdk_1.toGRT)('10000');
                await staking.connect(delegator).delegate(indexer.address, tokensToDelegate);
                const actualDelegation = tokensToDelegate.sub(tokensToDelegate.mul(delegationTaxPPM).div(1000000));
                await staking
                    .connect(indexer)
                    .transferStakeToL2(l2Indexer.address, tokensToStake, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                await staking.connect(delegator).undelegate(indexer.address, actualDelegation);
                const tx = await staking
                    .connect(delegator)
                    .unlockDelegationToTransferredIndexer(indexer.address);
                await (0, chai_1.expect)(tx)
                    .emit(staking, 'StakeDelegatedUnlockedDueToL2Transfer')
                    .withArgs(indexer.address, delegator.address);
                const tx2 = await staking.connect(delegator).withdrawDelegated(indexer.address, AddressZero);
                await (0, chai_1.expect)(tx2)
                    .emit(staking, 'StakeDelegatedWithdrawn')
                    .withArgs(indexer.address, delegator.address, actualDelegation);
            });
            it('rejects calls if the protocol is partially paused', async function () {
                await controller.connect(governor).setPartialPaused(true);
                const tx = staking.connect(delegator).unlockDelegationToTransferredIndexer(indexer.address);
                await (0, chai_1.expect)(tx).revertedWith('Partial-paused');
            });
            it('rejects calls if the indexer has not transferred their stake to L2', async function () {
                const tokensToDelegate = (0, sdk_1.toGRT)('10000');
                await staking.connect(delegator).delegate(indexer.address, tokensToDelegate);
                const tx = staking.connect(delegator).unlockDelegationToTransferredIndexer(indexer.address);
                await (0, chai_1.expect)(tx).revertedWith('indexer not transferred');
            });
            it('rejects calls if the indexer has only transferred part of their stake but not all', async function () {
                const tokensToDelegate = (0, sdk_1.toGRT)('10000');
                await staking.connect(delegator).delegate(indexer.address, tokensToDelegate);
                await staking
                    .connect(indexer)
                    .transferStakeToL2(l2Indexer.address, minimumIndexerStake, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                const tx = staking.connect(delegator).unlockDelegationToTransferredIndexer(indexer.address);
                await (0, chai_1.expect)(tx).revertedWith('indexer not transferred');
            });
            it('rejects calls if the delegator has not undelegated first', async function () {
                const tokensToDelegate = (0, sdk_1.toGRT)('10000');
                await staking.connect(delegator).delegate(indexer.address, tokensToDelegate);
                await staking
                    .connect(indexer)
                    .transferStakeToL2(l2Indexer.address, tokensToStake, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                const tx = staking.connect(delegator).unlockDelegationToTransferredIndexer(indexer.address);
                await (0, chai_1.expect)(tx).revertedWith('! locked');
            });
            it('rejects calls if the caller is not a delegator', async function () {
                await staking
                    .connect(indexer)
                    .transferStakeToL2(l2Indexer.address, tokensToStake, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                const tx = staking.connect(delegator).unlockDelegationToTransferredIndexer(indexer.address);
                // The function checks for tokensLockedUntil so this is the error we should get:
                await (0, chai_1.expect)(tx).revertedWith('! locked');
            });
        });
        describe('transferDelegationToL2', function () {
            it('rejects calls if the protocol is partially paused', async function () {
                await controller.connect(governor).setPartialPaused(true);
                const tx = staking
                    .connect(delegator)
                    .transferDelegationToL2(indexer.address, l2Delegator.address, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                await (0, chai_1.expect)(tx).revertedWith('Partial-paused');
            });
            it('rejects calls if the delegated indexer has not transferred stake to L2', async function () {
                const tokensToDelegate = (0, sdk_1.toGRT)('10000');
                await staking.connect(delegator).delegate(indexer.address, tokensToDelegate);
                const tx = staking
                    .connect(delegator)
                    .transferDelegationToL2(indexer.address, l2Delegator.address, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                await (0, chai_1.expect)(tx).revertedWith('indexer not transferred');
            });
            it('rejects calls if the beneficiary is zero', async function () {
                const tokensToDelegate = (0, sdk_1.toGRT)('10000');
                await staking.connect(delegator).delegate(indexer.address, tokensToDelegate);
                await staking
                    .connect(indexer)
                    .transferStakeToL2(l2Indexer.address, minimumIndexerStake, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                const tx = staking
                    .connect(delegator)
                    .transferDelegationToL2(indexer.address, AddressZero, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                await (0, chai_1.expect)(tx).revertedWith('l2Beneficiary == 0');
            });
            it('rejects calls if the delegator has tokens locked for undelegation', async function () {
                const tokensToDelegate = (0, sdk_1.toGRT)('10000');
                await staking.connect(delegator).delegate(indexer.address, tokensToDelegate);
                await staking
                    .connect(indexer)
                    .transferStakeToL2(l2Indexer.address, minimumIndexerStake, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                await staking.connect(delegator).undelegate(indexer.address, (0, sdk_1.toGRT)('1'));
                const tx = staking
                    .connect(delegator)
                    .transferDelegationToL2(indexer.address, l2Delegator.address, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                await (0, chai_1.expect)(tx).revertedWith('tokensLocked != 0');
            });
            it('rejects calls if the delegator has no tokens delegated to the indexer', async function () {
                await staking
                    .connect(indexer)
                    .transferStakeToL2(l2Indexer.address, minimumIndexerStake, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                const tx = staking
                    .connect(delegator)
                    .transferDelegationToL2(indexer.address, l2Delegator.address, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                await (0, chai_1.expect)(tx).revertedWith('delegation == 0');
            });
            it('sends all the tokens delegated to the indexer to the beneficiary on L2, using the gateway', async function () {
                const tokensToDelegate = (0, sdk_1.toGRT)('10000');
                await staking.connect(delegator).delegate(indexer.address, tokensToDelegate);
                const actualDelegation = tokensToDelegate.sub(tokensToDelegate.mul(delegationTaxPPM).div(1000000));
                await staking
                    .connect(indexer)
                    .transferStakeToL2(l2Indexer.address, minimumIndexerStake, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                const expectedFunctionData = utils_1.defaultAbiCoder.encode(['tuple(address,address)'], [[l2Indexer.address, l2Delegator.address]]);
                const expectedCallhookData = utils_1.defaultAbiCoder.encode(['uint8', 'bytes'], [(0, sdk_1.toBN)(1), expectedFunctionData]);
                const expectedL2Data = await l1GraphTokenGateway.getOutboundCalldata(grt.address, staking.address, l2StakingMock.address, actualDelegation, expectedCallhookData);
                const tx = staking
                    .connect(delegator)
                    .transferDelegationToL2(indexer.address, l2Delegator.address, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                // seqNum is 2 because the first bridge call was in transferStakeToL2
                await (0, chai_1.expect)(tx)
                    .emit(l1GraphTokenGateway, 'TxToL2')
                    .withArgs(staking.address, l2GRTGatewayMock.address, (0, sdk_1.toBN)(2), expectedL2Data);
                await (0, chai_1.expect)(tx)
                    .emit(staking, 'DelegationTransferredToL2')
                    .withArgs(delegator.address, l2Delegator.address, indexer.address, l2Indexer.address, actualDelegation);
            });
            it('sets the delegation shares to zero so cannot be called twice', async function () {
                const tokensToDelegate = (0, sdk_1.toGRT)('10000');
                await staking.connect(delegator).delegate(indexer.address, tokensToDelegate);
                await staking
                    .connect(indexer)
                    .transferStakeToL2(l2Indexer.address, minimumIndexerStake, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                await staking
                    .connect(delegator)
                    .transferDelegationToL2(indexer.address, l2Delegator.address, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                const tx = staking
                    .connect(delegator)
                    .transferDelegationToL2(indexer.address, l2Delegator.address, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                await (0, chai_1.expect)(tx).revertedWith('delegation == 0');
            });
            it('can be called again if the delegator added more delegation (edge case)', async function () {
                const tokensToDelegate = (0, sdk_1.toGRT)('10000');
                await staking.connect(delegator).delegate(indexer.address, tokensToDelegate);
                const actualDelegation = tokensToDelegate.sub(tokensToDelegate.mul(delegationTaxPPM).div(1000000));
                await staking
                    .connect(indexer)
                    .transferStakeToL2(l2Indexer.address, minimumIndexerStake, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                await staking
                    .connect(delegator)
                    .transferDelegationToL2(indexer.address, l2Delegator.address, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                await staking.connect(delegator).delegate(indexer.address, tokensToDelegate);
                const tx = staking
                    .connect(delegator)
                    .transferDelegationToL2(indexer.address, l2Delegator.address, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                await (0, chai_1.expect)(tx)
                    .emit(staking, 'DelegationTransferredToL2')
                    .withArgs(delegator.address, l2Delegator.address, indexer.address, l2Indexer.address, actualDelegation);
            });
            it('rejects calls if the ETH value is larger than expected', async function () {
                const tokensToDelegate = (0, sdk_1.toGRT)('10000');
                await staking.connect(delegator).delegate(indexer.address, tokensToDelegate);
                await staking
                    .connect(indexer)
                    .transferStakeToL2(l2Indexer.address, minimumIndexerStake, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                const tx = staking
                    .connect(delegator)
                    .transferDelegationToL2(indexer.address, l2Delegator.address, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)).add(1),
                });
                await (0, chai_1.expect)(tx).revertedWith('INVALID_ETH_AMOUNT');
            });
        });
        describe('transferLockedDelegationToL2', function () {
            it('rejects calls if the protocol is partially paused', async function () {
                await controller.connect(governor).setPartialPaused(true);
                const tx = staking
                    .connect(delegator)
                    .transferLockedDelegationToL2(indexer.address, maxGas, gasPriceBid, maxSubmissionCost);
                await (0, chai_1.expect)(tx).revertedWith('Partial-paused');
            });
            it('sends delegated tokens to L2 like transferDelegationToL2, but gets the beneficiary and ETH from the L1GraphTokenLockTransferTool', async function () {
                const tokensToDelegate = (0, sdk_1.toGRT)('10000');
                await staking.connect(delegator).delegate(indexer.address, tokensToDelegate);
                const actualDelegation = tokensToDelegate.sub(tokensToDelegate.mul(delegationTaxPPM).div(1000000));
                await staking
                    .connect(indexer)
                    .transferStakeToL2(l2Indexer.address, minimumIndexerStake, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                const expectedFunctionData = utils_1.defaultAbiCoder.encode(['tuple(address,address)'], [[l2Indexer.address, l2Delegator.address]]);
                const expectedCallhookData = utils_1.defaultAbiCoder.encode(['uint8', 'bytes'], [(0, sdk_1.toBN)(1), expectedFunctionData]);
                const expectedL2Data = await l1GraphTokenGateway.getOutboundCalldata(grt.address, staking.address, l2StakingMock.address, actualDelegation, expectedCallhookData);
                await l1GraphTokenLockTransferTool.setL2WalletAddress(delegator.address, l2Delegator.address);
                const oldTransferToolEthBalance = await graph.provider.getBalance(l1GraphTokenLockTransferTool.address);
                const tx = staking
                    .connect(delegator)
                    .transferLockedDelegationToL2(indexer.address, maxGas, gasPriceBid, maxSubmissionCost);
                // seqNum is 2 because the first bridge call was in transferStakeToL2
                await (0, chai_1.expect)(tx)
                    .emit(l1GraphTokenGateway, 'TxToL2')
                    .withArgs(staking.address, l2GRTGatewayMock.address, (0, sdk_1.toBN)(2), expectedL2Data);
                await (0, chai_1.expect)(tx)
                    .emit(staking, 'DelegationTransferredToL2')
                    .withArgs(delegator.address, l2Delegator.address, indexer.address, l2Indexer.address, actualDelegation);
                (0, chai_1.expect)(await graph.provider.getBalance(l1GraphTokenLockTransferTool.address)).to.equal(oldTransferToolEthBalance.sub(maxSubmissionCost).sub(gasPriceBid.mul(maxGas)));
            });
            it('rejects calls if the transfer tool contract returns a zero address beneficiary', async function () {
                const tokensToDelegate = (0, sdk_1.toGRT)('10000');
                await staking.connect(delegator).delegate(indexer.address, tokensToDelegate);
                await staking
                    .connect(indexer)
                    .transferStakeToL2(l2Indexer.address, minimumIndexerStake, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                const tx = staking
                    .connect(delegator)
                    .transferLockedDelegationToL2(indexer.address, maxGas, gasPriceBid, maxSubmissionCost);
                await (0, chai_1.expect)(tx).revertedWith('LOCK NOT TRANSFERRED');
            });
            it('rejects calls if the transfer tool contract does not provide enough ETH', async function () {
                const tokensToDelegate = (0, sdk_1.toGRT)('10000');
                await staking.connect(delegator).delegate(indexer.address, tokensToDelegate);
                await staking
                    .connect(indexer)
                    .transferStakeToL2(l2Indexer.address, minimumIndexerStake, maxGas, gasPriceBid, maxSubmissionCost, {
                    value: maxSubmissionCost.add(gasPriceBid.mul(maxGas)),
                });
                await staking
                    .connect(governor)
                    .setL1GraphTokenLockTransferTool(l1GraphTokenLockTransferToolBad.address);
                await l1GraphTokenLockTransferToolBad.setL2WalletAddress(delegator.address, l2Delegator.address);
                const tx = staking
                    .connect(delegator)
                    .transferLockedDelegationToL2(indexer.address, maxGas, gasPriceBid, maxSubmissionCost);
                await (0, chai_1.expect)(tx).revertedWith('ETH TRANSFER FAILED');
            });
        });
    });
});
